import { Injectable, inject } from '@angular/core';
import { Spell } from 'src/interfaces/interfaces';
import { Subject } from 'rxjs';
import { TranslateService } from '@ngx-translate/core';

@Injectable({
  providedIn: 'root',
})
export class SpellsService {
  private translateService = inject(TranslateService);
  // PROBABLY DEPRECATED BECAUSE OF THE SPELLBOOK COMPONENT
  private _closeSubject = new Subject<number>();

  public closeSubject$ = this._closeSubject.asObservable();

  public closeAllOthers(level: number): void {
    this._closeSubject.next(level);
  }
  //

  // Custom Spells

  public customSpells: Spell[] = [];

  /**
   * Returns an array of available spells for the given level and character class.
   * The results come from the official apells array and the custom spells array.
   * @param level The level of the spell.
   * @param characterClass The character class of the spell.
   * @returns An array of available spells.
   */
  public getAvailableSpells(level: number, characterClass?: string): Spell[] {
    let result: Spell[] = [];
    // If a character class is provided, the result will be filtered by the character class.
    if (characterClass !== undefined) {
      result = this.spells
        .filter((spell) => spell.level === level)
        .filter((spell) => spell.classes.includes(characterClass));
      result.push(
        ...this.customSpells.filter((spell) => spell.level === level),
      );
    }
    // Else all spells with the given level will be provided.
    else {
      result = this.spells.filter((spell) => spell.level === level);
      result.push(
        ...this.customSpells.filter((spell) => spell.level === level),
      );
    }
    return result;
  }

  /**
   * Returns a list of SPells filtered by a given list of classes and a level.
   * @param classList The list of classes to filter the spells.
   * @param level The level of the spells.
   * @return An array of spells.
   */
  public getSpellsByClasslistAndLevel(
    classList: string[],
    level: number,
  ): Spell[] {
    let result: Spell[] = [];
    classList.forEach((className) => {
      const classSpells: Spell[] = this.getAvailableSpells(level, className);
      // Falls noch nicht in result add
      classSpells.forEach((spell) => {
        if (!result.includes(spell)) {
          result.push(spell);
        }
      });
    });
    // sort by language specific name
    if (this.translateService.getDefaultLang() == 'de') {
      result.sort((a, b) => (a.german < b.german ? -1 : 1));
    } else {
      result.sort((a, b) => (a.english < b.english ? -1 : 1));
    }
    return result;
  }

  /**
   * Deletets a single spell from the custom spells array and the full spells array.
   * @param spellToDelete The spell to be deleted from the custom spells.
   */
  public deleteCustomSpell(spellToDelete: Spell): void {
    const customIndex = this.customSpells.findIndex(
      (spell) => spell.id === spellToDelete.id,
    );
    if (customIndex > -1) {
      this.customSpells.splice(customIndex, 1);
    }
    const fullIndex = this.spells.findIndex(
      (spell) => spell.id === spellToDelete.id,
    );
    if (fullIndex > -1) {
      this.spells.splice(fullIndex, 1);
    }
  }

  //

  private spells: Spell[] = [
    // #region CANTRIPS
    {
      id: 1,
      german: 'Ausbessern',
      english: 'Mending',
      image: 'mending',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Dieser Zauber repariert eine Bruchstelle oder einen Riss in einem Gegenstand, den du berührst, wie ein zerbrochenes Kettenglied, die beiden Hälften eines zerbrochenen Schlüssels, einen zerrissenen Umhang oder einen lecken Weinschlauch. Solange der Riss oder die Bruchstelle in keiner Dimension größer als 1 Fuß ist, kannst du sie flicken, sodass keine Spur des vorherigen Schadens übrigbleibt.</p>
        <p>Der Zauber kann einen magischen Gegenstand oder ein Konstrukt auf physische Weise reparieren, der Zauber kann einem solchen Gegenstand aber nicht seine Magie wiedergeben.</p>`,
      description_en: `
        <p>This spell repairs a single break or tear in an object you touch, such as broken chain link, two halves of a broken key, a torn clack, or a leaking wineskin. As long as the break or tear is no larger than 1 foot in any dimension, you mend it, leaving no trace of the former damage.</p>
        <p>This spell can physically repair a magic item or construct, but the spell can't restore magic to such an object.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2,
      german: 'Befall',
      english: 'Infestation',
      image: 'infestation',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'poison' }],
      doesHeal: false,
      description_de: `
        <p>Du lässt eine Wolke aus Parasiten auf ein Ziel in maximal 30 Fuß Distanz localStorage. Das Ziel muss einen Konstitutionsrettungswurf bestehen, oder 1W6 Giftschaden erleiden und 5 Fuß in eine zufällige Richtung laufen. Ein W4 gibt die Richtung vor:</p>
        <ul>
          <li>1: Norden</li>
          <li>2: Süden</li>
          <li>3: Westen</li>
          <li>4: Osten</li>
        </ul>
        <p>Der Zauber löst keine Gelegenheitsangriffe aus.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um 1W6, wenn du die 5. Stufe (2W6), die 11. Stufe (3W6) und die 17. Stufe (4W6) erreichst.</p>
      `,
      description_en: `
        <p>You cause a cloud of mites, fleas, and other parasites to appear momentarily on one creature you can see within range. The target must succeed on a Constitution saving throw, or it takes 1d6 poison damage and moves 5 feet in a random direction if it can move and its speed is at least 5 feet. Roll a d4 for the direction:
        <ul>
          <li>1: north</li>
          <li>2: south</li>
          <li>3: east</li>
          <li>4: west</li>
          </ul>
        <p>This movement doesn’t provoke opportunity attacks, and if the direction rolled is blocked, the target doesn’t move.</p>
        <p>The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
      areaOfEffectType: '',
    },
    {
      id: 3,
      german: 'Blitzköder',
      english: 'Lightning Lure',
      image: 'lightningLure',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
        <p>Du wirft eine Blitzfalle auf eine Kreatur in 15 Fuß Distanz. Das Ziel muss einen Stärkerettungswurf bestehen, oder 1W8 Blitzschaden erleiden und 10 Fuß in deine Richtung gezogen werden.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um 1W8, wenn du die 5. Stufe (2W8), die 11. Stufe (3W8) und die 17. Stufe (4W8) erreichst.</p>
      `,
      description_en: `
        <p>You create a lash of lightning energy that strikes at one creature of your choice that you can see within range. The target must succeed on a Strength saving throw or be pulled up to 10 feet in a straight line toward you and then take 1d8 lightning damage if it is within 5 feet of you.</p>
        <p><b>At Higher Levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>

      `,
      school: 'evocation',
      isRanged: true,
      range: 15,
      hasAreaOfEffect: false,
      areaOfEffectType: '',
    },
    {
      id: 4,
      german: 'Botschaft',
      english: 'Message',
      image: 'message',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du deutest mit dem Finger auf eine Kreatur in Reichweite und flüsterst eine Botschaft. Das Ziel (und nur das Ziel) hört die Botschaft und kann in einem Flüstern antworten, das nur du hören kannst.</p>
      <p>Du kannst diesen Zauber durch feste Gegenstände hindurch wirken, wenn du das Ziel kennst und weißt, dass es jenseits der Barriere liegt. "Magische Stille", 1 Fuß Stein, 15 Fuß Holz, eine dünne Schicht Blei oder gewöhnliches Metall blockieren den Zauber. Der Zauber muss nicht einer geraden Linie folgen und kann frei um Ecken oder durch Öffnungen reisen.</p>`,
      description_en: `
      <p>You point your finger toward a creature within range and whisper a message. The target (and only the target) hears the message and can reply in a whisper that only you can hear.</p>
      <p>You can cast this spell through solid objects if you are familiar with the target and know it is beyond the barrier. Magical silence, 1 foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of wood blocks the spell. The spell doesn’t have to follow a straight line and can travel freely around corners or through openings.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 5,
      german: 'Donnerschlag',
      english: 'Thunderclap',
      image: 'thunderclap',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
        <p>Du erzeugst einen Knall, der bis zu 100 Fuß weit zu hören ist. Jede Kreatur in Reichweite, außer dir, muss einen Konstitutionsrettungswurf schaffen oder 1W6 Donnerschaden erleiden.</p>
        <p>Der Schaden des Zaubers erhöht sich um 1W6, wenn du die 5. Stufe (2W6), die 11. Stufe (3W6) und die 17. Stufe (4W6) erreichst.</p>
        `,
      description_en: `
        <p>You create a burst of thunderous sound that can be heard up to 100 feet away. Each creature within range, other than you, must succeed on a Constitution saving throw or take 1d6 thunder damage.</p>
        <p>The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'circle',
      length: 200,
    },
    {
      id: 6,
      german: 'Dornenpeitsche',
      english: 'Thorn Whip',
      image: 'thornWhip',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'piercing' }],
      doesHeal: false,
      description_de: `
        <p>Du erschaffst eine lange, dornige Peitsche aus Efeu oder einem anderen Pflanzenmaterial, die du auf eine Kreatur in Reichweite richtest. Dein Angriffswurf mit dem Zauberangriffswurf bestimmt, ob der Zauber trifft oder nicht. Bei einem Treffer erleidet das Ziel 1W6 Piercingschaden und du ziehst das Ziel 10 Fuß in deine Richtung. Wenn das Ziel dadurch in deine Reichweite kommt, kannst du es mit einem Angriff mit einer Nahkampfwaffe deiner Wahl angreifen.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um 1W6, wenn du die 5. Stufe (2W6), die 11. Stufe (3W6) und die 17. Stufe (4W6) erreichst.</p>`,
      description_en: `
        <p>You create a long, vine-like whip covered in thorns that lashes out at your command toward a creature in range. Make a melee spell attack against the target. If the attack hits, the creature takes 1d6 piercing damage, and if the creature is Large or smaller, you pull the creature up to 10 feet closer to you.</p>
        <p>This spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 7,
      german: 'Druidenkunst',
      english: 'Druidcraft',
      image: 'druidcraft',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du flüsterst den Geistern der Natur zu und erschaffst innerhalb der Reichweite einen der folgenden Effekte:</p>
        <ul>
          <li>Du erschaffst einen kleinen, harmlosen sensorischen Effekt der vorhersagt, wie das Wetter an deinem Aufenthaltsort für die nächsten 24 Stunden sein wird. Der Effekt könnte sich als goldene Kugel für einen klaren Himmel, als Wolke für Regen, als fallende Schneeflocken für Schnee – und so weiter – manifestieren. Dieser Effekt hält eine Runde lang an.</li>
          <li>Du sorgst dafür, dass augenblicklich eine Blume erblüht, eine Samenkapsel sich öffnet oder eine Blattknospe aufblüht.</li>
          <li>Du erschaffst einen augenblicklichen, harmlosen sensorischen Effekt, wie fallende Blätter, einen Windhauch, die Geräusche eines kleinen Tieres oder den leichten Geruch eines Stinktiers. Der Effekt muss in einem Würfel mit 5 Fuß Kantenlänge passen.</li>
          <li>Du kannst augenblicklich eine Kerze, eine Fackel oder ein kleines Lagerfeuer entzünden oder löschen.</li>
        </ul>
          `,
      description_en: `
        <p>Whispering to the spirits of nature, you create one of the following effects within range:</p>
        <ul>
          <li>You create a tiny, harmless sensory effect that predicts what the weather will be at your location for the next 24 hours. The effect might manifest as a golden orb for clear skies, a cloud for rain, falling snowflakes for snow, and so on. This effect persists for 1 round.</li>
          <li>You instantly make a flower blossom, a seed pod open, or a leaf bud bloom.</li>
          <li>You create an instantaneous, harmless sensory effect, such as falling leaves, a puff of wind, the sound of a small animal, or the faint odor of skunk. The effect must fit in a 5-foot cube.</li>
          <li>You instantly light or snuff out a candle, a torch, or a small campfire.</li>
        </ul>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 8,
      german: 'Dröhnende Klinge',
      english: 'Booming Blade',
      image: 'boomingBlade',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
      <p>Du schwingst die Waffe, mit der der Zauber gewirkt wurde, und führst mit ihr einen Nahkampfangriff gegen eine Kreatur im Umkreis von 5 Fuß um dich herum aus. Bei einem Treffer erleidet das Ziel die normalen Effekte des Waffenangriffs und wird dann bis zum Beginn deines nächsten Zuges von dröhnender Energie umhüllt. Wenn sich das Ziel bis dahin freiwillig 5 Fuß oder mehr bewegt, erleidet das Ziel 1W8 Donnerschaden und der Zauber endet.</p>
      <p>Der Schaden des Zaubers erhöht sich um 1W8, wenn du die 5. Stufe (2W8), die 11. Stufe (3W8) und die 17. Stufe (4W8) erreichst.</p>
      `,
      description_en: `
      <p>You brandish the weapon used in the spell’s casting and make a melee attack with it against one creature within 5 feet of you. On a hit, the target suffers the weapon attack’s normal effects and then becomes sheathed in booming energy until the start of your next turn. If the target willingly moves 5 feet or more before then, the target takes 1d8 thunder damage, and the spell ends.</p>
      <p>At 5th level, the melee attack deals an extra 1d8 thunder damage to the target on a hit, and the damage the target takes for moving increases to 2d8. Both damage rolls increase by 1d8 at 11th level (2d8 and 3d8) and again at 17th level (3d8 and 4d8).</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 9,
      german: 'Einfache Illusion',
      english: 'Minor Illusion',
      image: 'minorIllusion',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'wizard', 'warlock', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du erschaffst ein Geräusch oder ein Bild eines Gegenstands in Reichweite, das für die Wirkungsdauer anhält. Die Illusion endet auch, wenn du sie als Aktion aufhebst oder diesen Zauber noch einmal wirkst. Wenn du ein Geräusch erschaffst, kann die Lautstärke von einem Flüstern bis zu einem Schrei reichen. Es kann deine Stimme sein, die Stimme eines anderen, das Brüllen eines Löwen, schlagende Trommeln oder ein anderes Geräusch deiner Wahl. Das Geräusch hält für die Wirkungsdauer an, oder du machst einzelne Laute zu verschiedenen Zeiten bis der Zauber endet.</p>
      <p>Wenn du das Bildnis eines Gegenstandes erschaffst - wie das eines Stuhles, schlammiger Fußabdrücke oder einer kleinen Truhe - kann es nicht größer als ein Würfel mit einer Seitenlänge von 1,50m sein. Das Bild kann keine Geräusche, kein Licht, keine Gerüche oder andere sensorische Effekte erzeugen. Körperliche Interaktion mit dem Trugbild offenbart, dass es sich um eine Illusion handelt, weil Dinge es einfach durchdringen können. Eine Kreatur, die ihre Aktion verwendet, um das Bildnis oder das Geräusch zu untersuchen, kann erkennen, dass es sich um eine Illusion handelt, indem sie einen erfolgreichen Wurf auf Intelligenz (Nachforschungen) gegen deinen Zauberrettungswurf-SG ablegt. Wenn die Kreatur die Illusion als das durchschaut, was sie ist, dann verblast sie für die Kreatur.</p>
      `,
      description_en: `
      <p>You create a sound or an image of an object within range that lasts for the duration. The illusion also ends if you dismiss it as an action or cast this spell again.</p>
      <p>If you create a sound, its volume can range from a whisper to a scream. It can be your voice, someone else’s voice, a lion’s roar, a beating of drums, or any other sound you choose. The sound continues unabated throughout the duration, or you can make discrete sounds at different times before the spell ends.</p>
      <p>If you create an image of an object - such as a chair, muddy footprints, or a small chest - it must be no larger than a 5-foot cube. The image can’t create sound, light, smell, or any other sensory effect. Physical interaction with the image reveals it to be an illusion, because things can pass through it. A creature that uses its action to examine the image can determine that it is an illusion with a successful Intelligence (Investigation) check against your spell save DC. If a creature discerns the illusion for what it is, the illusion becomes faint to the creature.</p>
      `,
      school: 'illusion',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 10,
      german: 'Erde Formen',
      english: 'Mold Earth',
      image: 'moldEarth',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'artificer'],
      timeToCast: 0,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      description_de: `
        <p>Du wählst ein Stück Erde oder Stein in Reichweite, das du sehen kannst und das in einen Würfel von 5 Fuß passt. Du kannst es auf eine der folgenden Arten manipulieren:</p>
        <ul>
          <li>Wenn du einen Bereich mit loser Erde anvisierst, kannst du sie augenblicklich ausheben, über den Boden bewegen und bis zu 5 Fuß entfernt ablegen. Diese Bewegung hat nicht genug Kraft, um Schaden zu verursachen.</li>
          <li>Du lässt Formen, Farben oder beides auf der Erde oder den Steinen erscheinen, indem du Wörter schreibst, Bilder kreierst oder Muster formst. Die Veränderungen halten 1 Stunde lang an.</li>
          <li>Befindet sich die Erde oder der Stein, auf den du zielst, auf dem Boden, verwandelst du ihn in schwieriges Terrain. Alternativ kannst du den Boden auch in normales Gelände verwandeln, wenn er bereits schwieriges Gelände ist. Diese Veränderung hält 1 Stunde lang an.</li>
        </ul>
        <p>Wenn du diesen Zauber mehrmals wirkst, kannst du nicht mehr als zwei seiner nicht-augenblicklichen Effekte gleichzeitig aktiv haben, und du kannst einen solchen Effekt als eine Aktion abbrechen.</p>
      `,
      description_en: `
        <p>You choose a portion of dirt or stone that you can see within range and that fits within a 5-foot cube. You manipulate it in one of the following ways:</p>
        <ul>
          <li>If you target an area of loose earth, you can instantaneously excavate it, move it along the ground, and deposit it up to 5 feet away. This movement doesn’t have enough force to cause damage.</li>
          <li>You cause shapes, colors, or both to appear on the dirt or stone, spelling out words, creating images, or shaping patterns. The changes last for 1 hour.</li>
          <li>If the dirt or stone you target is on the ground, you cause it to become difficult terrain. Alternatively, you can cause the ground to become normal terrain if it is already difficult terrain. This change lasts for 1 hour.</li>
        </ul>
        <p>If you cast this spell multiple times, you can have no more than two of its non-instantaneous effects active at a time, and you can dismiss such an effect as an action.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 11,
      german: 'Erfrierung',
      english: 'Frostbite',
      image: 'frostbite',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'warlock', 'artificer', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'cold' }],
      doesHeal: false,
      description_de: `
        <p>Du verursachst betäubenden Frost auf einer Kreatur in Reichweite, die du sehen kannst. Das Ziel muss einen Rettungswurf auf Konstitution machen. Bei einem misslungenen Rettungswurf erleidet das Ziel 1W6 Kälteschaden und hat Nachteil beim nächsten Angriffswurf, den es vor dem Ende seines nächsten Zuges macht.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W6 wenn du Stufe 5 (auf 2W6), Stufe 11 (auf 3W6) und Stufe 17 (auf 4W6) erreichst.</p>
      `,
      description_en: `
        <p>You cause numbing frost to form on one creature that you can see within range. The target must make a Constitution saving throw. On a failed save, the target takes 1d6 cold damage, and it has disadvantage on the next weapon attack roll it makes before the end of its next turn.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 12,
      german: 'Feuerpfeil',
      english: 'Fire Bolt',
      image: 'fireBolt',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 10, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
        <p>Du schleuderst einen Splitter aus Feuer auf eine Kreatur in Reichweite. Lege einen Fernkampf-Zauberangriff gegen das Ziel ab. Bei einem Treffer erleidet das Ziel 1W10 Feuerschaden. Ein brennbarer Gegenstand, der von diesem Zauber getroffen wird, geht in Flammen auf, wenn er nicht getragen oder in der Hand gehalten wird.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W10 wenn du Stufe 5 (auf 2W10), Stufe 11 (auf 3W10) und Stufe 17 (auf 4W10) erreichst.</p>
      `,
      description_en: `
        <p>You hurl a mote of fire at a creature or object within range. Make a ranged spell attack against the target. On a hit, the target takes 1d10 fire damage. A flammable object hit by this spell ignites if it isn’t being worn or carried.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d10 when you reach 5th level (2d10), 11th level (3d10), and 17th level (4d10).</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 13,
      german: 'Flammen Erzeugen',
      english: 'Produce Flame',
      image: 'produceFlames',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
        <p>Eine flackernde Flamme erscheint in deiner Hand. Die Flamme bleibt für die Wirkungsdauer bestehen und beschädigt weder dich noch deine Ausrüstung. Sie strahlt in einem Radius von 15 Fuß helles Licht und in einem zusätzlichen Radius von 15 Fuß dämmriges Licht ab. Der Zauber endet, wenn du ihn als Aktion aufhebst oder noch einmal wirkst.</p>
        <p>Du kannst mit der Flamme auch angreifen, doch dies beendet den Zauber. Wenn du diesen Zauber wirkst, oder als Aktion in einem späteren Zug, kannst du die Flamme auf eine Kreatur innerhalb von 30 Fuß werfen. Lege einen Fernkampf-Zauberangriff ab. Bei einem Treffer erleidet das Ziel 1W8 Feuerschaden.</p>
        <p>Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
        <p>A flickering flame appears in your hand. The flame remains there for the duration and harms neither you nor your equipment. The flame sheds bright light in a 10-foot radius and dim light for an additional 10 feet. The spell ends if you dismiss it as an action or if you cast it again.</p>
        <p>You can also attack with the flame, although doing so ends the spell. When you cast this spell, or as an action on a later turn, you can hurl the flame at a creature within 30 feet of you. Make a ranged spell attack. On a hit, the target takes 1d8 fire damage.</p>
        <p>The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 14,
      german: 'Flammen kontrollieren',
      english: 'Control Flames',
      image: 'controlFlames',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'artificer', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du wählst ein nicht-magisches Feuer in Reichweite, das du sehen kannst, und das in einem Würfel mit einer Seitenlänge von 5 Fuß passt. Du beeinflusst es auf eine der folgenden Arten:</p>
        <ul>
          <li>Du dehnst die Flamme augenblicklich 5 Fuß in eine Richtung aus, vorausgesetzt, dass Holz oder anderer Brennstoff an der neuen Stelle vorhanden ist.</li>
          <li>Du löschst die Flammen innerhalb des Würfels augenblicklich.</li>
          <li>Du verdoppelst oder halbierst den Bereich des hellen und schwachen Lichts, das die Flamme ausstrahlt, veränderst ihre Farbe oder beides. Die Veränderung hält 1 Stunde lang an.</li>
          <li>Du lässt einfache Formen - wie die vage Form einer Kreatur, eines unbelebten Objekts oder eines Ortes - in den Flammen erscheinen und nach Belieben beleben. Die Formen bleiben 1 Stunde lang bestehen.</li>
        </ul>
        <p>Wenn du diesen Zauber mehrmals wirkst, kannst du bis zu drei seiner nicht-augenblicklichen Effekte gleichzeitig aktiv haben, und du kannst einen solchen Effekt als Aktion abbrechen.</p>
      `,
      description_en: `
        <p>You choose nonmagical flame that you can see within range and that fits within a 5-foot cube. You affect it in one of the following ways:</p>
        <ul>
          <li>You instantaneously expand the flame 5 feet in one direction, provided that wood or other fuel is present in the new location.</li>
          <li>You instantaneously extinguish the flames within the cube.</li>
          <li>You double or halve the area of bright light and dim light cast by the flame, change its color, or both. The change lasts for 1 hour.</li>
          <li>You cause simple shapes - such as the vague form of a creature, an inanimate object, or a location - to appear within the flames and animate as you like. The shapes last for 1 hour.</li>
        </ul>
        <p>If you cast this spell multiple times, you can have up to three of its non-instantaneous effects active at a time, and you can dismiss such an effect as an action.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 15,
      german: 'Freundschaft',
      english: 'Friends',
      image: 'friends',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Für die Wirkungsdauer hast du einen Vorteil auf alle Charisma-Würfe, die auf eine Kreatur deiner Wahl gewirkt werden, die dir nicht feindlich gesonnen sind. Wenn der Zauber endet, begreift die Kreatur, dass du Magie verwendet hast, um ihre Stimmung zu beeinflussen und wird feindselig. Eine Kreatur, die zu Gewalt neigt, könnte dich angreifen. Andere Kreaturen könnten sich (nach Entscheidung des Spielleiters) andere Möglichkeiten suchen, um sich zu rächen, abhängig von der Art, wie du mit ihr interagiert hast.</p>
      `,
      description_en: `
        <p>For the duration, you have advantage on all Charisma checks directed at one creature of your choice that isn’t hostile toward you. When the spell ends, the creature realizes that you used magic to influence its mood and becomes hostile toward you. A creature prone to violence might attack you. Another creature might seek retribution in other ways (at the DM’s discretion), depending on the nature of your interaction with it.</p>
      `,
      school: 'enchantment',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 16,
      german: 'Gedankensplitter',
      english: 'Mind Sliver',
      image: 'mindSliver',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'intelligence',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
        <p>Du treibst einen verstörenden Splitter psychischer Energie in den Geist einer Kreatur, die du in Reichweite sehen kannst. Das Ziel muss einen erfolgreichen Rettungswurf auf Intelligenz durchführen, sonst nimmt es 1W6 psychischen Schaden und muss 1W4 von seinem nächsten Rettungswurf vor Ende deines nächsten Zuges abziehen.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W6 wenn du Stufe 5 (auf 2W6), Stufe 11 (auf 3W6) und Stufe 17 (auf 4W6) erreichst.</p>
      `,
      description_en: `
        <p>You drive a disorienting spike of psychic energy into the mind of one creature you can see within range. The target must succeed on an Intelligence saving throw or take 1d6 psychic damage and subtract 1d4 from the next saving throw it makes before the end of your next turn.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 17,
      german: 'Gehässiger Spott',
      english: 'Vicious Mockery',
      image: 'viciousMockery',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 4, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
        <p>Du rufst einer Kreatur in Richweite, die du sehen kannst, eine wahre Flut von Beleidigungen entgegen, in die subtile Verzauberungen gewebt sind. Wenn das Ziel dich hören kann (es braucht dich nicht zu verstehen), muss ihm ein Weisheitsrettungswurf gelingen, um nicht 1W4 psychischen Schaden zu erleiden und beim nächsten Angriffswurf im Nachteil zu sein, den es bis zum Einde seines nächsten Zuges ausführt.</p>
        <p><b>Auf höheren Graden:</b> Der Schaden des Zaubers steigt jeweils um 1W4 bei Erreichen der 5. (2W4), 11. (3W4) und 17. Stufe (4W4).</p>`,
      description_en: `
        <p>You unleash a string of insults laced with subtle enchantments at a creature you can see within range. If the target can hear you (though it need not understand you), it must succeed on a Wisdom saving throw or take 1d4 psychic damage and have disadvantage on the next attack roll it makes before the end of its next turn.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d4 when you reach 5th level (2d4), 11th level (3d4), and 17th level (4d4).</p>
      `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 18,
      german: 'Gift versprühen',
      english: 'Poison Spray',
      image: 'poisonSpray',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 12, damageType: 'poison' }],
      doesHeal: false,
      description_de: `
        <p>Du streckst deine Hand in Richtung einer Kreatur in Reichweite, die du sehen kannst, und erzeugst eine Wolke aus ekelhaftem Gas aus deiner Handfläche. Die Kreatur muss einen Rettungswurf auf Konstitution schaffen, sonst erleidet sie 1W12 Giftschaden.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W12 wenn du Stufe 5 (auf 2W12), Stufe 11 (auf 3W12) und Stufe 17 (auf 4W12) erreichst.</p>
      `,
      description_en: `
        <p>You extend your hand toward a creature you can see within range and project a puff of noxious gas from your palm. The creature must succeed on a Constitution saving throw or take 1d12 poison damage.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d12 when you reach 5th level (2d12), 11th level (3d12), and 17th level (4d12).</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 10,
      hasAreaOfEffect: false,
    },
    {
      id: 19,
      german: 'Grünfeuerklinge',
      english: 'Green-Flame Blade',
      image: 'greenFlameBlade',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Als Teil der Aktion, mit der du diesen Zauber wirkst, musst du einen Nahkampfangriff mit einer Waffe gegen eine Kreatur in der Reichweite des Zaubers ausführen, sonst misslingt der Zauber. Bei einem Treffer erleidet das Ziel die normalen Auswirkungen des Angriffs, und grünes Feuer springt vom Ziel auf eine andere Kreatur deiner Wahl, die du sehen kannst und die sich innerhalb von 5 Fuß zum Ziel befindet. Die zweite Kreatur erleidet Feuerschaden gleich dem Modifikator deines Attributs zum Zauberwirken.</p>
      <p><b>Auf höheren Stufen:</b> Auf der 5. Stufe fügt der Nahkampfangriff dem Ziel zusätzlich 1W8 Punkte Feuerschaden zu, und der Schaden, den das zweite Ziel erleidet, steigt um 1W8 + der Modifikator deines Attributs zum Zauberwirken. Beide Schadensarten steigen in der 11. und 17. Stufe um 1W8.</p>
      `,
      description_en: `
      <p>As part of the action used to cast this spell, you must make a melee attack with a weapon against one creature within the spell’s range, otherwise the spell fails. On a hit, the target suffers the weapon attack’s normal effects, and green fire leaps from the target to a different creature of your choice that you can see within 5 feet of it. The second creature takes fire damage equal to your spellcasting ability modifier.</p>
      <p><b>At higher levels:</b> At 5th level, the melee attack deals an extra 1d8 fire damage to the target, and the fire damage to the second creature increases to 1d8 + your spellcasting ability modifier. Both damage rolls increase by 1d8 at 11th level and 17th level.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 20,
      isCustom: false,
      german: 'Göttliche Führung',
      english: 'Guidance',
      image: 'guidance',
      classes: ['test', 'cleric', 'druid'],
      level: 0,
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        Du berührst eine bereitwillige Kreatur. Einmal vor dem Ende des Zaubers kann das Ziel mit einem W4 würfeln und das Ergebnis zu einem Attributswurf seiner Wahl addieren. Es kann mit dem W4 vor oder nach dem Attributswurf würfeln. Dann endet der Zauber.
      `,
      description_en: `
        You touch one willing creature. Once before the spell ends, the target can roll a d4 and add the number rolled to one ability check of its choice. It can roll the die before or after making the ability check. The spell then ends.
      `,
      school: 'divination',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 21,
      german: 'Heilige Flamme',
      english: 'Sacred Flame',
      image: 'sacredFlame',
      level: 0,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
      <p>Flammengleiches Licht senkt sich auf eine Kreatur in Reichweite, die du sehen kannst, herab. Die Kreatur muss einen Rettungswurf auf Geschicklichkeit schaffen, sonst erleidet sie 1W8 gleißenden Schaden. Bei einem Rettungswurf erhält das Ziel keinen Vorteil aus Deckung.</p>
      <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
      <p>Flame-like radiance descends on a creature that you can see within range. The target must succeed on a Dexterity saving throw or take 1d8 radiant damage. The target gains no benefit from cover for this saving throw.</p>
      <p><b>At higher levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 22,
      german: 'Kalte Hand',
      english: 'Chill Touch',
      image: 'chillTouch',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'necrotic' }],
      doesHeal: false,
      description_de: `
      <p>Du erschaffst eine gespenstische, skelettierte Hand im Bereich einer Kreatur in Reichweite. Mache einen Fernkampf-Zauberangriff gegen die Kreatur, um sie mit der Kälte des Grabes zu attackieren. Bei einem Treffer erleidet die Kreatur 1W8 nekrotischen Schaden und kann bis zu Beginn deines nächsten Zuges keine Trefferpunkt zurückbekommen. Bis dann hält sich die Hand an dem Ziel fest.</p>
      <p>Wenn du eine untote Kreatur triffst, hat sie außerdem Nachteil auf Angriffswürfe gegen dich bis zum Ende deines nächsten Zuges.</p>
      <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
      <p>You create a ghostly, skeletal hand in the space of a creature within range. Make a ranged spell attack against the creature to assail it with the chill of the grave. On a hit, the target takes 1d8 necrotic damage, and it can’t regain hit points until the start of your next turn. Until then, the hand clings to the target.</p>
      <p>If you hit an undead target, it also has disadvantage on attack rolls against you until the end of your next turn.</p>
      <p><b>At higher levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'necromancy',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 23,
      german: 'Klingenbann',
      english: 'Blade Ward',
      image: 'bladeWard',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du streckst die Hand aus und zeichnest ein Zeichen des Schutzes in die Luft. Bis zum Ende deines nächsten Zuges hast du Resistenz gegen Wucht-, Hieb- und Stichschaden, der von Angriffen mit Waffen verursacht wird.</p>
      `,
      description_en: `
      <p>You extend your hand and trace a sigil of warding in the air. Until the end of your next turn, you have resistance against bludgeoning, piercing, and slashing damage dealt by weapon attacks.</p>
      `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 24,
      german: 'Kältestrahl',
      english: 'Ray of Frost',
      image: 'rayOfFrost',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'cold' }],
      doesHeal: false,
      description_de: `
      <p>Ein eisiger Strahl aus blau-weißem Licht schießt auf eine Kreatur in Reichweite zu. Lege einen Fernkampf-Zauberangriff gegen das Ziel ab. Bei einem Treffer erleidet die Kreatur 1W8 Kälteschaden und ihre Bewegungsrate wird bis zum Beginn deines nächsten Zuges um 10 Fuß verringert.</p>
      <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
      <p>A frigid beam of blue-white light streaks toward a creature within range. Make a ranged spell attack against the target. On a hit, it takes 1d8 cold damage, and its speed is reduced by 10 feet until the start of your next turn.</p>
      <p><b>At higher levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 25,
      german: 'Lagerfeuer erschaffen',
      english: 'Create Bonfire',
      image: 'createBonfire',
      level: 0,
      isCustom: false,
      classes: ['test', 'artificer', 'druid', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
      <p>Du erschaffst ein Lagerfeuer in Reichweite auf dem Boden, den du sehen kannst. Bis der Zauber endet, füllt das Lagerfeuer einen Würfel von 5 Fuß. Jede Kreatur, die sich zum Zeitpunkt des Zaubers im Bereich des Feuers befindet, muss einen Rettungswurf auf Geschicklichkeit ablegen oder 1W8 Feuerschaden erleiden. Eine Kreatur muss den Rettungswurf auch machen, wenn sie den Raum des Lagerfeuers zum ersten Mal in einem Zug betritt oder ihren Zug dort beendet.</p>
      <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
      <p>You create a bonfire on ground that you can see within range. Until the spell ends, the magic bonfire fills a 5-foot cube. Any creature in the bonfire’s space when you cast the spell must succeed on a Dexterity saving throw or take 1d8 fire damage. A creature must also make the saving throw when it enters the bonfire’s space for the first time on a turn or ends its turn there.</p>
      <p><b>At higher levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 26,
      german: 'Licht',
      english: 'Light',
      image: 'light',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du berührst einen Gegenstand, der nicht größer als 10 Fuß in einer beliebigen Dimension ist. Bis der Zauber endet, strahlt das Objekt helles Licht in einem Radius von 20 Fuß und schwaches Licht für weitere 20 Fuß aus. Das Licht kann nach Belieben gefärbt werden. Wenn du das Objekt vollständig mit etwas Undurchsichtigem abdeckst, wird das Licht blockiert. Der Zauber endet, wenn du ihn erneut wirkst oder ihn als Aktion abbrichst.</p>
      <p>Wenn du ein Objekt anvisierst, das von einer feindlichen Kreatur gehalten oder getragen wird, muss diese Kreatur einen Rettungswurf auf Geschicklichkeit bestehen, um den Zauber zu vermeiden.</p>
      `,
      description_en: `
      <p>You touch one object that is no larger than 10 feet in any dimension. Until the spell ends, the object sheds bright light in a 20-foot radius and dim light for an additional 20 feet. The light can be colored as you like. Completely covering the object with something opaque blocks the light. The spell ends if you cast it again or dismiss it as an action.</p>
      <p>If you target an object held or worn by a hostile creature, that creature must succeed on a Dexterity saving throw to avoid the spell.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 27,
      german: 'Magierhand',
      english: 'Mage Hand',
      image: 'mageHand',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Eine gespenstische, schwebende Hand erscheint an einem Punkt deiner Wahl in Reichweite. Die Hand bleibt für die Wirkungsdauer bestehen oder bis du sie mit einer Aktion wegschickst. Die Hand verschwindet, wenn sie sich weiter als 30 Fuß von dir entfernt oder du den Zauber noch einmal wirkst.</p>
      <p>Du kannst eine Aktion verwenden, um die Hand zu kontrollieren. Du kannst die Hand verwenden, um Gegenstände zu manipulieren, verschlossene Türen oder Behälter zu öffnen, einen Gegenstand aus einem geöffneten Behälter zu holen oder ihn zu verstauen, oder den Inhalt einer Phiole auszugießen. Du kannst die Hand immer wenn du sie verwendest bis zu 9m weit bewegen.</p>
      <p>Die Hand kann nicht angreifen, keine magischen Gegenstände aktivieren oder mehr als 10 Pfund tragen.</p>
      `,
      description_en: `
      <p>A spectral, floating hand appears at a point you choose within range. The hand lasts for the duration or until you dismiss it as an action. The hand vanishes if it is ever more than 30 feet away from you or if you cast this spell again.</p>
      <p>You can use your action to control the hand. You can use the hand to manipulate an object, open an unlocked door or container, stow or retrieve an item from an open container, or pour the contents out of a vial. You can move the hand up to 30 feet each time you use it.</p>
      <p>The hand can’t attack, activate magic items, or carry more than 10 pounds.</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 28,
      german: 'Magiestein',
      english: 'Magic Stone',
      image: 'magicStone',
      level: 0,
      isCustom: false,
      classes: ['test', 'artificer', 'druid', 'Hexenmeister'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
      <p>Du berührst ein bis drei Kieselsteine und belegst sie mit Magie. Du oder eine andere Person kann mit einem der Kieselsteine einen Fernzauberangriff durchführen, indem du ihn wirfst oder mit einer Schleuder schleuderst. Wenn du ihn wirfst, hat er eine Reichweite von 60 Fuß. Wenn jemand anderes mit dem Kieselstein angreift, addiert dieser Angreifer deinen Zaubermodifikator, nicht den des Angreifers, zum Angriffswurf. Bei einem Treffer erleidet das Ziel Wuchtschaden in Höhe von 1W6 + deinem Zaubermodifikator. Treffer oder Fehlschuss, der Zauber endet dann auf dem Stein.</p>
      <p>Wenn du diesen Zauber erneut wirkst, endet der Zauber auf allen Steinen, die noch davon betroffen sind, vorzeitig.</p>
      `,
      description_en: `
      <p>You touch one to three pebbles and imbue them with magic. You or someone else can make a ranged spell attack with one of the pebbles by throwing it or hurling it with a sling. If thrown, it has a range of 60 feet. If someone else attacks with the pebble, that attacker adds your spellcasting ability modifier, not the attacker’s, to the attack roll. On a hit, the target takes bludgeoning damage equal to 1d6 + your spellcasting ability modifier. Hit or miss, the spell then ends on the stone.</p>
      <p>If you cast this spell again, the spell ends on any pebbles still affected by your previous casting.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 29,
      german: 'Resistenz',
      english: 'Resistance',
      image: 'resistance',
      level: 0,
      isCustom: false,
      classes: ['test', 'cleric', 'druid', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du berührst eine willige Kreatur. Einmal bevor der Zauber endet, kann das Ziel einen W4 werfen und die gewürfelte Zahl zu einem Rettungswurf seiner Wahl addieren. Es kann den Würfel werfen, bevor oder nachdem es den Rettungswurf gemacht hat. Der Zauber endet dann.</p>
      `,
      description_en: `
      <p>You touch one willing creature. Once before the spell ends, the target can roll a d4 and add the number rolled to one saving throw of its choice. It can roll the die before or after making the saving throw. The spell then ends.</p>
      `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 30,
      german: 'Schauriger Strahl',
      english: 'Eldritch Blast',
      image: 'eldritchBlast',
      level: 0,
      isCustom: false,
      classes: ['test', 'Hexenmeister'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 10, damageType: 'force' }],
      doesHeal: false,
      description_de: `
        <p>Ein Strahl aus knisternder Energie schießt auf eine Kreatur in Reichweite zu. Lege einen Fernkampf-Zauberangriff gegen das Ziel ab. Bei einem Treffer erleidet das Ziel 1W10 Energieschaden.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W10 wenn du Stufe 5 (auf 2W10), Stufe 11 (auf 3W10) und Stufe 17 (auf 4W10) erreichst.</p>
      `,
      description_en: `
        <p>A beam of crackling energy streaks toward a creature within range. Make a ranged spell attack against the target. On a hit, the target takes 1d10 force damage.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d10 when you reach 5th level (2d10), 11th level (3d10), and 17th level (4d10).</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 31,
      german: 'Schockgriff',
      english: 'Shocking Grasp',
      image: 'shockingGrasp',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
        <p>Blitze springen aus deinen Händen und verpassen einer Kreatur, die du zu berühren versuchst, einen Schock. Führe einen Nahkampf-Zauberangriff gegen das Ziel durch. Du hast einen Vorteil beim Angriffswurf, wenn das Ziel Rüstung trägt, die aus Metall besteht. Bei einem Treffer erleidet die Kreatur 1W8 Blitzschaden und kann bis zum Beginn deines nächsten Zuges keine Reaktion durchführen.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W8 wenn du Stufe 5 (auf 2W8), Stufe 11 (auf 3W8) und Stufe 17 (auf 4W8) erreichst.</p>
      `,
      description_en: `
        <p>Lightning springs from your hand to deliver a shock to a creature you try to touch. Make a melee spell attack against the target. You have advantage on the attack roll if the target is wearing armor made of metal. On a hit, the target takes 1d8 lightning damage, and it can’t take reactions until the start of its next turn.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and 17th level (4d8).</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 32,
      german: 'Schwertexplosion',
      english: 'Sword Burst',
      image: 'swordBurst',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'force' }],
      doesHeal: false,
      description_de: `
        <p>Für einen Moment fegt eine Schar spektraler Klingen in Kreisform um dich herum. Alle Kreaturen innerhalb von 5 Fuß müssen einen Rettungswurf auf Geschicklichkeit ausführen, ansonsten erleiden sie 1W6 Energieschaden.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W6 wenn du Stufe 5 (auf 2W6), Stufe 11 (auf 3W6) und Stufe 17 (auf 4W6) erreichst.</p>
      `,
      description_en: `
        <p>You create a momentary circle of spectral blades that sweep around you. Each creature within range, other than you, must succeed on a Dexterity saving throw or take 1d6 force damage.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'spere',
      length: 5,
    },
    {
      id: 33,
      german: 'Shillelagh',
      english: 'Shillelagh',
      image: 'shillelagh',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Das Holz einer Keule oder eines Kampfstabs, die du in der Hand hältst, ist mit der Kraft der Natur durchdrungen. Während dieser Zeit kannst du deine Zauberfähigkeit anstelle von Stärke für die Angriffs- und Schadenswürfe von Nahkampfangriffen mit dieser Waffe verwenden, und der Schadenswürfel der Waffe wird zu einem W8. Die Waffe wird außerdem magisch, falls sie es nicht schon ist. Der Zauber endet, wenn du ihn erneut wirkst oder wenn du die Waffe loslässt.</p>
      `,
      description_en: `
        <p>The wood of a club or quarterstaff you are holding is imbued with nature’s power. For the duration, you can use your spellcasting ability instead of Strength for the attack and damage rolls of melee attacks using that weapon, and the weapon’s damage die becomes a d8. The weapon also becomes magical, if it isn’t already. The spell ends if you cast it again or if you let go of the weapon.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 34,
      german: 'Säurespritzer',
      english: 'Acid Splash',
      image: 'acidSplash',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'acid' }],
      doesHeal: false,
      description_de: `
        <p>Du schleuderst eine Blase aus Säure. Wähle eine Kreatur in Reichweite oder zwei Kreaturen in Reichweite, die sich innerhalb von 5 Fuß voneinander befinden. Das Ziel muss einen Rettungswurf auf Geschicklichkeit bestehen oder erleidet 1W6 Säureschaden.</p>
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um je 1W6 wenn du Stufe 5 (auf 2W6), Stufe 11 (auf 3W6) und Stufe 17 (auf 4W6) erreichst.</p>
      `,
      description_en: `
        <p>You hurl a bubble of acid. Choose one creature within range, or choose two creatures within range that are within 5 feet of each other. A target must succeed on a Dexterity saving throw or take 1d6 acid damage.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 35,
      german: 'Tanzende Lichter',
      english: 'Dancing Lights',
      image: 'dancingLights',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du erschaffst bis zu vier fackelgroße Lichter in Reichweite, die als Fackeln, Laternen oder leuchtende Kugeln erscheinen, die für die Dauer der Aktion in der Luft schweben. Du kannst die vier Lichter auch zu einer leuchtenden, vage humanoiden Form von mittlerer Größe kombinieren. Unabhängig davon, welche Form du wählst, verströmt jedes Licht ein schwaches Licht in einem Radius von 10 Fuß. Als Bonusaktion in deinem Zug kannst du die Lichter bis zu 300 Fuß weit an einen neuen Platz in Reichweite bewegen. Ein Licht muss sich in einem Umkreis von 20 Fuß um ein anderes durch diesen Zauber erzeugtes Licht befinden, und ein Licht erlischt, wenn es den Wirkungsbereich des Zaubers verlässt.</p>
      `,
      description_en: `
        <p>You create up to four torch-sized lights within range, making them appear as torches, lanterns, or glowing orbs that hover in the air for the duration. You can also combine the four lights into one glowing vaguely humanoid form of Medium size. Whichever form you choose, each light sheds dim light in a 10-foot radius.</p>
        <p>As a bonus action on your turn, you can move the lights up to 60 feet to a new spot within range. A light must be within 20 feet of another light created by this spell, and a light winks out if it exceeds the spell’s range.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 36,
      german: 'Taschenspielerei',
      english: 'Prestidigitation',
      image: 'prestidigitation',
      level: 0,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Dies ist ein einfacher Zaubertrick, den Zauberlehrlinge zum Üben verwenden. Du erzeugst einen der folgenden magischen Effekte in Reichweite:</p>
        <ul>
          <li>Du erzeugst einen unmittelbaren, harmlosen sensorischen Effekt, etwa einen Funkenregen, einen Windstoß, eine leise Melodie oder einen merkwürdigen Geruch.</li>
          <li>Du entzündest oder löschst unmittelbar eine Kerze, eine Fackel oder ein kleines Lagerfeuer.</li>
          <li>Du kannst unmittelbar einen Gegenstand, der nicht größer als ein Würfel mit 1 Fuß Kantenlänge ist, säubern oder verschmutzen.</li>
          <li>Du kannst nichtlebendes Material, das nicht größer als ein Würfel mit 1 Fuß Kantenlänge ist, abkühlen, erhitzen oder würzen. Dies hält eine Stunde lang an.</li>
          <li>Du lässt einen Farbfleck, ein Mal oder ein Symbol eine Stunde lang auf einem Gegenstand oder einer Oberfläche erscheinen.</li>
          <li>Du erzeugst ein nichtmagisches Schmuckstück oder ein illusorisches Bild, das in deine Hand passt und das bis zum Ende deines nächsten Zuges erhalten bleibt.</li>
        </ul>
        <p>Wirkst du diesen Zauber mehrmals, können bis zu drei der langfristigen Effekte gleichzeitig aktiv sein. Du kannst einen solchen Effekt als Aktion beenden.</p>
      `,
      description_en: `
        <p>This spell is a minor magical trick that novice spellcasters use for practice. You create one of the following magical effects within range:</p>
        <ul>
          <li>You create an instantaneous, harmless sensory effect, such as a shower of sparks, a puff of wind, faint musical notes, or an odd odor.</li>
          <li>You instantaneously light or snuff out a candle, a torch, or a small campfire.</li>
          <li>You instantaneously clean or soil an object no larger than 1 cubic foot.</li>
          <li>You chill, warm, or flavor up to 1 cubic foot of nonliving material for 1 hour.</li>
          <li>You make a color, a small mark, or a symbol appear on an object or a surface for 1 hour.</li>
          <li>You create a nonmagical trinket or an illusory image that can fit in your hand and that lasts until the end of your next turn.</li>
        </ul>
        <p>If you cast this spell multiple times, you can have up to three of its non-instantaneous effects active at a time, and you can dismiss such an effect as an action.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 10,
      hasAreaOfEffect: false,
    },
    {
      id: 37,
      isCustom: false,
      german: 'Thaumaturgie',
      english: 'Thaumaturgy',
      image: 'thaumaturgy',
      classes: ['test', 'cleric'],
      level: 0,
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du manifestierst ein kleines Wunder, ein Zeichen übernatürlicher Macht, in Reichweite. Du kannst einen der folgenden magischen Effekte in Reichweite erschaffen:</p>
        <ul>
          <li>Deine Stimme dröhnt für 1 Minute dreimal so laut, wie normal.</li>
          <li>Du lässt Flammen flackern, heller oder dunkler werden oder die Farbe wechseln. All dies hält 1 Minute an.</li>
          <li>Du erzeugst für 1 Minute harmlose Erschütterungen im Boden.</li>
          <li>Du erschaffst ein kurzes Geräusch, das von einem Punkt deiner Wahl in Reichweite ertönt, wie ein Grollen von Donner, den Ruf eines Raben oder ein unheilvolles Flüstern.</li>
          <li>Du lässt eine nicht verriegelte Tür oder ein Fenster auffliegen oder zuschlagen.</li>
          <li>Du änderst für 1 Minute das Aussehen deiner Augen.</li>
        </ul>
        <p>Wenn du diesen Zauber mehrmals wirkst, kannst du bis zu drei der nicht sofortigen Effekte gleichzeitig aktiviert haben und du kannst solche Effekte als Aktion aufheben.</p>
      `,
      description_en: `
        <p>You manifest a minor wonder, a sign of supernatural power, within range. You create one of the following magical effects within range.</p>
        <ul>
          <li>Your voice booms up to three times as loud as normal for 1 minute.</li>
          <li>You cause flames to flicker, brighten, dim, or change color for 1 minute.</li>
          <li>> You cause harmless tremors in the ground for 1 minute.</li>
          <li>You create an instantaneous sound that originates from a point of your choice within range, such as a rumble of thunder, the cry of a raven, or ominous whispers.</li>
          <li>You instantaneously cause an unlocked door or window to fly open or slam shut.</li>
          <li>You alter the appearance of your eyes for 1 minute.</li>
        <p>If you cast this spell multiple times, you can have up to three of its 1-minute effects active at a time, and you can dismiss such an effect as an action.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 38,
      isCustom: false,
      german: 'Totenläuten',
      english: 'Toll the Dead',
      image: 'tollTheDead',
      classes: ['test', 'cleric', 'warlock', 'wizard'],
      level: 0,
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'necrotic' }],
      doesHeal: false,
      description_de: `
        <p>Du zeigen auf eine Kreatur, die Sie in Reichweite sehen können, und der Klang einer traurigen Glocke erfüllt für einen Moment die Luft um sie herum. Dem Ziel muss ein Weisheitswurf gelingen oder es erleidet 1W8 nekrotischen Schaden. Wenn dem Ziel einer seiner Trefferpunkte fehlt, erleidet es stattdessen 1W12 nekrotischen Schaden.
        <p><b>Auf höheren Stufen:</b> Der Schaden des Zaubers erhöht sich um einen Würfel, wenn du Stufe 5 (2W8 oder 2W12), 11 (3W8 oder 3W12) und 17 (4W8 oder 4W12) erreichst.</p>
      `,
      description_en: `
        <p>You point at one creature you can see within range, and the sound of a dolorous bell fills the air around it for a moment. The target must succeed on a Wisdom saving throw or take 1d8 necrotic damage. If the target is missing any of its hit points, it instead takes 1d12 necrotic damage.<p>
        <p> <b>At Higher Levels:</b> The spell’s damage increases by one die when you reach 5th level (2d8 or 2d12), 11th level (3d8 or 3d12), and 17th level (4d8 or 4d12).<p>
      `,
      school: 'necromancy',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 39,
      german: 'Verschonung der Toten',
      english: 'Spare the Dying',
      image: 'spareTheDying',
      level: 0,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du berührst eine lebende Kreatur mit 0 Trefferpunkten. Die Kreatur wird stabilisiert. Der Zauber hat keine Auswirkungen auf Untote oder Konstrukte.</p>
      `,
      description_en: `
        <p>You touch a living creature that has 0 hit points. The creature becomes stable. This spell has no effect on undead or constructs.</p>
      `,
      school: 'necromancy',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 40,
      german: 'Urtümliche Wildheit',
      english: 'Primal Savagery',
      image: 'primalSavagery',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 10, damageType: 'acid' }],
      doesHeal: false,
      description_de: `
        <p>Deine Zähne oder Fingernägel werden zu scharfen Klauen, die du als Nahkampfwaffe verwenden kannst. Du führst einen Nahkampf-Zauberangriff gegen eine Kreatur deiner Wahl durch. Bei einem Treffer erleidet die Kreatur 1W10 Säureschaden.</p>
        <p><b>Auf höheren Graden:</b> Der Schaden des Zaubers erhöht sich um je 1W10 wenn du Stufe 5 (auf 2W10), Stufe 11 (auf 3W10) und Stufe 17 (auf 4W10) erreichst.</p>
      `,
      description_en: `
        <p>You channel primal magic to cause your teeth or fingernails to sharpen, ready to deliver a corrosive attack. Make a melee spell attack against one creature within 5 feet of you. On a hit, the target takes 1d10 acid damage.</p>
        <p><b>At higher levels:</b> The spell’s damage increases by 1d10 when you reach 5th level (2d10), 11th level (3d10), and 17th level (4d10).</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 41,
      german: 'Wasser formen',
      english: 'Shape Water',
      image: 'shapeWater',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du wählst eine Wasserfläche, die du in Reichweite sehen kannst und die in einen Würfel von 5 Fuß passt. Du manipulierst es auf eine der folgenden Arten:</p>
        <ul>
          <li>Du bewegst das Wasser augenblicklich oder änderst es auf andere Weise, wie du es dir vorstellst, bis zu 5 Fuß in jede Richtung. Diese Bewegung hat nicht genug Kraft, um Schaden zu verursachen.</li>
          <li>Du bringst das Wasser dazu, sich zu einfachen Formen zu formen und sich auf deine Anweisung hin zu bewegen. Diese Veränderung hält 1 Stunde lang an.</li>
          <li>Du veränderst die Farbe oder Deckkraft des Wassers. Das Wasser muss immer auf dieselbe Weise verändert werden. Diese Veränderung hält 1 Stunde lang an.</li>
          <li>Du frierst das Wasser ein, vorausgesetzt, es befinden sich keine Lebewesen darin. Das Wasser taut in 1 Stunde wieder auf.</li>
        </ul>
        <p>Wenn du diesen Zauber mehrmals wirkst, kannst du nicht mehr als zwei seiner nicht sofortigen Effekte gleichzeitig aktiv haben, und du kannst einen solchen Effekt als Aktion abbrechen.</p>
      `,
      description_en: `
      <p>You choose an area of water that you can see within range and that fits within a 5-foot cube. You manipulate it in one of the following ways:</p>
      <ul>
        <li>You instantaneously move or otherwise change the flow of the water as you direct, up to 5 feet in any direction. This movement doesn’t have enough force to cause damage.</li>
        <li>You cause the water to form into simple shapes and animate at your direction. This change lasts for 1 hour.</li>
        <li>You change the water’s color or opacity. The water must be changed in the same way throughout. This change lasts for 1 hour.</li>
        <li>You freeze the water, provided that there are no creatures in it. The water unfreezes in 1 hour.</li>
      </ul>
      <p>If you cast this spell multiple times, you can have no more than two of its non-instantaneous effects active at a time, and you can dismiss such an effect as an action.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 42,
      german: 'Windböe',
      english: 'Gust',
      image: 'gust',
      level: 0,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du ergreifst die Luft und zwingst sie, einen der folgenden Effekte an einem Punkt in Reichweite, den du sehen kannst, zu erzeugen:</p>
        <ul>
          <li>Eine Kreatur mittlerer oder kleiner Größe, die du auswählst, muss einen Rettungswurf auf Stärke bestehen oder wird bis zu 5 Fuß von dir weggestoßen.</li>
          <li>Du erzeugst einen kleinen Luftstoß, der einen Gegenstand bewegen kann, der weder gehalten noch getragen wird und nicht mehr als 5 Pfund wiegt. Der Gegenstand wird bis zu 10 Fuß von dir weggeschoben. Er wird nicht mit so viel Kraft gestoßen, dass er Schaden anrichtet.</li>
          <li>Du erzeugst mit Luft einen harmlosen sensorischen Effekt, z.B. das Rascheln von Blättern, das Zuschlagen von Fensterläden durch den Wind oder das Kräuseln deiner Kleidung in einer Brise.</li>
        </ul>
      `,
      description_en: `
      <p>You seize the air and compel it to create one of the following effects at a point you can see within range:</p>
      <ul>
        <li>One Medium or smaller creature that you choose must succeed on a Strength saving throw or be pushed up to 5 feet away from you.</li>
        <li>You create a small blast of air capable of moving one object that is neither held nor carried and that weighs no more than 5 pounds. The object is pushed up to 10 feet away from you. It isn’t pushed with enough force to cause damage.</li>
        <li>You create a harmless sensory effect using air, such as causing leaves to rustle, wind to slam shutters shut, or your clothing to ripple in a breeze.</li>
      </ul>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 43,
      german: 'Wort des Strahlens',
      english: 'Word of Radiance',
      image: 'wordOfRadiance',
      level: 0,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
      <p>Du sprichst ein göttliches Wort aus, und ein brennendes Gleißen bricht aus dir heraus. Jede Kreatur deiner Wahl, welche du innerhalb der Reichweite sehen kannst, muss einen Konstitutionsrettungswurf ablegen oder erleidet 1W6 gleißenden Schaden.</p>
      <p><b>Auf höheren Graden:</b> Der Schaden des Zaubers erhöht sich um je 1W6 wenn du Stufe 5 (auf 2W6), Stufe 11 (auf 3W6) und Stufe 17 (auf 4W6) erreichst.</p>
      `,
      description_en: `
      <p>You utter a divine word, and burning radiance erupts from you. Each creature of your choice that you can see within range must succeed on a Constitution saving throw or take 1d6 radiant damage.</p>
      <p><b>At higher levels:</b> The spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 44,
      german: 'Zielsicherer Schlag',
      english: 'True Strike',
      image: 'trueStrike',
      level: 0,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du streckst deine Hand aus und deutest mit einem Finger auf das Ziel. Deine Magie gewährt dir einen kurzen Einblick in die Verteidigung des Ziels. In deinem nächsten Zug erhältst du Vorteil bei deinem ersten Angriffswurf gegen das Ziel, vorausgesetzt der Zauber ist nicht vorbei.</p>
      `,
      description_en: `
        <p>You extend your hand and point a finger at a target in range. Your magic grants you a brief insight into the target’s defenses. On your next turn, you gain advantage on your first attack roll against the target, provided that this spell hasn’t ended.</p>
      `,
      school: 'divination',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },

    // #endregion

    // #region Level 1
    {
      id: 1000,
      german: 'Alarm',
      english: 'Alarm',
      image: 'alarm',
      level: 1,
      isCustom: false,
      classes: ['test', 'ranger', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 8,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du platzierst einen Alarm gegen unerwünschtes Eindringen. Wähle eine Tür, ein Fenster oder einen Bereich innerhalb der Reichweite, der nicht größer als ein Würfel mit einer Kantenlänge von 6 Metern ist. Bis zum Ende des Zaubers wird ein Alarm ausgelöst, sobald eine winzige oder größere Kreatur den geschützten Bereich berührt oder betritt. Wenn du den Zauber sprichst, kannst du Kreaturen bestimmen, die den Alarm nicht auslösen sollen. Du wählst auch, ob der Alarm mental oder akustisch ist.</p>
        <p>Ein mentaler Alarm alarmiert dich mit einem Klingeln in deinem Kopf, wenn du dich im Umkreis von 1,5 Kilometern um den Schutzbereich befindest. Dieses Klingeln weckt dich, wenn du schläfst.</p>
        <p>Ein akustischer Alarm erzeugt den Klang einer Handglocke für 10 Sekunden im Umkreis von 18 Metern.</p>
      `,
      description_en: `
        <p>You set an alarm against unwanted intrusion. Choose a door, a window, or an area within range that is no larger than a 20-foot cube. Until the spell ends, an alarm alerts you whenever a tiny or larger creature touches or enters the warded area. When you cast the spell, you can designate creatures that won’t set off the alarm. You also choose whether the alarm is mental or audible.</p>
        <p>A mental alarm alerts you with a ping in your mind if you are within 1 mile of the warded area. This ping awakens you if you are sleeping.</p>
        <p>An audible alarm produces the sound of a hand bell for 10 seconds within 60 feet.</p>
      `,
      school: 'abjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 1001,
      german: 'Angst verursachen',
      english: 'Cause Fear',
      image: 'causeFear',
      level: 1,
      isCustom: false,
      classes: ['test', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du zwingst eine Kreatur in Reichweite, die du sehen kannst, einen Rettungswurf auf Weisheit zu machen. Wenn das Ziel scheitert, ist das Ziel verängstigt. Konstrukte und Untote sind immun gegen den Effekt. Am Ende jedes Zuges kann das Ziel den Rettungswurf wiederholen, um den Effekt zu brechen</p>
        <p><b>Auf höheren Leveln:</b> Wenn du diesen Zauber auf einem höheren Level wirkst, kannst du für jedes zusätzliche Level den Zauber auf eine weitere Person wirken.</p>
      `,
      description_en: `
        <p>You awaken the sense of mortality in one creature you can see within range. A construct or an undead is immune to this effect. The target must succeed on a Wisdom saving throw or become frightened of you until the spell ends. The frightened target can repeat the saving throw at the end of each of its turns, ending the effect on itself on a success.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st. The creatures must be within 30 feet of each other when you target them.</p>
      `,
      school: 'necromancy',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1002,
      german: 'Arme von Hadar',
      english: 'Arms of Hadar',
      image: 'armsOfHadar',
      level: 1,
      isCustom: false,
      classes: ['test', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 6, damageType: 'necrotic' }],
      doesHeal: false,
      description_de: `
        <p>Du berufst dich auf die Macht Hadars, des Dunklen Hungers. Tentakel aus finsterer Energie brechen aus dir hervor und schlagen auf alle Kreaturen innerhalb von 3 m ein. Jede Kreatur in diesem Bereich muss einen Stärkerettungswurf ablegen. Bei einem Misserfolg erleidet ein Ziel 2W6 nekrotischen Schaden und kann bis zu seinem nächsten Zug keine Reaktion verwenden. Bei einem erfolgreichen Rettungswurf nimmt die Kreatur halben Schaden, erleidet aber keine anderen Auswirkungen.</p>
        <p><b>Auf höheren Graden</b>: Wenn du diesen Zauber mit einem Zauberplatz der 2. Stufe oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You invoke the power of Hadar, the Dark Hunger. Tendrils of dark energy erupt from you and batter all creatures within 10 feet of you. Each creature in that area must make a Strength saving throw. On a failed save, a target takes 2d6 necrotic damage and can’t take reactions until its next turn. On a successful save, the creature takes half damage, but suffers no other effect.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 1003,
      german: 'Befehl',
      english: 'Command',
      image: 'command',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du sprichst einen Ein-Wort-Befehl zu einer Kreatur in Reichweite, die du sehen kannst. Das Ziel muss einen Rettungswurf auf Weisheit bestehen oder dem Befehl in seinem nächsten Zug folgen. Der Zauber hat keine Wirkung, wenn das Ziel untot ist, wenn es deine Sprache nicht versteht oder wenn dein Befehl ihm direkt schadet.</p>
          <p>Ein paar typische Befehle und ihre Effekte folgen. Du kannst auch einen anderen als den hier beschriebenen Befehl geben. Wenn du das tust, bestimmt der Spielleiter, wie sich das Ziel verhält. Wenn das Ziel deinen Befehl nicht befolgen kann, endet der Zauber.</p>
          <p><b>Annähern</b>: Das Ziel bewegt sich auf dem kürzesten und direktesten Weg auf dich zu und beendet seinen Zug, wenn es sich weniger als 1,5 Meter von dir entfernt.</p>
          <p><b>Abwerfen</b>: Das Ziel lässt alles fallen, was es in der Hand hält und beendet dann seinen Zug.</p>
          <p><b>Fliehen</b>: Das Ziel verbringt seinen Zug damit, sich auf dem schnellsten verfügbaren Weg von dir weg zu bewegen.</p>
          <p><b>Knien</b>: Das Ziel fällt auf den Boden und beendet dann seinen Zug.</p>
          <p><b>Halt</b>: Das Ziel bewegt sich nicht und führt keine Aktionen aus. Eine fliegende Kreatur bleibt in der Luft, sofern sie dazu in der Lage ist. Wenn sie sich bewegen muss, um in der Luft zu bleiben, fliegt sie die Mindestdistanz, die nötig ist, um in der Luft zu bleiben.</p>
          <p><b>Auf höheren Graden</b>: Wenn du diesen Zauber mit einem Zauberplatz des 2ten Grades oder höher wirkst, kannst du eine zusätzliche Kreatur für jeden Grad über dem ersten beeinflussen. Die Kreaturen müssen sich im Umkreis von 9 Metern befinden, wenn du sie als Ziel wählst.</p>
        `,
      description_en: `
          <p>You speak a one-word command to a creature you can see within range. The target must succeed on a Wisdom saving throw or follow the command on its next turn. The spell has no effect if the target is undead, if it doesn’t understand your language, or if your command is directly harmful to it. Some typical commands and their effects follow. You might issue a command other than one described here. If you do so, the DM determines how the target behaves. If the target can’t follow your command, the spell ends.</p>
          <p><b>Approach</b>. The target moves toward you by the shortest and most direct route, ending its turn if it moves within 5 feet of you.</p>
          <p><b>Drop</b>. The target drops whatever it is holding and then ends its turn.</p>
          <p><b>Flee</b>. The target spends its turn moving away from you by the fastest available means.</p>
          <p><b>Grovel</b>. The target falls prone and then ends its turn.</p>
          <p><b>Halt</b>. The target doesn’t move and takes no actions. A flying creature stays aloft, provided that it is able to do so. If it must move to stay aloft, it flies the minimum distance needed to remain in the air.</p>
          <p><b>At Higher Levels</b>. When you cast this spell using a spell slot of 2nd level or higher, you can affect one additional creature for each slot level above 1st. The creatures must be within 30 feet of each other when you target them.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1004,
      german: 'Brennende Hände',
      english: 'Burning Hands',
      image: 'burningHands',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
        <p>Deine Hände werden von Flammen umhüllt, die du in einem Kegel aus Feuer entfesselst. Jede Kreatur in einem 3 Meter langen Kegel, der von dir ausgeht, muss einen Geschicklichkeitsrettungswurf ablegen. Eine Kreatur nimmt 3W6 Feuerschaden bei einem Misserfolg oder die Hälfte so viel Schaden bei einem Erfolg.</p>
        <p>Das Feuer entzündet alle entzündbaren Gegenstände in dem Bereich, die nicht getragen werden.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>As you hold your hands with thumbs touching and fingers spread, a thin sheet of flames shoots forth from your outstretched fingertips. Each creature in a 15-foot cone must make a Dexterity saving throw. A creature takes 3d6 fire damage on a failed save, or half as much damage on a successful one.</p>
        <p>The fire ignites any flammable objects in the area that aren't being worn or carried.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cone',
      length: 15,
    },
    {
      id: 1005,
      german: 'Chaospfeil',
      english: 'Chaos Bolt',
      image: 'chaosBolt',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 8, damageType: 'special' }],
      doesHeal: false,
      description_de: `
        <p>Du schleuderst einen chaotischen, energiegeladenen Projektilzauber auf eine Kreatur in Reichweite. Führe einen Fernkampfangriff mit einem Zauberangriffswurf gegen das Ziel durch. Wenn du triffst, erleidet das Ziel 2W8 Schaden. Wähle dann einen der beiden Würfel. Die Zahl bestimmt die Schadensart basierend auf folgender Tabelle:</p>
        <table>
        <tr>
        <th>W8</th>
        <th>Schadensart</th>
        </tr>
        <tr>
        <td>1</td>
        <td>Säure</td>
        </tr>
        <tr>
        <td>2</td>
        <td>Kälte</td>
        </tr>
        <tr>
        <td>3</td>
        <td>Feuer</td>
        </tr>
        <tr>
        <td>4</td>
        <td>Energie</td>
        </tr>
        <tr>
        <td>5</td>
        <td>Blitz</td>
        </tr>
        <tr>
        <td>6</td>
        <td>Gift</td>
        </tr>
        <tr>
        <td>7</td>
        <td>Psychisch</td>
        </tr>
        <tr>
        <td>8</td>
        <td>Donner</td>
        </tr>
        </table>
        <p>Wenn du auf beiden W8 die gleiche Zahl würfelst, springt die chaotische Energie auf ein anderes Ziel deiner Wahl innerhalb von 30 Fuß davon. Führe einen neuen Angriffswurf gegen das neue Ziel durch und führe einen neuen Schadenswurf durch, der die chaotische Energie erneut springen lassen könnte.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden des Zaubers um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You hurl an undulating, warbling mass of chaotic energy at one creature in range. Make a ranged spell attack against the target. On a hit, the target takes 2d8 + 1d6 damage. Choose one of the d8s. The number rolled on that die determines the attack's damage type, as shown below.</p>
        <table class="wiki-content-table">
        <tr>
        <th>d8</th>
        <th>Damage Type</th>
        </tr>
        <tr>
        <td>1</td>
        <td>Acid</td>
        </tr>
        <tr>
        <td>2</td>
        <td>Cold</td>
        </tr>
        <tr>
        <td>3</td>
        <td>Fire</td>
        </tr>
        <tr>
        <td>4</td>
        <td>Force</td>
        </tr>
        <tr>
        <td>5</td>
        <td>Lightning</td>
        </tr>
        <tr>
        <td>6</td>
        <td>Poison</td>
        </tr>
        <tr>
        <td>7</td>
        <td>Psychic</td>
        </tr>
        <tr>
        <td>8</td>
        <td>Thunder</td>
        </tr>
        </table>
        <p>If you roll the same number on both d8s, the chaotic energy leaps from the target to a different creature of your choice within 30 feet of it. Make a new attack roll against the new target, and make a new damage roll, which could cause the chaotic energy to leap again.</p>
        <p>A creature can be targeted only once by each casting of this spell.</p>
        <p><strong><em>At Higher Levels.</em></strong> When you cast this spell using a spell slot of 2nd level or higher, each target takes 1d6 extra damage of the type rolled for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 1006,
      german: 'Chromatische Kugel',
      english: 'Chromatic Orb',
      image: 'chromaticOrb',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'special' }],
      doesHeal: false,
      description_de: `
        <p>Du schleuderst eine Kugel aus Energie mit einem Durchmesser von 10 Zentimetern auf eine Kreatur, in Reichweite, die du sehen kannst. Du wählst Säure, Kälte, Feuer, Blitz, Gift oder Donner für die Art der Kugel, die du erschaffst, und führst dann einen Fernkampf-Zauberangriff gegen das Ziel aus. Wenn der Angriff trifft, erleidet die Kreatur 3W8 des von dir gewählten Typs.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden des Zaubers um 1W8 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You hurl a 4-inch-diameter sphere of energy at a creature that you can see within range. You choose acid, cold, fire, lightning, poison, or thunder for the type of orb you create, and then make a ranged spell attack against the target. If the attack hits, the creature takes 3d8 damage of the type you chose.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d8 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 1007,
      german: 'Dissonantes Flüstern',
      english: 'Dissonant Whispers',
      image: 'dissonantWhispers',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 6, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
        <p>Du stimmst flüsternd eine missklingende Melodie an, die nur eine Kreatur deiner Wahl in Reichweite hören kann. Die Kreatur wird daraufhin von schrecklichen Schmerzen erfüllt. Das Ziel muss einen Weisheitsrettungswurf ablegen. Bei einem Misserfolg erleidet es 3W6 psychischen Schaden und muss sofort seine Reaktion verwenden (falls sie noch verfügbar ist), um sich so weit von dir zu entfernen, wie es seine Bewegungsrate erlaubt. Die Kreatur bewegt sich nicht in offensichtlich gefährliches Gelände, wie Feuer oder eine Grube. Bei einem erfolgreichen Rettungswurf erleidet das Ziel nur den halben Schaden und muss sich nicht wegbewegen. Eine taube Kreatur ist bei ihrem Rettungswurf automatisch erfolgreich.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden des Zaubers um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You whisper a discordant melody that only one creature of your choice within range can hear, wracking it with terrible pain. The target must make a Wisdom saving throw. On a failed save, it takes 3d6 psychic damage and must immediately use its reaction, if available, to move as far as its speed allows away from you. The creature doesn’t move into obviously dangerous ground, such as a fire or a pit. On a successful save, the target takes half as much damage and doesn’t have to move away. A deafened creature automatically succeeds on the save.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1008,
      german: 'Donnerndes Niederstrecken',
      english: 'Thunderous Smite',
      image: 'thunderousSmite',
      level: 1,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 6, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
          <p>Das nächste Mal, wenn du eine Kreatur mit einem Waffenangriff triffst, ehe dieser Zauber endet, hallt deine Waffe mit Donner wider, und der Angriff fügt dem Ziel zusätzlich 2W6 Schallschaden zu. Wenn das Ziel eine Kreatur ist, muss sie außerdem einen Rettungswurf auf Stärke schaffen, um nicht 3 Meter von dir fortgestoßen zu werden und den Zustand liegend zu erleiden.</p>
        `,
      description_en: `
          <p>The first time you hit with a melee weapon attack during this spell’s duration, your weapon rings with thunder that is audible within 300 feet of you, and the attack deals an extra 2d6 thunder damage to the target. Additionally, if the target is a creature, it must succeed on a Strength saving throw or be pushed 10 feet away from you and knocked prone.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1009,
      german: 'Donnerwoge',
      english: 'Thunderwave',
      image: 'thunderwave',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 8, damageType: 'thunder' }],
      doesHeal: false,
      description_de: `
        <pEine Woge aus donnernder Kraft geht von dir aus. Jede Kreatur in einem Würfel mit 4,5 Metern Seitenlänge muss einen Rettungswurf auf Konstitution machen. Bei einem misslungenen Rettungswurf erleidet eine Kreatur 2W8 Schallschaden und wird 3 Meter von dir weggestoßen. Bei einem erfolgreichen Rettungswurf erleidet die Kreatur den halben Schaden und wird nicht gestoßen.</p>
        <p>Außerdem werden nicht gesicherte Gegenstände, die sich vollständig innerhalb des Wirkungsbereichs aufhalten, vom Effekt des Zaubers automatisch 3 Meter von dir weggestoßen. Der Zauber gibt ein donnerndes Dröhnen ab, das auf 90 Meter hörbar ist.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>A wave of thunderous force sweeps out from you. Each creature in a 15-foot cube originating from you must make a Constitution saving throw. On a failed save, a creature takes 2d8 thunder damage and is pushed 10 feet away from you. On a successful save, the creature takes half as much damage and isn’t pushed.</p>
        <p>In addition, unsecured objects that are completely within the area of effect are automatically pushed 10 feet away from you by the spell’s effect, and the spell emits a thunderous boom audible out to 300 feet.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d8 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 15,
    },
    {
      id: 1010,
      german: 'Dornenhagel',
      english: 'Hail of Thorns',
      image: 'hailOfThorns',
      level: 1,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'Dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 10, damageType: 'piercing' }],
      doesHeal: false,
      description_de: `
        <p>Das nächste Mal, wenn du eine Kreatur mit einer Fernkampfwaffe triffst, ehe der Zauber endet, erschafft dieser Zauber einen Hagel aus Dornen, der aus deiner Fernkampfwaffe oder einem Projektil wächst. Neben dem normalen Effekt des Angriffs müssen das Ziel des Angriffs und alle Kreaturen in 1,5 Metern Umkreis einen Rettungswurf auf Geschicklichkeit machen. bei einem misslungenen Rettungswurf erleidet eine Kreatur 1W10 Stichschaden, halb so viel Schaden bei einem erfolgreichen Rettungswurf.</p>
        <p><b>Auf höheren Graden:</b>Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann steigt der Schaden für jeden Zauberplatz-grad über dem ersten um 1W10 (maximaler Schaden 6W10).</p>
      `,
      description_en: `
        <p>The next time you hit a creature with a ranged weapon attack before the spell ends, this spell creates a rain of thorns that sprouts from your ranged weapon or ammunition. In addition to the normal effect of the attack, the target of the attack and each creature within 5 feet of it must make a Dexterity saving throw. A creature takes 1d10 piercing damage on a failed save, or half as much damage on a successful one.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage (both initial and extra) increases by 1d10 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 1011,
      german: 'Eismesser',
      english: 'Ice Knife',
      image: 'iceKnife',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [
        { diceNumber: 1, diceType: 10, damageType: 'piercing' },
        { diceNumber: 2, diceType: 6, damageType: 'cold' },
      ],
      doesHeal: false,
      description_de: `
        <p>Du erschaffst einen Eissplitter und schleuderst ihn auf eine Kreatur in Reichweite. Führe einen Fernkampfzauberangriff gegen das Ziel aus. Wenn du triffst, erleidet das Ziel 1W10 Stichschaden. Der Splitter explodiert, egal ob du triffst oder nicht. Das Ziel und jede Kreatur im Umkreis von 1,5 Metern um die Stelle, an der das Eis explodierte, müssen einen Rettungswurf auf Geschicklichkeit bestehen oder 2W6 Kälteschaden erleiden.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Kälteschadens um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You create a shard of ice and fling it at one creature within range. Make a ranged spell attack against the target. On a hit, the target takes 1d10 piercing damage. Hit or miss, the shard then explodes. The target and each creature within 5 feet of it must succeed on a Dexterity saving throw or take 2d6 cold damage.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the cold damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 1012,
      german: 'Elemente absorbieren',
      english: 'Absorb Elements',
      image: 'absorbElements',
      level: 1,
      isCustom: false,
      classes: ['test', 'artificer', 'druid', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'reaction',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Der Zauber fängt einen Teil der eintreffenden Energie ein, schwächt ihre Wirkung auf dich ab und speichert sie für deinen nächsten Nahkampfangriff. Du hast bis zum Beginn deines nächsten Zuges Widerstand gegen die auslösende Schadensart. Wenn du in deinem nächsten Zug zum ersten Mal mit einem Nahkampfangriff triffst, erleidet das Ziel zusätzlich 1W6 Schaden der auslösenden Schadensart, und der Zauber endet.</p>
        <p><b>Auf höheren Graden</b>: Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>The spell captures some of the incoming energy, lessening its effect on you and storing it for your next melee attack. You have resistance to the triggering damage type until the start of your next turn. Also, the first time you hit with a melee attack on your next turn, the target takes an extra 1d6 damage of the triggering type, and the spell ends.</p>
        <p><b>At Higher Levels</b>: When you cast this spell using a spell slot of 2nd level or higher, the extra damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1013,
      german: 'Erdrütteln',
      english: 'Earth Tremor',
      image: 'earthTremor',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
        <p>Du verursachst ein Beben des Bodens in einem Radius von 3 Metern. Jede Kreatur außer dir in diesem Gebiet muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 1W6 Wuchtschaden und wird auf den Boden geworfen. Wenn der Boden in diesem Gebiet aus loser Erde oder Stein besteht, wird es zu schwierigem Gelände, bis es geräumt ist.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Slot über dem 1. Stufe.</p>
      `,
      description_en: `
        <p>You cause a tremor in the ground in a 10-foot radius. Each creature other than you in that area must make a Dexterity saving throw. On a failed save, a creature takes 1d6 bludgeoning damage and is knocked prone. If the ground in that area is loose earth or stone, it becomes difficult terrain until cleared.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 1014,
      german: 'Erzwungenes Duell',
      english: 'Compelled Duel',
      image: 'compelledDuel',
      level: 1,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du versuchst, eine Kreatur in ein Duell zu zwingen. Wähle eine Kreatur in Reichweite, die du sehen kannst. Sie muss einen Rettungswurf auf Weisheit machen. Bei einem misslungenen Rettungswurf wird die Kreatur von dir angezogen, gezwungen von deiner göttlichen Herausforderung. Für die Wirkungsdauer hat sie einen Nachteil bei Angriffswürfen gegen Kreaturen außer dir, und muss einen Rettungswurf auf Weisheit machen, immer wenn sie sich in einen Bereich bewegen will, der mehr als 9 Meter von dir entfernt ist, wenn der Rettungswurf erfolgreich ist, Schränkt der Zauber die Bewegung des Ziels für diesen Zug nicht ein.</p>
          <p>Der Zauber endet, wenn du eine andere Kreatur angreifst, wenn du einen Zauber wirkst, der auf eine feindliche Kreatur wirkt, die nicht das Ziel ist, wenn eine Kreatur, die mit dir verbündet ist, das Ziel verletzt oder einen schädigenden Zauber auf das Ziel wirkt, oder wenn du deinen Zug weiter als 9 Meter vom Ziel entfernt beendest.</p>
        `,
      description_en: `
          <p>You attempt to compel a creature into a duel. One creature that you can see within range must make a Wisdom saving throw. On a failed save, the creature is drawn to you, compelled by your divine demand. For the duration, it has disadvantage on attack rolls against creatures other than you, and must make a Wisdom saving throw each time it attempts to move to a space that is more than 30 feet away from you; if it succeeds on this saving throw, this spell doesn’t restrict the target’s movement for that turn.</p>
          <p>The spell ends if you attack any other creature, if you cast a spell that targets a hostile creature other than the target, if a creature friendly to you damages the target or casts a harmful spell on it, or if you end your turn more than 30 feet away from the target.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1015,
      german: 'Falsches Leben',
      english: 'False Life',
      image: 'falseLife',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 1, diceType: 4, additionalHeal: 4 },
      description_de: `
        <p>ndem du dich mit einem nekromantischen Abbild des Lebens verstärkst, erhältst du für die Dauer des Zaubers 1W4 + 4 temporäre Trefferpunkte.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhältst du für jeden Slot über dem 1. Stufe 5 zusätzliche temporäre Trefferpunkte.</p>
      `,
      description_en: `
        <p>Bolstering yourself with a necromantic facsimile of life, you gain 1d4 + 4 temporary hit points for the duration.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you gain 5 additional temporary hit points for each slot level above 1st.</p>
      `,
      school: 'necromancy',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1016,
      german: 'Fangschlinge',
      english: 'Snare',
      image: 'snare',
      level: 1,
      isCustom: false,
      classes: ['test', 'artificer', 'druid', 'ranger', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'permanent',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Wenn du diesen Zauber wirkst nutzt du das Seil, um einen Kreis mit 5 ft Radius zu schaffen, welcher auf dem Grund oder dem Boden liegt. Bei Abschluss des Wirkens des Zaubers verschwindet das Seil, und der Kreis wird zu einer magischen Falle. </p>
        <p> Diese Falle ist beinahe unsichtbar, und es ist ein erfolgreicher Wurf auf Intelligenz (Nachforschungen) gegen deinen Zauberrettungswurf-SG nötig, um sie zu erkennen.</p>
        <p>Die Falle löst aus, wenn eine kleine, mittelgroße oder große Kreatur sich auf den Untergrund oder den Boden im Bereich der Falle bewegt. Diese Kreatur muss einen erfolgreichen Geschicklichkeitsrettungswurf ablegen oder wird auf magische Weise in die Luft gehoben, wo sie dann kopfüber 1 m über dem Boden hängt. Die Kreatur ist bis zum Ende des Zaubers festgesetzt.</p>
        <p>Eine festgesetzte Kreatur kann ablegen am Ende jeder ihrer Züge einen Geschicklichkeitsrettungswurf ablegen, um den Effekt auf sich selbst bei einem Erfolg zu beenden.</p>
      `,
      description_en: `
        <p>While you cast this spell, you use the cord or rope to create a circle with a 5-foot radius on a flat surface within your reach. When you finish casting, the cord or rope disappears to become a magical trap.</p>
        <p>The trap is nearly invisible, requiring a successful Wisdom (Perception) check against your spell save DC to be discerned.</p>
        <p>The trap triggers when a Small, Medium, or Large creature moves onto the ground or the floor in the spell’s radius. That creature must succeed on a Dexterity saving throw or be hoisted into the air, leaving it hanging upside down 3 feet above the ground or the floor. The creature is restrained there until the spell ends.</p>
        <p>The restrained creature can make a Dexterity saving throw at the end of each of its turns, ending the effect on itself on a success.</p>
      `,
      school: 'abjuration',
      isRanged: true,
      range: 5,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 1017,
      german: 'Federfall',
      english: 'Feather Fall',
      image: 'featherFall',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'reaction',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du kannst bis zu fünf fallende Kreaturen in Reichweite auswählen und ihre Fallgeschwin-digkeit für die Zauberdauer auf 18 Meter pro Runde verlangsamen. Wenn eine Krea-tur landet bevor der Zauber endet, erleidet sie keinen Fallschaden und landet auf ihren Füßen, wonach der Zauber für diese Kreatur endet.</p>
      `,
      description_en: `
        <p>Choose up to five falling creatures within range. A falling creature’s rate of descent slows to 60 feet per round until the spell ends. If the creature lands before the spell ends, it takes no falling damage and can land on its feet, and the spell ends for that creature.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1018,
      german: 'Feenfeuer',
      english: 'Faerie Fire',
      image: 'faerieFire',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Jeder Gegenstand in einem Würfel mit 6 Metern Kantenlänge in Reichweite wird von blauem, grünen oder violettem Licht umgeben (deine Wahl). Alle Kreaturen, die sich in dem Bereich aufhalten, wenn der Zauber gewirkt wird, werden ebenfalls von Licht umgeben, wenn sie einen Geschicklichkeits-RW nicht schaffen. Für die Wirkungsdauer geben Gegenstände und betroffene Kreaturen Schwaches Licht in einem Radius von 3 Metern ab.</p>
        <p>Angriffswürfe gegen betroffene Kreaturen oder Gegenstände haben Vorteil, wenn der Angreifer sie sehen kann, und die betroffenen Kreaturen oder Gegenstände können keinen Nutzen aus Unsichtbarkeit ziehen.</p>
      `,
      description_en: `
        <p>Each object in a 20-foot cube within range is outlined in blue, green, or violet light (your choice). Any creature in the area when the spell is cast is also outlined in light if it fails a Dexterity saving throw. For the duration, objects and affected creatures shed dim light in a 10-foot radius.</p>
        <p>Any attack roll against an affected creature or object has advantage if the attacker can see it, and the affected creature or object can’t benefit from being invisible.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 1019,
      german: 'Fesselnder Schlag',
      english: 'Ensnaring Strike',
      image: 'ensnaringStrike',
      level: 1,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Das nächste Mal, dass du eine Kreatur mit einem Waffenangriff triffst, ehe dieser Zauber endet, erscheint eine sich windende Masse dorniger Ranken am Einschlagspunkt. Das Ziel muss einen Rettungswurf auf Stärke machen, um nicht bis zum Ende des Zaubers von den magischen Ranken festgesetzt zu werden. Eine große oder größere Kreatur hat Vorteil bei diesem Rettungswurf. Wenn das Ziel seinen Rettungswurf schafft, verdorren die Ranken sofort.</p>
        <p>Solange das Ziel von diesem Zauber festgesetzt ist, erleidet es zu Beginn eines jeden seiner Züge 1W6 Stichschaden. Eine Kreatur, die von den Ranken festgesetzt ist oder die die Kreatur berühren kann, kann einen Stärkewurf gegen deinen Zauberrettungswurf-SG machen. Bei einem Erfolg wird das Ziel befreit.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann steigt der Schaden für jeden Zauberplatz-Grad über dem ersten um 1W6.</p>
      `,
      description_en: `
        <p>The next time you hit a creature with a weapon attack before this spell ends, a writhing mass of thorny vines appears at the point of impact and the target must succeed on a Strength saving throw or be restrained by the magical vines. A Large or larger creature has advantage on this saving throw. While restrained by this spell, the target takes 1d6 piercing damage at the start of each of its turns. A creature restrained by the vines or one that can touch the creature can use its action to make a Strength check against your spell save DC. On a success, the target is freed.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1020,
      german: 'Gift und Krankheiten entdecken',
      english: 'Detect Poison and Disease',
      image: 'detectPoisonAndDisease',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'cleric', 'druid', 'paladin', 'ranger'],
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de:
        'Während der Wirkungsdauer nimmst du die Gegenwart und die Position von Giften, giftigen Kreaturen und Krankheiten im Abstand von bis zu neun Metern von dir wahr. Du kannst auch die Art des Gifts, der giftigen Kreatur oder der Krankheit bestimmen. Dieser Zauber durchdringt die meisten Barrieren, wird aber von 1 Fuß Stein, 2,5 Zentimetern gewöhnlichem Metall, dünnem Bleiblech sowie von einem Meter Holz oder Erde blockiert.',
      description_en:
        'For the duration, you can sense the presence and location of poisons, poisonous creatures, and diseases within 30 feet of you. You also identify the kind of poison, poisonous creature, or disease in each case. The spell can penetrate most barriers, but it is blocked by 1 foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of wood or dirt.',
      school: 'divination',
      isRanged: true,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 1021,
      german: 'Gute Beeren',
      english: 'Goodberry',
      image: 'goodberry',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 0, diceType: 0, additionalHeal: 1 },
      description_de: `
        <p>Bis zu 10 Beeren erscheinen in deiner Hand und sind bis zum Ende des Zaubers mit Magie versehen. Eine Kreatur kann eine Beere essen und erhält dadurch 1 Trefferpunkt. Die Beere bietet Nahrung für einen Tag.</p>
      `,
      description_en: `
        <p>Up to ten berries appear in your hand and are infused with magic for the duration. A creature can use its action to eat one berry. Eating a berry restores 1 hit point, and the berry provides enough nourishment to sustain a creature for one day.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1022,
      german: 'Gutes und Böses entdecken',
      english: 'Detect Evil and Good',
      image: 'detectEvilAndGood',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Für die Dauer des Zaubers weißt du, ob sich im Umkreis von 9 Metern ein Abnormaler, ein Himmlischer, ein Elementarer, ein Fee, ein Unhold oder ein Untoter aufhält und wo sich die Kreatur befindet. Ebenso weißt du, ob sich im Umkreis von 9 Metern ein magisch geweihter oder entweihter Ort oder Gegenstand befindet.</p>
          <p>Der Zauber kann die meisten Barrieren durchdringen, wird aber durch 30 Zentimeter Stein, 2,5 Zentimeter gewöhnliches Metall, eine dünne Bleiplatte oder 1 Meter Holz oder Erde blockiert.</p>
        `,
      description_en: `
          <p>For the duration, you know if there is an aberration, celestial, elemental, fey, fiend, or undead within 30 feet of you, as well as where the creature is located. Similarly, you know if there is a place or object within 30 feet of you that has been magically consecrated or desecrated.</p>
          <p>The spell can penetrate most barriers, but it is blocked by 1 foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of wood or dirt.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 1023,
      german: 'Göttliche Gunst',
      english: 'Divine Favor',
      image: 'divineFavor',
      level: 1,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 4, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
          <p>Dein Gebet verleiht dir eine göttliche Ausstrahlung. Bis der Zauber endet, verursachen deine Waffenangriffe bei einem Treffer zusätzlich 1W4 gleißenden SChaden.</p>
        `,
      description_en: `
          <p>Your prayer empowers you with divine radiance. Until the spell ends, your weapon attacks deal an extra 1d4 radiant damage on a hit.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1024,
      german: 'Heilendes Wort',
      english: 'Healing Word',
      image: 'healingWord',
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid'],
      level: 1,
      timeToCast: 1,
      cost: 'bonus',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 1, diceType: 4, additionalHeal: 0 },
      description_de: `
        <p>Eine Kreatur deiner Wahl in Reichweite, die du sehen kannst, erhält Trefferpunkte in Höhe von 1W4 + deinem Zauberwirken-Attributsmodifikator zurück. Der Zauber hat keine Auswirkungen auf Untote oder Konstrukte.</p>
        <p>Auf höheren Graden: Wirkst du diesen Zauber, indem du einen Zauberplatz des 2. Grades oder höher nutzt, steigt die Heilung für jeden Grad über dem 1. um 1W4.</p>
      `,
      description_en: `
        <p>A creature of your choice that you can see within range regains hit points equal to 1d4 + your spellcasting ability modifier. This spell has no effect on undead or constructs.</p>
        <p>At higher levels: When you cast this spell using a spell slot of 2nd level or higher, the healing increases by 1d4 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1025,
      german: 'Heiligtum',
      english: 'Sanctuary',
      image: 'sanctuary',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du schützt eine Kreatur in Reichweite vor Angriffen. Bis der Zauber endet, müssen alle Kreaturen, die die geschützte Kreatur mit einem Angriff oder einem Schädigenden Zauber attackieren wollen, zunächst einen Weisheitsrettungswurf ablegen. Bei einem misslungenen Rettungswurfmuss die Kreatur ein neues Ziel wählen, sonst verliert sie den Angriff oder Zauber. Der Zauber schützt die Kreatur nicht vor Flächeneffekten, wie der Explosion eines Feuerballs.</p>
        <p>Wenn die geschützte Kreatur einen Angriff macht oder einen Zauber wirkt, der eine gegnerische Kreatur betrifft, endet dieser Zauber.</p>
      `,
      description_en: `
        <p>You ward a creature within range against attack. Until the spell ends, any creature who targets the warded creature with an attack or a harmful spell must first make a Wisdom saving throw. On a failed save, the creature must choose a new target or lose the attack or spell. This spell doesn’t protect the warded creature from area effects, such as the explosion of a fireball.</p>
        <p>If the warded creature makes an attack or casts a spell that affects an enemy creature, this spell ends.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1026,
      german: 'Heldenmut',
      english: 'Heroism',
      image: 'heroism',
      level: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'bard', 'paladin'],
      timeToCast: 0,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Eine willige Kreatur, die du berührst, wird mit Tapferkeit durchtränkt. Bis der Zauber endet, ist die Kreatur immun gegen das Erschrecken und erhält zu Beginn jeder ihrer Runden temporäre Trefferpunkte in Höhe deines Zaubermodifikators. Wenn der Zauber endet, verliert das Ziel alle verbleibenden temporären Trefferpunkte aus diesem Zauber.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, kannst du eine zusätzliche Kreatur für jeden Zauberplatz-Grad über der Stufe 1 anvisieren.</p>
        `,
      description_en: `
        <p>A willing creature you touch is imbued with bravery. Until the spell ends, the creature is immune to being frightened and gains temporary hit points equal to your spellcasting ability modifier at the start of each of its turns. When the spell ends, the target loses any remaining temporary hit points from this spell.</p>
        <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st.</p>
      `,
      school: 'enchantment',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1027,
      german: 'Hexenpfeil',
      english: 'Witch Bolt',
      image: 'witchBolt',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 12, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
        <p>Ein Blitzbogen schießt auf eine Kreatur deiner Wahl in Reichweite zu. Führe einen Fernkampf-Zauberangriff gegen das Ziel aus. Bei einem Erfolg erleidet das Ziel 1W12 Blitzschaden, und du kannst den Zauber in jedem deiner Züge erneut wirken, ohne einen weiteren Zauberplatz zu verbrauchen.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann steigt der Schaden für jeden Zauberplatz-Grad über dem ersten um 1W12.</p>
      `,
      description_en: `
        <p>A beam of crackling, blue energy lances out toward a creature within range, forming a sustained arc of lightning between you and the target. Make a ranged spell attack against that creature. On a hit, the target takes 1d12 lightning damage, and on each of your turns for the duration, you can use your action to deal 1d12 lightning damage to the target automatically. The spell ends if you use your action to do anything else. The spell also ends if the target is ever outside the spell’s range or if it has total cover from you.
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the initial damage increases by 1d12 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1028,
      german: 'Höllischer Tadel',
      english: 'Hellish Rebuke',
      image: 'hellishRebuke',
      level: 1,
      cost: 'reaction',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'warlock'],
      timeToCast: 0,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 10, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
        <p>Du deutest mit deinem Finger und die Kreatur, die dich verletzt hat, wird für einen kurzen Moment von höllischen Flammen umzüngelt. Die Kreatur muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet das Ziel 2W10 Feuerschaden, halb so viel Schaden bei einem erfolgreichen Rettungswurf.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, steigt der Schaden für jeden Zauberplatz-Grad über dem ersten um 1W10.</p>
      `,
      description_en: `
        <p>You point your finger, and the creature that damaged you is momentarily surrounded by hellish flames. The creature must make a Dexterity saving throw. It takes 2d10 fire damage on a failed save, or half as much damage on a successful one.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d10 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1029,
      german: 'Identifizieren',
      english: 'Identify',
      image: 'identify',
      level: 1,
      timeToCast: 1,
      cost: 'minutes',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'bard', 'wizard'],
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du wählst einen Gegenstand, den du während des Zaubers berühren musst. Wenn es sich um einen magischen Gegenstand handelt, erfährst du, welche Eigenschaften er hat, wie er benutzt werden kann, ob er eine Einstimmung benötigt und wie viele Aufladungen er hat, wenn überhaupt. Du erfährst, ob der Gegenstand von Zaubern beeinflusst wird und welche das sind. Wenn der Gegenstand durch einen Zauber erschaffen wurde, erfährst du, welcher Zauber ihn erschaffen hat.</p>
        <p>Wenn du stattdessen während des Zauberns eine Kreatur berührst, erfährst du, welche Zauber, falls vorhanden, sie gerade beeinflussen.</p>
      `,
      description_en: `
        <p>You choose one object that you must touch throughout the casting of the spell. If it is a magic item or some other magic-imbued object, you learn its properties and how to use them, whether it requires attunement to use, and how many charges it has, if any. You learn whether any spells are affecting the item and what they are. If the item was created by a spell, you learn which spell created it.</p>
        <p>If you instead touch a creature throughout the casting, you learn what spells, if any, are currently affecting it.</p>
      `,
      school: 'divination',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1030,
      german: 'Illusionsschrift',
      english: 'Illusory Script',
      image: 'illusoryScript',
      level: 1,
      timeToCast: 1,
      cost: 'minutes',
      duration: 10,
      durationType: 'days',
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'wizard'],
      isRitual: true,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du schreibst einen Text auf Pergament, Papier oder ein anderes passendes Material und erfüllst die Schrift mit einer mächtigen Illusion, die für die Wirkungsdauer anhält.</p>
        <p>Für dich und alle Kreaturen, die du beim Wirken des Zaubers bestimmst, erscheint der Text normal, verfasst in deiner Handschrift. Sein Inhalt entspricht dem, den du niedergeschrieben hast. Für alle anderen erscheint der Text, als sei er in einer unbekannten oder magischen Schrift verfasst, die unverständlich ist. Alternativ kannst du den Text als vollkommen andere Nachricht erscheinen lassen, die in einer anderen Handschrift und Sprache verfasst ist, welche dir allerdings bekannt sein muss. Sollte der Zauber gebannt werden, verschwinden sowohl die echte Schrift als auch die Illusion.</p>
        <p>Eine Kreatur mit wahrer Sicht kann die verborgene Nachricht lesen.</p>
      `,
      description_en: `
        <p>You write on parchment, paper, or some other suitable writing material and imbue it with a potent illusion that lasts for the duration.</p>
        <p>To you and any creatures you designate when you cast the spell, the writing appears normal, written in your hand, and conveys whatever meaning you intended when you wrote the text. To all others, the writing appears as if it were written in an unknown or magical script that is unintelligible. Alternatively, you can cause the writing to appear to be an entirely different message, written in a different hand and language, though the language must be one you know. Should the spell be dispelled, both the original script and the illusion disappear.</p>
        <p>A creature with true sight can read the hidden message.</p>
      `,
      school: 'illusion',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1031,
      german: 'Katapult',
      english: 'Catapult',
      image: 'catapult',
      level: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'artificer', 'sorcerer', 'wizard'],
      timeToCast: 0,
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
        <p>Wähle ein unverankertes Objekt, das weniger als 5 Pfund wiegt und das du berühren kannst. Das Objekt wird in einer geraden Linie in eine Richtung deiner Wahl geschleudert, die du beim Wirken des Zaubers bestimmst. Wenn das Objekt ein Kreatur trifft, muss diese einen Geschicklichkeitsrettungswurf machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 3W8 Schaden.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, steigt das maximale Gewicht der Objekte, die du mit diesem Zauber anvisieren kannst, um 5 Pfund, und der Schaden steigt für jeden Zauberplatz-Grad über dem ersten um 1W8.</p>
        `,
      description_en: `
        <p>Choose one object weighing 1 to 5 pounds within range that isn’t being worn or carried. The object flies in a straight line up to 90 feet in a direction you choose before falling to the ground, stopping early if it impacts against a solid surface. If the object would strike a creature, that creature must make a Dexterity saving throw. On a failed save, the object strikes the target and stops moving. When the object strikes something, the object and what it strikes each take 3d8 bludgeoning damage.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the maximum weight of objects that you can target with this spell increases by 5 pounds, and the damage increases by 1d8, for each slot level above 1st.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1032,
      german: 'Lange Schritte',
      english: 'Longstrider',
      image: 'longstrider',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger', 'wizard'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Deine Schrittlänge verlängert sich für die Wirkungsdauer. Deine Geschwindigkeit erhöht sich um 10 Fuß, bis der Zauber endet.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich die Geschwindigkeit um 10 Fuß für jeden Zauberplatz-Grad über dem ersten.</p>
      `,
      description_en: `
        <p>You touch a creature. The target’s speed increases by 10 feet until the spell ends.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st.</p>
      `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1033,
      german: 'Lautloses Trugbild',
      english: 'Silent Image',
      image: 'silentImage',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du erschaffst das Trugbild eines Gegenstands, einer Kreatur oder eines anderen sichtbaren Phänomens, das nicht größer sein kann als ein Würfel mit 4,5 Metern Seitenlänge. Das Bild erscheint an einem Punkt innerhalb der Reichweite und hält für die Wirkungsdauer an. Das Bildnis ist rein visuell: es wird nicht von Geräuschen, Gerüchen oder anderen Sinneseindrücken begleitet.</p>
        <p>Du kannst deine Aktion verwenden, um das Bildnis an einen beliebigen Punkt in Reichweite zu bewegen. Wenn das Bild seine Position wechselt, kannst du sein Aussehen verändern, so dass seine Bewegung natürlich für das Bildnis erscheint.</p>
        <p>Körperliche Interaktion mit dem Trugbild offenbart, dass es sich um eine Illusion handelt, weil Dinge es einfach durchdringen können. Eine Kreatur, die ihre Aktion verwendet, um das Bildnis zu untersuchen, kann erkennen, dass es sich um eine Illusion handelt, indem sie einen erfolgreichen Wurf auf Intelligenz (Nachforschungen) ablegt. Wenn die Kreatur die Illusion als das durchschaut, was sie ist, dann kann sie durch sie hindurch blicken.</p>
      `,
      description_en: `
        <p>You create the image of an object, a creature, or some other visible phenomenon that is no larger than a 15-foot cube. The image appears at a spot within range and lasts for the duration. The image is purely visual; it isn’t accompanied by sound, smell, or other sensory effects.</p>
        <p>You can use your action to cause the image to move to any spot within range. As the image changes location, you can alter its appearance so that its movements appear natural for the image. For example, if you create an image of a creature and move it, you can alter the image so that it appears to be walking.</p>
        <p>Physical interaction with the image reveals it to be an illusion because things can pass through it. A creature that uses its action to examine the image can determine that it is an illusion with a successful Intelligence (Investigation) check against your spell save DC. If a creature discerns the illusion for what it is, the creature can see through the image, and any noise it makes sounds hollow to the creature.</p>
      `,
      school: 'illusion',
      isRanged: false,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1034,
      german: 'Lenkendes Geschoss',
      english: 'Guiding Bolt',
      image: 'guidingBolt',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'cleric'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 4, diceType: 6, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
        <p>Ein Lichtblitz schießt auf eine Kreatur deiner Wahl in Reichweite zu. Mache einen Fernkampf-Zauberangriff gegen das Ziel. Bei einem Treffer erleidet das Ziel 4W6 gleißenden Schaden, und der nächste Angriffswurf, der vor Ende deines nächsten Zuges gegen das Ziel durchgeführt wird, hat einen Vorteil, weil mysterisches Licht das Ziel zum Schimmern bringt.<p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, steigt der Schaden für jeden Zauberplatz-Grad über dem ersten um 1W6.</p>
      `,
      description_en: `
        <p>A flash of light streaks toward a creature of your choice within range. Make a ranged spell attack against the target. On a hit, the target takes 4d6 radiant damage, and the next attack roll made against this target before the end of your next turn has advantage, thanks to the mystical dim light glittering on the target until then.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d6 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 1035,
      german: 'Magie entdecken',
      english: 'Detect Magic',
      image: 'detectMagic',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isCustom: false,
      classes: [
        'test',
        'bard',
        'cleric',
        'druid',
        'paladin',
        'ranger',
        'sorcerer',
        'wizard',
      ],
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Für die Wirkungsdauer spürst du die Anwesenheit von Magie im Umkreis von 9 Metern um dich herum. Wenn du Magie auf diese Weise spürst, kannst du deine Aktion verwenden, um eine schwache Aura um eine sichtbare beliebige Kreatur oder einen Gegenstand im Wirkungsbereich zu sehen, der magisch ist, und die Schule der Magie in Erfahrung zu bringen, falls es eine gibt.</p>
        <p>Der Zauber kann die meisten Barrieren durchdringen, wird aber von 30 Zentimetern Stein, 2,5 Zentimetern gewöhnlicher Metalle, einer dünnen Schicht Blei oder 90 Zentimetern Holz oder Erde blockiert.</p>
      `,
      description_en: `
        <p>For the duration, you sense the presence of magic within 30 feet of you. If you sense magic in this way, you can use your action to see a faint aura around any visible creature or object in the area that bears magic, and you learn its school of magic, if any.</p>
        <p>The spell can penetrate most barriers, but it is blocked by 1 foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of wood or dirt.</p>
      `,
      school: 'divination',
      isRanged: true,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 1036,
      german: 'Magierrüstung',
      english: 'Mage Armor',
      image: 'mageArmor',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du berührst eine bereitwillige Kreatur, die keine Rüstung trägt. Eine schützende magische Energie umgibt sie bis der Zauber endet. Die Basis-Rüstungsklasse des Ziels entspricht 13 + Geschicklichkeitsmodifikator. Der Zauber endet, wenn das Ziel Rüstung anlegt oder du den Zauber mit einer Aktion aufhebst.</p>
      `,
      description_en: `
        <p>You touch a willing creature who isn’t wearing armor, and a protective magical force surrounds it until the spell ends. The target’s base AC becomes 13 + its Dexterity modifier. The spell ends if the target dons armor or if you dismiss the spell as an action.</p>
      `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1037,
      german: 'Magisches Geschoß',
      english: 'Magic Missile',
      image: 'magicMissile',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 4, damageType: 'force' }],
      doesHeal: false,
      description_de: `
        <p>Du erschaffst drei leuchtende Pfeile aus magischer Energie. Jeder Pfeil trifft eine Kreatur deiner Wahl in Reichweite, die du sehen kannst. Ein Pfeil fügt dem Ziel 1W4 + 1 Energieschaden zu. Die Pfeile Schlagen alle gleichzeitig ein, und du kannst sie auf eine oder mehrere Kreaturen aufteilen.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden für jeden Zauberplatz-Grad über dem ersten um 1W4.</p>
      `,
      description_en: `
        <p>You create three glowing darts of magical force. Each dart hits a creature of your choice that you can see within range. A dart deals 1d4 + 1 force damage to its target. The darts all strike simultaneously and you can direct them to hit one creature or several.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the spell creates one more dart for each slot level above 1st.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 1038,
      german: 'Mit Tieren sprechen',
      english: 'Speak with Animals',
      image: 'speakWithAnimals',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger'],
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du erhältst die Fähigkeit, für die Wirkungsdauer Tiere zu verstehen und verbal mit ihnen zu kommunizieren. Das Wissen und das Bewusstsein vieler Tiere wird durch ihre Intelligenz eingeschränkt, aber Tiere können mindestens Informationen über nahe Orte und Monsters übermitteln, auch was sie wahrnehmen können oder innerhalb des letzten Tages wahrgenommen haben. Du könntest ein Tier auch überzeugen, dir einen kleinen Gefallen zu erweisen, nach Maßgabe des Spielleiters.</p>
      `,
      description_en: `
        <p>You gain the ability to comprehend and verbally communicate with beasts for the duration. The knowledge and awareness of many beasts is limited by their intelligence, but at a minimum, beasts can give you information about nearby locations and monsters, including whatever they can perceive or have perceived within the past day. You might be able to persuade a beast to perform a small favor for you, at the DM’s discretion.</p>
      `,
      school: 'divination',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1039,
      german: 'Nahrung und Wasser reinigen',
      english: 'Purify Food and Drink',
      image: 'purifyFoodAndDrink',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'cleric', 'paladin', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Alle nichtmagischen Nahrungsmittel und Getränke in einer Sphäre mit einem Radius von 1,5 Metern, zentriert um einen Punkt deiner Wahl in Reichweite, werden gereinigt und von Giften und Krankheiten befreit.</p>
        `,
      description_en: `
          <p>All nonmagical food and drink within a 5-foot-radius sphere centered on a point of your choice within range is purified and rendered free of poison and disease.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 10,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 1040,
      german: 'Nebelwolke',
      english: 'Fog Cloud',
      image: 'fogCloud',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst einen Bereich aus Nebel mit einem Radius von 6 Metern, der auf einen Punkt deiner Wahl in Reichweite zentriert ist. Der Bereich breitet sich um Ecken aus, und der Bereich gilt als komplett verschleiert. Er bleibt für die Wirkungsdauer des Zaubers oder bis ein mittelstarker oder stärkerer Wind (mindestes 15 Kilometer pro Stunde) den Nebel auflöst.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Radius des Nebels um 3 Meter für jeden Zauberplatz-Grad über dem ersten.</p>
        `,
      description_en: `
          <p>You create a 20-foot-radius sphere of fog centered on a point within range. The sphere spreads around corners, and its area is heavily obscured. It lasts for the duration or until a wind of moderate or greater speed (at least 10 miles per hour) disperses it.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the radius of the fog increases by 20 feet for each slot level above 1st.</p>
          `,
      school: 'conjuration',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 1041,
      german: 'Person bezaubern',
      english: 'Charm Person',
      image: 'charmPerson',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du versuchst einen Humanoiden in Reichweite, den du sehen kannst, zu bezaubern. Das Ziel muss einen Rettungswurf auf Weisheit machen, was es mit Vorteil tut, wenn du oder deine Gefährten gegen es kämpfen. Wenn es den Rettungswurf nicht schafft, wird es von dir bezaubert bis der Zauber endet, oder du oder eine deiner Gefährten etwas tut, um ihm zu schaden. Die bezauberte Kreatur betrachtet dich als freundschaftliche Bekanntschaft. Wenn der Zauber endet, weiß die Kreatur, dass sie von dir bezaubert worden ist.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann kannst du den Zauber für jeden Zauberplatz-Grad über dem ersten auf eine zusätzliche Kreatur wirken. Die Kreaturen dürfen nicht weiter als 9 Meter voneinander entfernt sein, wenn du den Zauber auf sie wirkst.</p>
        `,
      description_en: `
          <p>You attempt to charm a humanoid you can see within range. It must make a Wisdom saving throw, and does so with advantage if you or your companions are fighting it. If it fails the saving throw, it is charmed by you until the spell ends or until you or your companions do anything harmful to it. The charmed creature regards you as a friendly acquaintance. When the spell ends, the creature knows it was charmed by you.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st. The creatures must be within 30 feet of each other when you target them.</p>
          `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1042,
      german: 'Rascher Rückzug',
      english: 'Expeditious Retreat',
      image: 'expeditiousRetreat',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber erlaubt es dir, dich mit einer unglaublicher Geschwindigkeit zu bewegen. Wenn du diesen Zauber wirkst und danach als Bonusaktion in jedem deiner Züge, bis der Zauber endet, kannst du die Spurt-Aktion verwenden.</p>
        `,
      description_en: `
          <p>This spell allows you to move at an incredible pace. When you cast this spell, and then as a bonus action on each of your turns until the spell ends, you can take the Dash action.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1043,
      german: 'Rüstung des Agathys',
      english: 'Armor of Agathys',
      image: 'armorOfAgathys',
      level: 1,
      isCustom: false,
      classes: ['test', 'warlock', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      heal: { diceNumber: 5, diceType: 0, additionalHeal: 0 },
      description_de: `
          <p>Schützende magische Energie umgibt dich, die sich als geisterhafter Frost manifestiert, der dich und deine Ausrüstung bedeckt. Du erhältst für die Wirkungsdauer 5 temporäre Trefferpunkte. Wenn dich eine Kreatur mit einem Nahkampfangriff trifft, während du über diese Trefferpunkte verfügst, erleidet die Kreatur 5 Kälteschaden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich die Anzahl der temporären Trefferpunkte um 5 für jeden Zauberplatz-Grad über dem ersten.</p>
        `,
      description_en: `
          <p>A protective magical force surrounds you, manifesting as a spectral frost that covers you and your gear. You gain 5 temporary hit points for the duration. If a creature hits you with a melee attack while you have these hit points, the creature takes 5 cold damage.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the amount of temporary hit points increases by 5 for each slot level above 1st.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1044,
      german: 'Schild',
      english: 'Shield',
      image: 'shield',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'reaction',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine unsichtbare Barriere aus magischer Kraft erscheint und schützt dich. Bis zum Beginn deines nächsten Zuges hast du einen Bonus von +5 auf deine RK, auch gegen den auslösenden Angriff, und du erleidest keinen Schaden durch magische Geschosse.</p> 
        `,
      description_en: `
          <p>An invisible barrier of magical force appears and protects you. Until the start of your next turn, you have a +5 bonus to AC, including against the triggering attack, and you take no damage from magic missile.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1045,
      german: 'Schild des Glaubens',
      english: 'Shield of Faith',
      image: 'shieldOfFaith',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein Schimmerndes Feld erscheint und umhüllt eine Kreatur deiner Wahl in Reichweite. Es gewährt ihr für die Wirkungsdauer einen Bonus von +2 auf RK.</p>
        `,
      description_en: `
          <p>A shimmering field appears and surrounds a creature of your choice within range, granting it a +2 bonus to AC for the duration.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1046,
      german: 'Schlaf',
      english: 'Sleep',
      image: 'sleep',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber versetzt Kreaturen in einen magischen Schlaf. Wirf 5W8, das Ergebnis ist die Gesamtzahl an Trefferpunkten an Kreaturen, die dieser Zauber beeinflussen kann. Kreaturen innerhalb von 6 Metern um einen Punkt deiner Wahl in Reichweite werden betroffen, in aufsteigender Reihenfolge nach ihren aktuellen Trefferpunkten (bewusstlose Kreaturen werden ignoriert).</p>
          <p>Beginnend mit der Kreatur mit den niedrigsten aktuellen Trefferpunkten werden alle Kreaturen, die von diesem Zauber betroffen werden, bewusstlos bis der Zauber endet, der Schlafende Schaden erleidet oder jemand eine Aktion verwendet, um den Schlafenden zu schütteln oder zu ohrfeigen. Ziehe die Trefferpunkte der Kreatur von der Gesamtsumme ab, ehe du mit der nächsten Kreatur mit den niedrigsten Trefferpunkten weitermachst. Die Trefferpunkte einer Kreatur müssen gleich oder niedriger als die verbleibenden Gesamttrefferpunkte sein, damit sie betroffen wird. Untote und Kreaturen, die nicht bezaubert werden können, sind von diesem Zauber nicht betroffen.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich die Gesamtzahl der Trefferpunkte für jeden Zauberplatz-Grad über dem ersten um 2W8.</p>
          `,
      description_en: `
          <p>This spell sends creatures into a magical slumber. Roll 5d8; the total is how many hit points of creatures this spell can affect. Creatures within 20 feet of a point you choose within range are affected in ascending order of their current hit points (ignoring unconscious creatures).</p>
          <p>Starting with the creature that has the lowest current hit points, each creature affected by this spell falls unconscious until the spell ends, the sleeper takes damage, or someone uses an action to shake or slap the sleeper awake. Subtract each creature’s hit points from the total before moving on to the creature with the next lowest hit points. A creature’s hit points must be equal to or less than the remaining total for that creature to be affected. Undead and creatures immune to being charmed aren’t affected by this spell.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, roll an additional 2d8 for each slot level above 1st.</p>
          `,
      school: 'enchantment',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 1047,
      german: 'Schmieren',
      english: 'Grease',
      image: 'grease',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Schmieriges Fett bedeckt den Boden in einem Quadrat mit 3 Metern Seitenlänge, das um einen Punkt in Reichweite zentriert ist, und verwandelt diesen Bereich für die Wirkungsdauer in schwieriges Gelände.</p>
          <p>Wenn die Schmiere erscheint, müssen alle Kreaturen, die in dem Bereich stehen, einen Geschicklichkeitsrettungswurfschaffen, um nicht zu Boden zu fallen und liegend zu sein. Eine Kreatur, die den Bereich betritt oder ihren Zug dort beendet, muss ebenfalls einen Geschicklichkeitsrettungswurfschaffen oder zu Boden gehen und ist liegend.</p>
        `,
      description_en: `
          <p>Slick grease covers the ground in a 10-foot square centered on a point within range and turns it into difficult terrain for the duration.</p>
          <p>When the grease appears, each creature standing in its area must succeed on a Dexterity saving throw or fall prone. A creature that enters the area or ends its turn there must also succeed on a Dexterity saving throw or fall prone.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'square',
      length: 10,
    },
    {
      id: 1048,
      german: 'Schutz vor Gut und Böse',
      english: 'Protection from Evil and Good',
      image: 'protectionFromEvilAndGood',
      level: 1,
      isCustom: false,
      classes: ['test', 'warlock', 'cleric', 'wizard', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Bis der Zauber endet, wird eine bereitwillige Kreatur, die du berührst, vor bestimmten Arten von Kreaturen geschützt: Aberrationen, himmlische Wesen, Elementare, Feen, Unholde und Untote. </p>
          <p>Der Schutz bietet mehrere Vorteile: Kreaturen dieser Art haben einen Nachteil auf Angriffswürfe gegen das Ziel. Das Ziel kann außerdem nicht von ihnen bezaubert, verängstigt oder durch sie besessen werden. Wenn das Ziel bereits verzaubert, verängstigt oder von einer solchen Kreatur besessen ist, hat das Ziel Vorteil bei jedem neuen Rettungswurf gegen den relevanten Effekt.</p>
        `,
      description_en: `
          <p>Until the spell ends, one willing creature you touch is protected against certain types of creatures: aberrations, celestials, elementals, fey, fiends, and undead.</p>
          <p>The protection grants several benefits. Creatures of those types have disadvantage on attack rolls against the target. The target also can't be charmed, frightened, or possessed by them. If the target is already charmed, frightened, or possessed by such a creature, the target has advantage on any new saving throw against the relevant effect.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1049,
      german: 'Segnen',
      english: 'Bless',
      image: 'bless',
      level: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 0,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du segnest bis zu drei Kreaturen deiner Wahl in Reichweite. Immer wenn ein Ziel einen Angriffs- oder Rettungswurf macht, bis der Zauber endet, kann es einen W4 werfen und das Ergebnis auf den Angriffswurf oder Rettungswurf addieren.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann kannst du den Zauber für jeden Zauberplatz-Grad über dem ersten auf eine zusätzliche Kreatur wirken.</p>
      `,
      description_en: `
        <p>You bless up to three creatures of your choice within range. Whenever a target makes an attack roll or a saving throw before the spell ends, the target can roll a d4 and add the number rolled to the attack roll or saving throw.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st.</p>
      `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1050,
      german: 'Selbstverkleidung',
      english: 'Disguise Self',
      image: 'disguiseSelf',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du lässt dich selbst - einschließlich deiner Kleidung, Rüstung, Waffen und anderer Gegenstände an deiner Person - anders aussehen, bis der Zauber endet oder bis du deine Aktion nutzt, um ihn abzubrechen. Du kannst 30 Zentimeter kleiner oder größer erscheinen und du kannst dünn, fett oder dazwischen erscheinen. Du kannst deinen Körpertyp nicht ändern, also musst du eine Form annehmen, die die gleiche Grundanordnung der Gliedmaßen hat. Ansonsten ist das Ausmaß der Illusion dir überlassen.</p>
          <p>Die Veränderungen, die dieser Zauber bewirkt, halten einer physischen Überprüfung nicht stand. Wenn du diesen Zauber zum Beispiel verwendest, um deiner Kleidung einen Hut hinzuzufügen, gehen Objekte durch den Hut hindurch, und jeder, der ihn berührt, würde nichts spüren oder deinen Kopf und deine Haare fühlen. Wenn du diesen Zauber verwendest, um dünner zu erscheinen als du bist, würde die Hand von jemandem, der die Hand ausstreckt, um dich zu berühren, gegen dich stoßen, während sie sich scheinbar noch in der Luft befindet.</p>
          <p>Um zu erkennen, dass du verkleidet bist, kann eine Kreatur ihre Aktion nutzen, um deine Erscheinung zu untersuchen und muss einen Intelligenztest (Nachforschung) gegen deinen Zauberrettungswurf SG bestehen.</p>
        `,
      description_en: `
          <p>You make yourself - including your clothing, armor, weapons, and other belongings on your person - look different until the spell ends or until you use your action to dismiss it. You can seem 1 foot shorter or taller and can appear thin, fat, or in between. You can't change your body type, so you must adopt a form that has the same basic arrangement of limbs. Otherwise, the extent of the illusion is up to you.</p>
          <p>The changes wrought by this spell fail to hold up to physical inspection. For example, if you use this spell to add a hat to your outfit, objects pass through the hat, and anyone who touches it would feel nothing or would feel your head and hair. If you use this spell to appear thinner than you are, the hand of someone who reaches out to touch you would bump into you while it was seemingly still in midair.</p>
          <p>To discern that you are disguised, a creature can use its action to inspect your appearance and must succeed on an Intelligence (Investigation) check against your spell save DC.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1051,
      german: 'Sengendes Niederstrecken',
      english: 'Searing Smite',
      image: 'searingSmite',
      level: 1,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Das nächste Mal, wenn du innerhalb der Wirkungsdauer des Zaubers eine Kreatur mit einer Nahkampfwaffe triffst, blitzt deine Waffe mit weißglühender Intensität auf.</p>
          <p>Der Angriff fügt dem Ziel zusätzlich 1W6 Feuerschaden zu und lässt es in Flammen aufgehen.</p>
          <p>Zu Beginn eines jeden seiner Züge, bis der Zauber endet, muss das Ziel einen Konstitutionsrettungswurf ablegen. Misslingt er, erleidet es 1W6 Feuerschaden. Bei einem erfolgreichen Rettungswurf endet der Zauber.</p>
          <p>Wenn das Ziel oder eine Kreatur im Umkreis von 1,5m von ihm eine Aktion verwendet, um die Flammen zu ersticken, oder ein anderer Effekt das Feuer löscht (beispielweise wenn die betroffene Kreatur unter Wasser getaucht wird oder es regnet), endet der Zauber.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Spruch mit einem Zauberplatz des 2. oder eines höheren Grades wirkst, steigt der anfängliche Bonusschaden beim Angriff für jeden Grad über den 1. hinaus um 1W6.</p>
        `,
      description_en: `
          <p>The next time you hit a creature with a melee weapon attack during the spell’s duration, your weapon flares with white-hot intensity, and the attack deals an extra 1d6 fire damage to the target and causes the target to ignite in flames.</p>
          <p>At the start of each of its turns until the spell ends, the target must make a Constitution saving throw. On a failed save, it takes 1d6 fire damage. On a successful save, the spells ends. If the target or a creature within 5 feet of it uses an action to put out the flames, or if some other effect douses the flames (such as the target being submerged in water), the spell ends.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the initial extra damage dealt by the attack increases by 1d6 for each slot.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1052,
      german: 'Sprachen verstehen',
      english: 'Comprehend Languages',
      image: 'comprehendLanguages',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Für die Wirkungsdauer verstehst du die wörtliche Bedeutung jeglicher gesprochene Sprache, die du hörst. Du kannst auch jede geschriebene Sprache lesen, die du sehen kannst, allerdings musst du die Oberfläche berühren, auf der die Worte geschrieben sind. Es dauert ungefähr eine Minute, eine Seite Text zu lesen.</p>
          <p>Der Zauber hilft nicht bei der Entschlüsselung von Geheimbotschaften in einem Text oder Glyphen, wie arkane Zeichen, die nicht Teil einer geschriebenen Sprache sind.</p>
        `,
      description_en: `
          <p>For the duration, you understand the literal meaning of any spoken language that you hear. You also understand any written language that you see, but you must be touching the surface on which the words are written. It takes about 1 minute to read one page of text.</p>
          <p>This spell doesn’t decode secret messages in a text or a glyph, such as an arcane sigil, that isn’t part of a written language.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1053,
      german: 'Springen',
      english: 'Jump',
      image: 'jump',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur. Die Sprungdistanz der Kreatur wird verdreifacht, bis der Zauber endet.</p>
        `,
      description_en: `
          <p>You touch a creature. The creature's jump distance is tripled until the spell ends.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1054,
      german: 'Sprühende Farben',
      english: 'Color Spray',
      image: 'colorSpray',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein schillerndes Feld aus blinkendem, farbigem Licht entspringt deiner Hand. Wirf 6W10; die Summe gibt an, wie viele Trefferpunkte von Kreaturen dieser Zauber bewirken kann. Kreaturen in einem 4,5 Meter Kegel, der von dir ausgeht, werden in aufsteigender Reihenfolge ihrer aktuellen Trefferpunkte betroffen (bewusstlose Kreaturen und Kreaturen, die nicht sehen können, werden ignoriert).</p>
          <p>Beginnend mit der Kreatur, die die niedrigsten aktuellen Trefferpunkte hat, wird jede von diesem Zauber betroffene Kreatur geblendet, bis der Zauber endet. Ziehe die Trefferpunkte jeder Kreatur von der Gesamtzahl ab, bevor du mit der Kreatur mit den nächstniedrigen Trefferpunkten weitermachst. Die Trefferpunkte einer Kreatur müssen gleich oder kleiner als die verbleibende Gesamtzahl sein, damit diese Kreatur betroffen ist.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich die Gesamtzahl der Trefferpunkte für jeden Zauberplatz-Grad über dem ersten um 2W10.</p>
        `,
      description_en: `
          <p>A dazzling array of flashing, colored light springs from your hand. Roll 6d10; the total is how many hit points of creatures this spell can affect. Creatures in a 15-foot cone originating from you are affected in ascending order of their current hit points (ignoring unconscious creatures and creatures that can't see).</p>
          <p>Starting with the creature that has the lowest current hit points, each creature affected by this spell is blinded until the spell ends. Subtract each creature’s hit points from the total before moving on to the creature with the next lowest hit points. A creature’s hit points must be equal to or less than the remaining total for that creature to be affected.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, roll an additional 2d10 for each slot level above 1st.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cone',
      length: 15,
    },
    {
      id: 1055,
      german: 'Strahl der Übelkeit',
      english: 'Ray of Sickness',
      image: 'rayOfSickness',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 8, damageType: 'poison' }],
      doesHeal: false,
      description_de: `
          <p>Ein Strahl aus kränklich grüner Energie Schlägt auf eine Kreatur in Reichweite über Lege einen Fernkampf-Zauberangriff gegen das Ziel ab. Das Ziel erleidet bei einem Treffer 2W6 Giftschaden und muss einen Rettungswurf auf Konstitution machen. Bei einem misslungenen Rettungswurf ist das Ziel außerdem bis zum Ende seines nächsten Zuges vergiftet.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Zauberplatz-Grad über dem ersten.</p>
        `,
      description_en: `
          <p>A ray of sickening greenish energy lashes out toward a creature within range. Make a ranged spell attack against the target. On a hit, the target takes 2d8 poison damage and must make a Constitution saving throw. On a failed save, it is also poisoned until the end of your next turn.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d8 for each slot level above 1st.</p>
        `,
      school: 'necromancy',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1056,
      german: 'Tashas fürchterliches Lachen',
      english: "Tasha's Hideous Laughter",
      image: 'tashasHideousLaughter',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine Kreatur deiner Wahl, in Reichweite, die du sehen kannst, nimmt alles als unglaublich lustig wahr und verfällt in einen Lachanfall, wenn der Zauber in Kraft tritt. Misslingt dem Ziel ein Weisheitsrettungswurf, erhält es den Zustand Liegend, wird Kampfunfähig und kann für die Wirkungsdauer nicht aufstehen. Kreaturen mit Intelligenz von 4 oder niedriger sind von diesem Zauber nicht betroffen.</p>
          <p>Wenn das Ziel Schaden erleidet, sowei am Ende eines jeden seiner Züge, darf es einen weiteren Weisheitsrettungsweurf ablegen. Es ist bei dem Rettungswurf im Vorteil, wenn dieser durch Schaden ausgelöst wurde. Bei einem Erfolg endet der Zauber.</p>
        `,
      description_en: `
          <p>A creature of your choice that you can see within range perceives everything as hilariously funny and falls into fits of laughter if this spell affects it. It must succeed on a Wisdom saving throw or fall prone, becoming incapacitated and unable to stand up for the duration. A creature with an Intelligence score of 4 or less isn't affected.</p>
          <p>At the end of each of its turns and each time it takes damage, the target can make another Wisdom saving throw. The target has advantage on the saving throw if it's triggered by damage. On a success, the spell ends.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1057,
      german: 'Tashas ätzenden Gebräu',
      english: "Tasha's Caustic Brew",
      image: 'tashasCausticBrew',
      level: 1,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 4, damageType: 'acid' }],
      doesHeal: false,
      description_de: `
          <p>Von dir aus spritzt ein neun Meter langer und 1,5 Meter breiter Säurestrahl in eine Richtung deiner Wahl. Alle Kreaturen in der Ziellinie müssen einen erfolgreichen Rettungswurf auf Geschicklichkeit ausführen, sonst sind sie für die Wirkungsdauer des Zaubers mit Säure bedeckt. Kreaturen können die Säure als Aktion von sich oder anderen abschrubben oder abwaschen. Eine mit Säure bedeckte Kreatur nimmt zu Beginn jedes ihrer Züge 2W4 Säureschaden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, erhöht sich der Schaden um 1W4 für jeden Zauberplatz-Grad über dem ersten.</p>
        `,
      description_en: `
          <p>You hurl a 5-foot-diameter ball of acid at a creature or object within range. Any creature in the target's space must succeed on a Dexterity saving throw or take 2d4 acid damage. A creature can take this damage only once per turn.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d4 for each slot level above 1st.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 30,
    },
    {
      id: 1058,
      german: 'Tensers schwebende Scheibe',
      english: "Tenser's Floating Disk",
      image: 'tensersFloatingDisk',
      level: 1,
      isCustom: false,
      classes: ['test', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Der Zauber erschafft eine kreisförmige, waagerechte Fläche aus Energie mit 90 cm Durchmesser und 2,5 cm Dicke. Sie schwebt in einem nicht besetzten Bereich deiner Wahl, den du sehen kannst, in 90 cm Höhe über dem Boden. Die Scheibe bleibt für die Wirkungsdauer bestehen und kann bis zu 500 Pfund tragen. Wenn mehr Gewicht auf ihr lastet, endet der Zauber, und alles, das sich auf ihr befand, fällt zu Boden.</p>
          <p>Die Scheibe ist unbeweglich, solange du dich innerhalb von 6m um sie befindest. Entfernst du dich mehr als 6 m, folgt dir die Scheibe, sodass sie innerhalb von 6 m zu dir bleibt. Sie ist in der Lage, sich über unebenes Gelände, Treppen, Abhänge und Ähnliches zu bewegen, kann aber keine Höhenunterschiede von 3 m oder mehr überwinden. Zum Beispiel kann sich die Scheibe nicht über eine 3 m tiefe Grube bewegen oder eine solche verlassen, wenn sie in ihr erschaffen wurde.</p>
          <p>Entfernst du dich weiter als 30 m von der Scheibe (für gewöhnlich weil sie sich nicht um ein Hindernis bewegen kann, um dir zu folgen), endet der Zauber.</p>
        `,
      description_en: `
          <p>This spell creates a circular, horizontal plane of force, 3 feet in diameter and 1 inch thick, that floats 3 feet above the ground in an unoccupied space of your choice that you can see within range. The disk remains for the duration and can hold up to 500 pounds. If more weight is placed on it, the spell ends, and everything on the disk falls to the ground.</p>
          <p>The disk is immobile while you are within 20 feet of it. If you move more than 20 feet away from it, the disk follows you so that it remains within 20 feet of you. It can move across uneven terrain, up or down stairs, slopes and the like, but it can't cross an elevation change of 10 feet or more. For example, the disk can't move across a 10-foot-deep pit, nor could it leave such a pit if it was created at the bottom.</p>
          <p>If you move more than 100 feet from the disk (typically because it can't move around an obstacle to follow you), the spell ends.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1059,
      german: 'Tierbindung',
      english: 'Beast Bond',
      image: 'beastBond',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du stellst eine telepathische Verbindung mit einem Tier her, das du berührst und das dir freundlich gesinnt ist oder von dir verzaubert wurde. Der Zauber schlägt fehl, wenn die Intelligenz des Tieres 4 oder höher ist. Bis zum Ende des Zaubers ist die Verbindung aktiv, solange du und das Tier sich in Sichtweite zueinander befinden. Durch die Verbindung kann das Biest deine telepathischen Botschaften an es verstehen, und es kann dir einfache Emotionen und Konzepte telepathisch zurückmelden. Während die Verbindung aktiv ist, erhält die Bestie einen Vorteil bei Angriffswürfen gegen jede Kreatur im Umkreis von 1,5 Metern von dir, die du sehen kannst.</p>
        `,
      description_en: `
          <p>This spell lets you convince a beast that you mean it no harm. Choose a beast that you can see within range. It must see and hear you. If the beast's Intelligence is 4 or higher, the spell fails. Otherwise, the beast must succeed on a Wisdom saving throw or be charmed by you for the spell's duration. If you or one of your companions harms the target, the spell ends.</p>
        `,
      school: 'enchantment',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1060,
      german: 'Tierfreundschaft',
      english: 'Animal Friendship',
      image: 'animalFriendship',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger'],
      timeToCast: 0,
      cost: 'action',
      duration: 1,
      durationType: 'days',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber erlaubt es dir, ein Tier davon zu überzeugen, dass du ihm nicht schaden willst. Wähle ein Tier in Reichweite, das du sehen kannst. Es muss dich sehen und hören können. Wenn die Intelligenz des Tieres 4 oder höher beträgt, dann schlägt der Zauber fehl. Ansonsten muss das Tier einen Rettungswurf auf Weisheits machen, um nicht für die Wirkungsdauer bezaubert zu werden. Wenn du oder einer deiner Gefährten das Ziel verletzt, endet der Zauber.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann kannst du ein zusätzliches Tier für jeden Zauberplatzgrad über dem ersten bezaubern. Die Tiere müssen sich innerhalb von 9 Metern voneinander befinden, wenn du den Zauber wirkst.</p>
        `,
      description_en: `
          <p>This spell lets you convince a beast that you mean it no harm. Choose a beast that you can see within range. It must see and hear you. If the beast's Intelligence is 4 or higher, the spell fails. Otherwise, the beast must succeed on a Wisdom saving throw or be charmed by you for the spell's duration. If you or one of your companions harms the target, the spell ends.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can affect one additional beast for each slot level above 1st. The animals must be within 30 feet of each other when you target them.</p>
        `,
      school: 'enchantment',
      isRanged: false,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1061,
      german: 'Unsichtbarer Diener',
      english: 'Unseen Servant',
      image: 'unseenServant',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein unsichtbarer, geisterhafter Diener erscheint in einem unbesetzten Raum, den du sehen kannst, innerhalb der Reichweite. Der Diener erscheint mit einer Geschwindigkeit von 9 Metern und kann einfache Befehle ausführen, die du ihm gibst. Der Diener kann einfache Aufgaben erledigen, die keine besondere Fertigkeit erfordern. Der Diener kann nicht angreifen, Zaubersprüche wirken oder Gegenstände verwenden, die magische Energie erfordern. Wenn du ihm keine Befehle gibst, führt der Diener die letzte ihm erteilte Anweisung aus, bis der Zauber endet.</p>
          <p>Einmal in jedem deiner Züge kannst du als Bonusaktion dem Diener den geistigen Befehl geben, sich bis zu 4,50 Meter zu bewegen und mit einem Gegenstand zu interagieren. Der Diener kann einfache Aufgaben übernehmen, die ein menschlicher Diener leisten könnte. Er kann Dinge holen, reinigen und reparieren, Kleider zusammenlegen, Feuer anzünden, Essen servieren, und Wein einschenken. Sobald du den Befehl erteilst, erfüllt der Diener sie nach besten Möglichkeiten bis er die Aufgabe erfüllt. Dann wartet er auf den nächsten Befehl.</p>
          <p>Wenn du dem Diener befiehlst, eine Aufgabe zu erfüllen, für die er sich mehr als 18 Meter von dir entfernen müsste, dann endet der Zauber.</p>
          `,
      description_en: `
          <p>This spell creates an invisible, mindless, shapeless force that performs simple tasks at your command until the spell ends. The servant springs into existence in an unoccupied space on the ground within range. It has AC 10, 1 hit point, and a Strength of 2, and it can’t attack. If it drops to 0 hit points, the spell ends.</p>
          <p>Once on each of your turns as a bonus action, you can mentally command the servant to move up to 15 feet and interact with an object. The servant can perform simple tasks that a human servant could do, such as fetching things, cleaning, mending, folding clothes, lighting fires, serving food, and pouring wine. Once you give the command, the servant performs the task to the best of its ability until it completes the task, then waits for your next command.</p>
          <p>If you command the servant to perform a task that would move it more than 60 feet away from you, the spell ends.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 1062,
      german: 'Verderben',
      english: 'Bane',
      image: 'bane',
      level: 1,
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'charisma',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Bis zu drei Kreaturen deiner Wahl in Reichweite, die du sehen kannst, müssen Rettungswürfe auf Charisma machen. Immer wenn ein Ziel, das seinen Rettungswurf nicht geschafft hat, einen Angriffs- oder Rettungswurf macht bis der Zauber endet, muss es einen W4 werfen und das Ergebnis vom Angriffs- oder Rettungswurf abziehen.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, dann kannst du den Zauber für jeden Zauberplatzgrad über dem ersten auf eine zusätzliche Kreatur wirken.</p>
        `,
      description_en: `
          <p>Up to three creatures of your choice that you can see within range must make Charisma saving throws. Whenever a target that fails this saving throw makes an attack roll or a saving throw before the spell ends, the target must roll a d4 and subtract the number rolled from the attack roll or saving throw.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you can target one additional creature for each slot level above 1st.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 1063,
      german: 'Verstricken',
      english: 'Entangle',
      image: 'entangle',
      level: 1,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Greifende Unkräuter und Ranken sprießen aus dem Boden in einem 20-Fuß-Quadrat, beginnend von einem Punkt in Reichweite. Für die Dauer des Zaubers verwandeln diese Pflanzen den Boden in dem Gebiet in schwieriges Terrain.</p>
          <p>Eine Kreatur, die sich in dem Gebiet befindet, wenn du den Zauber sprichst, muss einen Rettungswurf auf Stärke bestehen oder wird von den umschlingenden Pflanzen gefangen gehalten, bis der Zauber endet. Eine Kreatur, die von den Pflanzen gefesselt ist, kann ihre Aktion nutzen, um eine Stärkeprobe gegen deinen Zauberrettungswurf SG zu machen. Bei einem Erfolg befreit sie sich selbst. Wenn der Zauber endet, welken die beschworenen Pflanzen weg.</p>
        `,
      description_en: `
          <p>Grasping weeds and vines sprout from the ground in a 20-foot square starting from a point within range. For the duration, these plants turn the ground in the area into difficult terrain.</p>
          <p>A creature in the area when you cast the spell must succeed on a Strength saving throw or be restrained by the entangling plants until the spell ends. A creature restrained by the plants can use its action to make a Strength check against your spell save DC. On a success, it frees itself.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'square',
      length: 20,
    },
    {
      id: 1064,
      german: 'Vertrauten finden',
      english: 'Find Familiar',
      image: 'findFamiliar',
      level: 1,
      isCustom: false,
      classes: ['test', 'wizard'],
      timeToCast: 1,
      cost: 'hours',
      duration: 0,
      durationType: 'days',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erhältst die Dienste eines Vertrauten, eines Geistes, der die Gestalt eines Tieres deiner Wahl annimmt: Eidechse, Eule, Falke, Fisch (Quipper), Fledermaus, Frosch (Kröte), Giftschlange, Katze, Krabbe, Oktopus, Rabe, Ratte, Seepferdchen, Spinne oder Wiesel. Der Vertraute erscheint in einem nicht besetzten Bereich in Reichweite und hat die Spielwerte der ausgewählten Gestalt, doch ist er ein himmlisches Wesen, ein Feenwesen oder ein Unhold (deine Wahl) und kein Tier.</p>
          <p>Dein Vertrauter agiert unabhängig von dir, folgt aber immer deinen Befehlen. Im Kampf hat er seine eigene Initiative und handelt selbstständig. Ein Vertrauter kann nicht angreifen, aber alle anderen Aktionen ausführen.</p>
          <p>Wenn der Vertraute auf 0 TP fällt, verschwindet er und lässt keine physische Form zurück. Er taucht wieder auf, wenn du diesen Zauber erneut wirkst.</p>
          <p>Solange sich dein Vertrauter im Umkreis von 30 Metern um dich aufhält, kannst du telepathisch mit ihm kommunizieren. Außerdem kannst du als Aktion bis zum Beginn deines nächsten Zuges durch die Augen deines Vertrauten sehen und hören, was er hört, wobei du den Vorteil aller besonderen Sinne des Vertrauten nutzen kannst. Während dieser Zeit bist du taub und blind was deine eigenen Sinne angeht.</p>
          <p>Als Aktion kannst du deinen Vertrauten kurzzeitig fortschicken. Er verschwindet in einer Taschendimension, wo er auf deinen Ruf wartet. Alternativ kannst du ihn für immer fortschicken. Solange der Vertraute fortgeschickt ist, kannst du deine Aktion verwenden, um ihn in einen nicht besetzten Bereich innerhalb von 9 Metern um dich auftauchen zu lassen.</p>
          <p>Du kannst nur einen Vertrauten auf einmal haben. Wenn du diesen Zauber wirkst, während du bereits über einen Vertrauten verfügst, lässt du ihn stattdessen eine neue Gestalt annehmen. Wähle eine der Formen aus der oben genannten Liste. Dein Vertrauter verwandelt sich in die gewählte Kreatur.</p>
          <p>Wenn du außerdem einen Zauber mit der Reichweite Berührung wirkst, kann dein Vertrauter den Zauber überbringen, als hättest du ihn gewirkt. Dein Vertrauter muss sich innerhalb von 30 Metern um dich befinden und seine Reaktion verwenden, um den Zauber zu überbrigen, wenn du ihn wirkst. Wenn der Zauber einen Angriffswurf erfordert, verwende deinen Angriffsmodifikator für den Wurf.</p>
        `,
      description_en: `
          <p>You gain the service of a familiar, a spirit that takes an animal form you choose: bat, cat, crab, frog (toad), hawk, lizard, octopus, owl, poisonous snake, fish (quipper), rat, raven, sea horse, spider, or weasel. Appearing in an unoccupied space within range, the familiar has the statistics of the chosen form, though it is a celestial, fey, or fiend (your choice) instead of a beast.</p>
          <p>Your familiar acts independently of you, but it always obeys your commands. In combat, it rolls its own initiative and acts on its own turn. A familiar can't attack, but it can take other actions as normal.</p>
          <p>When the familiar drops to 0 hit points, it disappears, leaving behind no physical form. It reappears when you cast this spell again.</p>
          <p>While your familiar is within 100 feet of you, you can communicate with it telepathically. Additionally, as an action, you can see through your familiar's eyes and hear what it hears until the start of your next turn, gaining the benefits of any special senses that the familiar has. During this time, you are deaf and blind with regard to your own senses.</p>
          <p>As an action, you can temporarily dismiss your familiar. It disappears into a pocket dimension where it awaits your summons. Alternatively, you can dismiss it forever. As an action while it is temporarily dismissed, you can cause it to reappear in any unoccupied space within 30 feet of you.</p>
          <p>You can't have more than one familiar at a time. If you cast this spell while you already have a familiar, you instead cause it to adopt a new form. Choose one of the forms from the above list. Your familiar transforms into the chosen creature.</p>
          <p>Finally, when you cast a spell with a range of touch, your familiar can deliver the spell as if it had cast the spell. Your familiar must be within 100 feet of you, and it must use its reaction to deliver the spell when you cast it. If the spell requires an attack roll, you use your attack modifier for the roll.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 10,
      hasAreaOfEffect: false,
    },
    {
      id: 1065,
      german: 'Verwünschen',
      english: 'Hex',
      image: 'hex',
      level: 1,
      isCustom: false,
      classes: ['test', 'warlock'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du verwünschst eine Kreatur in Reichweite, die du sehen kannst. Bis der Zauber endet, fügst du dem Ziel zusätzlich 1W6 nekrotischen Schaden zu, wenn du es mit einem Angriff triffst. Wähle außerdem ein Attribut, wenn du den Zauber wirkst. Das Ziel hat Nachteil bei Attributswürfen mit dem gewählten Attribut.</p>
          <p>Wenn das Ziel auf 0 Trefferpunkte fällt, ehe der Zauber endet, kannst du in einem späteren Zug eine Bonusaktion verwenden, um eine neue Kreatur zu verwünschen. Der Zauber "Fluch brechen" hebt diesen Zauber vorzeitig auf.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. oder 4. Grades wirkst, kannst du deine Konzentration bis zu 8 Stunden aufrechterhalten. Wenn du einen Zauberplatz des 5. Grades oder höher verwendest, beträgt die Konzentration auf den Zauber 24 Stunden.</p>
        `,
      description_en: `
          <p>You place a curse on a creature that you can see within range. Until the spell ends, you deal an extra 1d6 necrotic damage to the target whenever you hit it with an attack. Also, choose one ability when you cast the spell. The target has disadvantage on ability checks made with the chosen ability.</p>
          <p>If the target drops to 0 hit points before this spell ends, you can use a bonus action on a subsequent turn of yours to curse a new creature. A remove curse cast on the target ends this spell early.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd or 4th level, you can maintain your concentration on the spell for up to 8 hours. When you use a spell slot of 5th level or higher, you can maintain your concentration on the spell for up to 24 hours.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 1066,
      german: 'Wasser erschaffen und zerstören',
      english: 'Create or Destroy Water',
      image: 'createOrDestroyWater',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p><b>Wasser erschaffen:</b> Du erschaffst bis zu 45 Liter sauberes Wasser in Reichweite in einem offenen Behälter. Alternativ fällt das Wasser als Regen in einem Würfel mit 9 Metern Kantenlänge in Reichweite und löscht offene Flammen in der Umgebung.</p>
          <p><b>Wasser zerstören:</b> Du zerstörst bis zu 45 Liter Wasser in einem offenen Behälter in Reichweite. Alternativ zerstörst du Nebel in einem Würfel mit 9 Metern Kantenlänge in Reichweite.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grads oder höher wirkst, erschaffst oder zerstörst du 45 zusätzliche Liter Wasser oder die Größe des Würfels vergrößert sich um 1,5 Meter, für jeden zusätzlichen Zaubergrad.</p>
          `,
      description_en: `
          <p><b>Create Water:</b> You create up to 10 gallons of clean water within range in an open container. Alternatively, the water falls as rain in a 30-foot cube within range, extinguishing exposed flames in the area.</p>
          <p><b>Destroy Water:</b> You destroy up to 10 gallons of water in an open container within range. Alternatively, you destroy fog in a 30-foot cube within range.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, you create or destroy 10 additional gallons of water, or the size of the cube increases by 5 feet, for each slot level above 1st.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 30,
    },
    {
      id: 1067,
      german: 'Wunden Heilen',
      english: 'Cure Wounds',
      image: 'cureWounds',
      level: 1,
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'paladin', 'ranger'],
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 1, diceType: 8, additionalHeal: 4 },
      description_de: `
        <p>Eine Kreatur, die du berührst, gewinnt Trefferpunkte in Höhe von 1W8 + deinem Attributsmodifikator im Zauberwirken zurück. Dieser Zauber wirkt nicht auf Untote oder Konstrukte.</p>
        <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, steigt die Heilung für jeden Zauberplatzgrad über dem 1. um 1W8.</p>
      `,
      description_en: `
        <p>A creature you touch regains a number of hit points equal to 1d8 + your spellcasting ability modifier. This spell has no effect on undead or constructs.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the healing increases by 1d8 for each slot level above 1st.</p>
      `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1068,
      german: 'Wunden verursachen',
      english: 'Inflict Wounds',
      image: 'inflictWounds',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 10, damageType: 'necrotic' }],
      doesHeal: false,
      description_de: `
        <p>Mache einen Nahkampfangriff gegen eine Kreatur innerhalb deiner Reichweite. Bei einem Treffer erleidet das Ziel 3W10 nekrotischen Schaden.</p>
        <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 2. Grades oder höher wirkst, steigt der Schaden für jeden Zauberplatzgrad über dem 1. um 1W10.</p>
      `,
      description_en: `
        <p>Make a melee spell attack against a creature you can reach. On a hit, the target takes 3d10 necrotic damage.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 2nd level or higher, the damage increases by 1d10 for each slot level above 1st.</p>
      `,
      school: 'necromancy',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 1069,
      german: 'Zeichen des Jägers',
      english: "Hunter's Mark",
      image: 'huntersMark',
      level: 1,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du wählst eine Kreatur in Reichweite aus, die du sehen kannst, und markierst sie auf mystische Art als deine Beute. Bis der Zauber endet, fügst du dem Ziel jedes Mal 1W6 Schaden zu, wenn du es mit einem Waffenangriff triffst, und du hast Vorteil bei Würfen auf Wahrnehmung oder Überleben, wenn du das Ziel finden willst. Wenn das Ziel auf 0 Trefferpunkte fällt, ehe der Zauber endet, kannst du in einem späteren Zug eine Bonusaktion verwenden, um eine neue Kreatur zu markieren.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. oder 4. Grades wirkst, kannst du deine Konzentration für bis zu 8 Stunden aufrechterhalten. Wenn du diesen Zauber mit einem Zauberplatz des 5. Grades oder höher wirkst, kannst du deine Konzentration für bis zu 24 Stunden aufrechterhalten.</p>
        `,
      description_en: `
          <p>You choose a creature you can see within range and mystically mark it as your quarry. Until the spell ends, you deal an extra 1d6 damage to the target whenever you hit it with a weapon attack, and you have advantage on any Wisdom (Perception) or Wisdom (Survival) check you make to find it.</p>
          <p>If the target drops to 0 hit points before this spell ends, you can use a bonus action on a subsequent turn of yours to mark a new creature.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd or 4th level, you can maintain your concentration on the spell for up to 8 hours. When you use a spell slot of 5th level or higher, you can maintain your concentration on the spell for up to 24 hours.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 1070,
      german: 'Zephyrschlag',
      english: 'Zephyr Strike',
      image: 'zephyrStrike',
      level: 1,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dein Körper wird von Winden umhüllt und du löst keine Gelegenheitsangriffe aus. Einmalig verursacht dein nächster Angriffzusätzlich 1W8 Kraftschaden. DU hast außerdem Vorteil auf deinen nächsten Waffenangriff.</p>
        `,
      description_en: `
          <p>You move like the wind. Until the spell ends, your movement doesn’t provoke opportunity attacks. Once before the spell ends, you can give yourself advantage on one weapon attack roll on your turn. That attack deals an extra 1d8 force damage on a hit. Whether you hit or miss, your walking speed increases by 30 feet until the end of that turn.</p>

        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1071,
      german: 'Zeremonie',
      english: 'Ceremony',
      image: 'ceremony',
      level: 1,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'hours',
      duration: 0,
      durationType: 'instant',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dein Zauber führt eine Zeremonie durch, die eine der folgenden Auswirkungen hat:</p>
          <ul>
            <li>Du führst eine Hochzeitszeremonie durch, die zwei Kreaturen innerhalb von 10 Fuß voneinander verbindet. Für die nächsten 7 Tage erhalten die Kreaturen einen Bonus von 2 auf ihre Rettungswürfe gegen Furcht und können nicht durch Zauber oder Effekte gezwungen werden, sich voneinander zu entfernen.</li>
            <li>Du führst eine Einweihungszeremonie durch, die eine Fläche von 10 Fuß Radius segnet. Wenn die Kreaturen in der Fläche einen Rettungswurf gegen Furcht oder einen Fluch machen, erhalten sie einen Bonus von 2 auf den Wurf.</li>
            <li>Du führst eine Bestattungszeremonie durch, die die sterblichen Überreste einer Kreatur in einem 5 Fuß Radius reinigt. Die Kreatur wird von Flüchen, Krankheiten und Vergiftungen befreit. Der Zauber hebt auch die Wirkung aller Zauber auf, die die Kreatur in einem untoten Zustand halten.</li>
            <li>Du führst eine Weihezeremonie durch, die eine Fläche von 30 Fuß Radius segnet. Die Fläche wird als heiliges Gebiet behandelt, bis der Zauber endet. Die Kreaturen erhalten Vorteil auf Rettungswürfe gegen Furcht und Untote können die Fläche nicht betreten.</li>
            <li>Du führst eine Segnungszeremonie durch, die eine Kreatur innerhalb von 30 Fuß segnet. Die Kreatur erhält einen Bonus von 2 auf Angriffs- und Rettungswürfe für die nächsten 7 Tage.</li>
            <li>Du führst eine Reinigungszeremonie durch, die eine Kreatur innerhalb von 5 Fuß reinigt. Die Kreatur wird von Flüchen, Krankheiten und Vergiftungen befreit.</li>
          </ul>
        `,
      description_en: `
      <p>You perform one of several religious ceremonies. When you cast the spell, choose one of the following ceremonies, the target of which must be within 10 feet of you throughout the casting.</p>
      <ul>
        <li>Atonement. You touch one willing creature whose alignment has changed, and you make a DC 20 Wisdom (Insight) check. On a successful check, you restore the target to its original alignment.</li>
        <li>Bless Water. You touch one vial of water and cause it to become holy water.</li>
        <li>Coming of Age. You touch one humanoid who is a young adult. For the next 24 hours, whenever the target makes an ability check, it can roll a d4 and add the number rolled to the ability check. A creature can benefit from this rite only once.</li>
        <li>Ceremony of Marriage. You touch adult humanoids willing to be bonded together in marriage. For the next 7 days, each target gains a +2 bonus to AC while they are within 30 feet of each other. A creature can benefit from this rite again only if widowed.</li>
        <li>Funeral Rite. You touch one corpse, and for the next 7 days, the target can’t become undead by any means short of a wish spell.</li>
        <li>Dedication. You touch one humanoid who wishes to be dedicated to your god’s service. For the next 24 hours, whenever the target makes a saving throw, it can roll a d4 and add the number rolled to the save. A creature can benefit from this rite only once.</li>
        <li>Rite of Baptism. You touch one humanoid or a creature that is a monstrosity, a celestial, or a fiend, and you cause it to become a celestial.</li>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 1072,
      german: 'Zorniges Niederstrecken',
      english: 'Wrathful Smite',
      image: 'wrathfulSmite',
      level: 1,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
          <p>Das nächste Mal, wenn du eine Kreatur mit einem Nahkampfwaffenangriff triffst, ehe dieser Zauber endet, fügt dein Angriff zusätzlich 1W6 psychischen Schaden zu. Wenn das Ziel außerdem eine Kreatur ist, muss sie einen Rettungswurf auf Weisheit bestehen, sonst ist sie bis zum Ende des Zaubers Verängstigt. Als Aktion kann eine Kreatur einen Weisheitswurf gegen deinen Zauberrettungswurf-SG ablegen, um ihren Willen zu stählen und diesen Zauber zu beenden.</p>
        `,
      description_en: `
          <p>The next time you hit with a melee weapon attack during this spell’s duration, your attack deals an extra 1d6 psychic damage. Additionally, if the target is a creature, it must make a Wisdom saving throw or be frightened of you until the spell ends. As an action, the creature can make a Wisdom check against your spell save DC to steel its resolve and end this spell.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },

    // #endregion

    // #region level 2

    {
      id: 2000,
      german: 'Nathairs Unheil',
      english: 'Nathair’s Mischief',
      image: 'nathairsMischief',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>You fill a 20-foot cube you can see within range with fey and draconic magic. Roll on the Mischievous Surge table to determine the magical effect produced, and roll again at the start of each of your turns until the spell ends. You can move the cube up to 10 feet before you roll.</p>
        <table class="wiki-content-table">
        <tr>
        <th colspan="2"><strong>Mischievous Surge</strong></th>
        </tr>
        <tr>
        <th><strong>d4</strong></th>
        <th><strong>Effect</strong></th>
        </tr>
        <tr>
        <td>1</td>
        <td>The smell of apple pie fills the air, and each creature in the cube must succeed on a Wisdom saving throw or become charmed by you until the start of your next turn.</td>
        </tr>
        <tr>
        <td>2</td>
        <td>Bouquets of flowers appear all around, and each creature in the cube must succeed on a Dexterity saving throw or be blinded until the start of your next turn as the flowers spray water in their faces.</td>
        </tr>
        <tr>
        <td>3</td>
        <td>Each creature in the cube must succeed on a Wisdom saving throw or begin giggling until the start of your next turn. A giggling creature is incapacitated and uses all its movement to move in a random direction.</td>
        </tr>
        <tr>
        <td>4</td>
        <td>Drops of molasses appear and hover in the cube, turning it into difficult terrain until the start of your next turn.</td>
        </tr>
        </table>
        `,
      description_en: `
          <p>You fill a 20-foot cube you can see within range with fey and draconic magic. Roll on the Mischievous Surge table to determine the magical effect produced, and roll again at the start of each of your turns until the spell ends. You can move the cube up to 10 feet before you roll.</p>
          <table class="wiki-content-table">
          <tr>
          <th colspan="2"><strong>Mischievous Surge</strong></th>
          </tr>
          <tr>
          <th><strong>d4</strong></th>
          <th><strong>Effect</strong></th>
          </tr>
          <tr>
          <td>1</td>
          <td>The smell of apple pie fills the air, and each creature in the cube must succeed on a Wisdom saving throw or become charmed by you until the start of your next turn.</td>
          </tr>
          <tr>
          <td>2</td>
          <td>Bouquets of flowers appear all around, and each creature in the cube must succeed on a Dexterity saving throw or be blinded until the start of your next turn as the flowers spray water in their faces.</td>
          </tr>
          <tr>
          <td>3</td>
          <td>Each creature in the cube must succeed on a Wisdom saving throw or begin giggling until the start of your next turn. A giggling creature is incapacitated and uses all its movement to move in a random direction.</td>
          </tr>
          <tr>
          <td>4</td>
          <td>Drops of molasses appear and hover in the cube, turning it into difficult terrain until the start of your next turn.</td>
          </tr>
          </table>
        `,
      school: 'illusion',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 2001,
      german: 'Rimes eisiger Griff',
      english: 'Rimes Binding Ice',
      image: 'rimesBindingIce',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'cold' }],
      doesHeal: false,
      description_de: `
          <p>A burst of cold energy emanates from you in a 30-foot cone. Each creature in that area must make a Constitution saving throw. On a failed save, a creature takes 3d8 cold damage and is hindered by ice formations for 1 minute, or until it or another creature within reach of it uses an action to break away the ice. A creature hindered by ice has its speed reduced to 0. On a successful save, a creature takes half as much damage and isn’t hindered by ice.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for each slot level above 2.</p>
          `,
      description_en: `
          <p>A burst of cold energy emanates from you in a 30-foot cone. Each creature in that area must make a Constitution saving throw. On a failed save, a creature takes 3d8 cold damage and is hindered by ice formations for 1 minute, or until it or another creature within reach of it uses an action to break away the ice. A creature hindered by ice has its speed reduced to 0. On a successful save, a creature takes half as much damage and isn’t hindered by ice.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for each slot level above 2.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cone',
      length: 30,
    },
    {
      id: 2002,
      german: 'Aganazzars Versenger',
      english: "Aganazzar's Scorcher",
      image: 'aganazzarsScorcher',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Eine brüllende Stichflamme von 9 m Länge und 1,50 m Breite entspringt in gerader Linie aus dir in eine Richtung deiner Wahl.</p>
          <p>Jede Kreatur auf dieser Linie muss einen Geschicklichkeitsrettungswurf durchführen. Bei einem Fehlschlag erleidet eine Kreatur 3d8 Feuerschaden, die Hälfte davon bei einem Erfolg.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1d8 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>A line of roaring flame 30 feet long and 5 feet wide emanates from you in a direction you choose. Each creature in the line must make a Dexterity saving throw. A creature takes 3d8 fire damage on a failed save, or half as much damage on a successful one.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for each slot level above 2.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 30,
    },
    {
      id: 2003,
      german: 'Arkanes Schloss',
      english: 'Arcane Lock',
      image: 'arcaneLock',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'permanent',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wenn du eine geschlossene Tür, ein Fenster, ein Tor, eine Truhe oder einen anderen Eingang berührst, wird das Ziel für die Dauer des Zaubers verschlossen. Du und die Kreaturen, die du beim Sprechen dieses Zaubers bestimmst, können das Objekt normal öffnen. Du kannst auch ein Passwort festlegen, das, wenn es im Umkreis von 1,5 Metern um das Objekt gesprochen wird, diesen Zauber für 1 Minute unterdrückt. Andernfalls ist das Objekt unpassierbar, bis es zerstört wird oder der Zauber gebannt oder unterdrückt wird. Der Zauber "Klopfen" auf das Objekt unterdrückt das arkane Schloss für 10 Minuten.</p>
          <p>Während dieser Zauber auf das Objekt wirkt, ist es schwieriger, es aufzubrechen oder gewaltsam zu öffnen; der SG, um es aufzubrechen oder irgendwelche Schlösser zu knacken, erhöht sich um 10.</p>
          `,
      description_en: `
          <p>You touch a closed door, window, gate, chest, or other entryway, and it becomes locked for the duration. You and the creatures you designate when you cast this spell can open the object normally. You can also set a password that, when spoken within 5 feet of the object, suppresses this spell for 1 minute. Otherwise, it is impassable until it is broken or the spell is dispelled or suppressed. Casting knock on the object suppresses arcane lock for 10 minutes.</p>
          <p>While affected by this spell, the object is more difficult to break into or force open; the DC to break it or pick any locks on it increases by 10.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2004,
      german: 'Attribut verbessern',
      english: 'Enhance Ability',
      image: 'enhanceAbility',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur und verleihst ihr eine magische Verstärkung. Wähle einen der folgenden Effekte; das Ziel erhält diesen Effekt bis zum Ende des Zaubers.</p>
          <p>Ausdauer des Bären: Das Ziel hat Vorteil auf Konstitutionsproben. Außerdem erhält es 2W6 temporäre Trefferpunkte, die bei Beendigung des Zaubers verloren gehen.</p>
          <p>Stärke des Stiers: Die Zielperson hat Vorteil auf Stärkeproben, und ihre Tragfähigkeit verdoppelt sich.</p>
          <p>Anmut der Katze: Die Zielperson hat Vorteil auf Geschicklichkeitsproben. Außerdem erleidet es keinen Schaden, wenn es 6 Meter oder weniger fällt, sofern es nicht außer Gefecht gesetzt ist.</p>
          <p>Pracht des Adlers: Das Ziel hat Vorteil auf Charismaproben.</p>
          <p>Schlauheit des Fuchses: Das Ziel hat Vorteil auf Intelligenzproben.</p>
          <p>Weisheit der Eule: Das Ziel hat Vorteil auf Weisheitsproben.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, kannst du eine zusätzliche Kreatur für jeden Zauberplatz-Grad über dem 2 als Ziel wählen.</p>
        `,
      description_en: `
          <p>You touch a creature and bestow it with a magical enhancement. Choose one of the following effects; the target gains that effect until the spell ends.</p>
          <p>Bear’s Endurance: The target has advantage on Constitution checks. It also gains 2d6 temporary hit points, which are lost when the spell ends.</p>
          <p>Bull’s Strength: The target has advantage on Strength checks, and its carrying capacity doubles.</p>
          <p>Cat’s Grace: The target has advantage on Dexterity checks. It also doesn’t take damage from falling 20 feet or less if it isn’t incapacitated.</p>
          <p>Eagle’s Splendor: The target has advantage on Charisma checks.</p>
          <p>Fox’s Cunning: The target has advantage on Intelligence checks.</p>
          <p>Owl’s Wisdom: The target has advantage on Wisdom checks.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, you can target one additional creature for each slot level above 2.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2005,
      german: 'Beistand',
      english: 'Aid',
      image: 'aid',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dein Zauber stärkt die Entschlossenheit und Zähigkeit deiner Verbündeten. Wähle bis zu drei Kreaturen in Reichweite. Die maximalen und aktuellen Trefferpunkte aller Ziele steigen für die Wirkungsdauer um 5.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, dann steigen die Trefferpunkte des Ziels um zusätzlich 5 für jeden Zauberplatz-Grad über dem zweiten.</p>
        `,
      description_en: `
          <p>Your spell bolsters your allies with toughness and resolve. Choose up to three creatures within range. Each target’s hit point maximum and current hit points increase by 5 for the duration.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, a target’s hit points increase by an additional 5 for each slot level above 2nd.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2006,
      german: 'Blindheit/Taubheit',
      english: 'Blindness/Deafness',
      image: 'blindnessDeafness',
      level: 2,
      isCustom: false,
      classes: ['test', 'bar', 'cleric', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du kannst einen Feind blind oder taub werden lassen. Wähle eine Kreatur in Reichweite, die du sehen kannst. Sie muss einen Rettungswurf auf Konstitution machen. Wenn dieser misslingt, ist das Ziel für die Wirkungsdauer entweder blind oder taub (deine Wahl). Zu Beginn eines jeden seiner Züge kann das Ziel einen Rettungswurf auf Konstitution machen. Bei einem Erfolg endet der Zauber.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, kannst du ein zusätzliches Ziel für jeden Zauberplatz-Grad über dem 2 wählen.</p>
        `,
      description_en: `
          <p>You can blind or deafen a foe. Choose one creature that you can see within range to make a Constitution saving throw. If it fails, the target is either blinded or deafened (your choice) for the duration. At the beginning of each of its turns, the target can make a Constitution saving throw. On a success, the spell ends.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, you can target one additional creature for each slot level above 2.</p>
        `,
      school: 'necromancy',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2007,
      german: 'Brandmarkendes Niederstrecken',
      english: 'Branding Smite',
      image: 'brandingSmite',
      level: 2,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
          <p>Wenn du das nächste Mal eine Kreatur mit einem Waffenangriff triffst, bevor dieser Zauber endet, schimmert die Waffe beim Schlag in astralem Glanz. Der Angriff fügt dem Ziel zusätzlich 2W6 gleißender Schaden zu, es wird sichtbar, falls es unsichtbar ist, und das Ziel wirft in einem Radius von 1,5 Metern ein schwaches Licht ab und kann nicht unsichtbar werden, bis der Zauber endet.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der zusätzliche Schaden um 1W6 für jeden Grad des Zauberplatzes über dem 2.</p>
        `,
      description_en: `
          <p>The next time you hit a creature with a weapon attack before this spell ends, the weapon gleams with astral radiance as you strike. The attack deals an extra 2d6 radiant damage to the target, which becomes visible if it is invisible, and the target sheds dim light in a 5-foot radius and can’t become invisible until the spell ends.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the extra damage increases by 1d6 for each slot level above 2nd.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2008,
      german: 'Dauerhafte Flamme',
      english: 'Continual Flame',
      image: 'continualFlame',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'permanent',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Aus einem Gegenstand, den du berührst, entspringt eine Flamme, die so hell ist wie eine Taschenlampe. Der Effekt sieht aus wie eine normale Flamme, aber sie erzeugt keine Hitze und verbraucht keinen Sauerstoff. Eine kontinuierliche Flamme kann abgedeckt oder versteckt werden, aber nicht erstickt oder ausgelöscht werden.</p>
          `,
      description_en: `
          <p>A flame, equivalent in brightness to a torch, springs forth from an object that you touch. The effect looks like a regular flame, but it creates no heat and doesn’t use oxygen. A continual flame can be covered or hidden but not smothered or quenched.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2009,
      german: 'Dolchwolke',
      english: 'Cloud of Daggers',
      image: 'cloudOfDaggers',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 4, diceType: 4, damageType: 'slashing' }],
      doesHeal: false,
      description_de: `
          <p>Du füllst die Luft in einem Würfel mit 1,5 Meter Kantenlänge, der um einen Punkt deiner Wahl in Reichweite zentriert ist, mit wirbelnden Dolchen. Eine Kreatur erleidet 4W4 Hiebschaden, wenn sie den Bereich des Zaubers das erste Mal in einem Zug betritt oder ihren Zug dort beginnt.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 2W4 für jeden Zauberplatz-Grad über dem 2.</p>
        `,
      description_en: `
          <p>You fill the air with spinning daggers in a cube 5 feet on each side, centered on a point you choose within range. A creature takes 4d4 slashing damage when it enters the spell’s area for the first time on a turn or starts its turn there.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 2d4 for each slot level above 2.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 2010,
      german: 'Dornenwuchs',
      english: 'Spike Growth',
      image: 'spikeGrowth',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 4, damageType: 'piercing' }],
      doesHeal: false,
      description_de: `
          <p>Der Boden in einem Radius von 6 Metern um einen Punkt innerhalb der Reichweite verdreht sich und treibt harte Stacheln und Dornen aus. Das Gebiet wird für die Dauer des Angriffs zu schwierigem Terrain. Wenn sich eine Kreatur in das Gebiet oder innerhalb des Gebiets bewegt, erleidet sie alle 1,5 Meter 2W4 Stichschäden.</p>
          <p>Die Verwandlung des Bodens ist so getarnt, dass sie natürlich aussieht. Jede Kreatur, die das Gebiet zum Zeitpunkt des Zaubers nicht sehen kann, muss eine Weisheitsprobe (Wahrnehmung) gegen deinen Zauberrettungswurf SG machen, um das Gelände als gefährlich zu erkennen, bevor sie es betritt.</p>
        `,
      description_en: `
          <p>The ground in a 20-foot radius centered on a point within range twists and sprouts hard spikes and thorns. The area becomes difficult terrain for the duration. When a creature moves into or within the area, it takes 2d4 piercing damage for every 5 feet it travels.</p>
          <p>The transformation of the ground is camouflaged to look natural. Any creature that can’t see the area at the time the spell is cast must make a Wisdom (Perception) check against your spell save DC to recognize the terrain as hazardous before entering it.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 150,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 211,
      german: 'Drachenodem',
      english: 'Dragon’s Breath',
      image: 'dragonsBreath',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>You touch one willing creature and imbue it with the power to spew magical energy from its mouth, provided it has one. Choose acid, cold, fire, lightning, or poison. Until the spell ends, the creature can use an action to exhale energy of the chosen type in a 15-foot cone. Each creature in that area must make a Dexterity saving throw, taking 3d6 damage of the chosen type on a failed save, or half as much damage on a successful one.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2nd.</p>
        `,
      description_en: `
          <p>You touch one willing creature and imbue it with the power to spew magical energy from its mouth, provided it has one. Choose acid, cold, fire, lightning, or poison. Until the spell ends, the creature can use an action to exhale energy of the chosen type in a 15-foot cone. Each creature in that area must make a Dexterity saving throw, taking 3d6 damage of the chosen type on a failed save, or half as much damage on a successful one.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2nd.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2012,
      german: 'Dunkelheit',
      english: 'Darkness',
      image: 'darkness',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Magische Dunkelheit breitet sich von einem Punkt deiner Wahl innerhalb der Reichweite aus und füllt für die Dauer des Zaubers eine Kugel mit einem Radius von 5 Metern. Die Dunkelheit breitet sich um Ecken herum aus. Eine Kreatur mit Nachtsicht kann diese Dunkelheit nicht durchschauen, und nichtmagisches Licht kann sie nicht erhellen. Befindet sich der von dir gewählte Punkt auf einem Gegenstand, den du in der Hand hältst oder der nicht getragen wird, geht die Dunkelheit von diesem Gegenstand aus und bewegt sich mit ihm. Wenn du die Quelle der Dunkelheit mit einem undurchsichtigen Gegenstand, wie einer Schale oder einem Helm, vollständig abdeckst, wird die Dunkelheit blockiert. Überschneidet sich ein Bereich dieses Zaubers mit einem Lichtbereich, der durch einen Zauber des 2. Grades oder niedriger erzeugt wurde, wird der Zauber, der das Licht erzeugt hat, gebannt.</p>
          `,
      description_en: `
          <p>Magical darkness spreads from a point you choose within range to fill a 15-foot-radius sphere for the duration. The darkness spreads around corners. A creature with darkvision can’t see through this darkness, and nonmagical light can’t illuminate it. If the point you choose is on an object you are holding or one that isn’t being worn or carried, the darkness emanates from the object and moves with it. Completely covering the source of the darkness with an opaque object, such as a bowl or a helm, blocks the darkness. If any of this spell’s area overlaps with an area of light created by a spell of 2nd level or lower, the spell that created the light is dispelled.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 15,
    },
    {
      id: 2013,
      german: 'Dunkelsicht',
      english: 'Darkvision',
      image: 'darkvision',
      level: 2,
      isCustom: false,
      classes: ['test', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dein Zauber verleiht einem willigen Geschöpf die Fähigkeit, in der Dunkelheit zu sehen. Für die Dauer des Zaubers erhält das Ziel Dunkelsicht auf eine Reichweite von 18 Metern.</p>
        `,
      description_en: `
          <p>Your spell grants a willing creature the ability to see in the dark. For the duration, that creature has darkvision out to a range of 60 feet.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2014,
      german: 'Einflüsterung',
      english: 'Suggestion',
      image: 'suggestion',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du schlägst eine Handlung vor (begrenzt auf ein oder zwei Sätze) und beeinflusst auf magische Weise eine Kreatur, die du in Reichweite siehst, die dich hören und verstehen kann. Kreaturen, die nicht verzaubert werden können, sind gegen diesen Effekt immun. Die Suggestion muss so formuliert sein, dass die Handlung vernünftig klingt. Wenn du die Kreatur aufforderst, sich selbst zu erstechen, auf einen Speer zu werfen, sich zu opfern oder eine andere offensichtlich schädliche Handlung auszuführen, wird der Zauber beendet.</p>
          <p>Das Ziel muss einen Rettungswurf auf Weisheit machen. Bei einem misslungenen Rettungswurf verfolgt es die von dir beschriebene Handlung nach bestem Wissen und Gewissen. Die vorgeschlagene Handlung kann während der gesamten Dauer fortgesetzt werden. Wenn die vorgeschlagene Aktivität in kürzerer Zeit abgeschlossen werden kann, endet der Zauber, wenn die Zielperson das getan hat, worum sie gebeten wurde.</p>
          <p>Du kannst auch Bedingungen angeben, die während der Dauer eine besondere Aktivität auslösen. Du könntest zum Beispiel vorschlagen, dass eine Ritterin ihr Schlachtross dem ersten Bettler gibt, dem sie begegnet. Wenn die Bedingung nicht erfüllt wird, bevor der Zauber abläuft, wird die Aktivität nicht ausgeführt.</p>
          <p>Wenn du oder einer deiner Gefährten dem Ziel Schaden zufügt, endet der Zauber.</p>
          `,
      description_en: `
          <p>You suggest a course of activity (limited to a sentence or two) and magically influence a creature you can see within range that can hear and understand you. Creatures that can’t be charmed are immune to this effect. The suggestion must be worded in such a manner as to make the course of action sound reasonable. Asking the creature to stab itself, throw itself onto a spear, immolate itself, or do some other obviously harmful act ends the spell.</p>
          <p>The target must make a Wisdom saving throw. On a failed save, it pursues the course of action you described to the best of its ability. The suggested course of action can continue for the entire duration. If the suggested activity can be completed in a shorter time, the spell ends when the subject finishes what it was asked to do.</p>
          <p>You can also specify conditions that will trigger a special activity during the duration. For example, you might suggest that a knight give her warhorse to the first beggar she meets. If the condition isn’t met before the spell expires, the activity isn’t performed.</p>
          <p>If you or any of your companions damage the target, the spell ends.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2015,
      german: 'Erdbindung',
      english: 'Earthbind',
      image: 'earthbind',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wähle eine Kreatur in Reichweite, die du siehst. Gelbe Streifen aus magischer Energie umkreisen die Kreatur. Das Ziel muss einen Rettungswurf auf Stärke bestehen oder seine Fluggeschwindigkeit (falls vorhanden) wird für die Dauer des Zaubers auf 0 reduziert. Eine Kreatur in der Luft, die von diesem Zauber betroffen ist, sinkt mit 18 Metern pro Runde, bis sie den Boden erreicht oder der Zauber endet.</p>
          `,
      description_en: `
          <p>Choose a creature you can see within range. Yellow strips of magical energy loop around the creature. The target must succeed on a Strength saving throw or its flying speed (if any) is reduced to 0 for the spell’s duration. An airborne creature affected by this spell descends at 60 feet per round until it reaches the ground or the spell ends.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 300,
      hasAreaOfEffect: false,
    },
    {
      id: 2016,
      german: 'Fallen finden',
      english: 'Find Traps',
      image: 'findTraps',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du spürst das Vorhandensein jeder Falle, die sich in Sichtweite befindet. Eine Falle im Sinne dieses Zaubers ist alles, was eine plötzliche oder unerwartete Wirkung hat, die du als schädlich oder unerwünscht ansiehst und die von ihrem Schöpfer ausdrücklich als solche beabsichtigt wurde. Der Zauber würde also einen Bereich aufspüren, der von dem Zauber "Alarm (Stufe 1)", einer "Glyphe der Abwehr (Stufe 3)" oder einer mechanischen Grubenfalle betroffen ist, aber er würde keine natürliche Schwachstelle im Boden, eine instabile Decke oder ein verstecktes Erdloch aufdecken.</p>
          <p>Dieser Zauber zeigt lediglich an, dass eine Falle vorhanden ist. Du erfährst nicht, wo sich die einzelnen Fallen befinden, aber du erfährst die allgemeine Art der Gefahr, die von einer Falle ausgeht, die du spürst.</p>
          `,
      description_en: `
          <p>You sense the presence of any trap within range that is within line of sight. A trap, for the purpose of this spell, includes anything that would inflict a sudden or unexpected effect you consider harmful or undesirable, which was specifically intended as such by its creator. Thus, the spell would sense an area affected by the alarm spell, a glyph of warding, or a mechanical pit trap, but it would not reveal a natural weakness in the floor, an unstable ceiling, or a hidden sinkhole.</p>
          <p>This spell merely reveals that a trap is present. You don’t learn the location of each trap, but you do learn the general nature of the danger posed by a trap you sense.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 2017,
      german: 'Fesseln',
      english: 'Hold Person',
      image: 'holdPerson',
      level: 2,
      isCustom: false,
      classes: [
        'test',
        'bard',
        'cleric',
        'paladin',
        'sorcerer',
        'warlock',
        'wizard',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du spinnst einen ablenkenden Wortstrang, der Kreaturen deiner Wahl in Reichweite, die du siehst und die dich hören können, zu einem Rettungswurf auf Weisheit zwingt. Jede Kreatur, die nicht verzaubert werden kann, schafft diesen Rettungswurf automatisch, und wenn du oder deine Gefährten gegen eine Kreatur kämpfen, hat sie Vorteil bei diesem Rettungswurf. Bei einem misslungenen Rettungswurf hat das Ziel Nachteil bei Weisheitsproben (Wahrnehmung), um eine andere Kreatur als dich wahrzunehmen, bis der Zauber endet oder bis das Ziel dich nicht mehr hören kann. Der Zauber endet, wenn du außer Gefecht gesetzt bist oder nicht mehr sprechen kannst.</p>
          `,
      description_en: `
          <p>You attempt to paralyze a creature you can see within range. It must succeed on a Wisdom saving throw or be paralyzed for the duration. This spell has no effect on undead. At the end of each of its turns, the target can make another Wisdom saving throw. On a success, the spell ends on the target.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2018,
      german: 'Feuerwerk',
      english: 'Pyrotechnics',
      image: 'pyrotechnics',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      damage: [{ diceNumber: 1, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Wähle einen Bereich mit einer nichtmagischen Flamme in Reichweite, den du sehen kannst und der in einem Würfel mit maximal 1,5 Meter Kantenlänge passt. Du kannst das Feuer in diesem Bereich löschen und erzeugst dabei entweder ein Feuerwerk oder Rauch.</p>
          <p>Feuerwerk: Das Ziel explodiert mit einem schillernden Farbenspiel. Jede Kreatur in 3 Metern Entfernung vom Ziel muss einen Rettungswurf auf Konstitution bestehen oder wird bis zum Ende deines nächsten Zuges geblendet.</p>
          <p>Rauch: Dichter schwarzer Rauch breitet sich vom Ziel in einem Radius von 6 Metern aus und bewegt sich um Ecken. Das Gebiet des Rauchs ist stark vernebelt. Der Rauch bleibt 1 Minute lang bestehen oder bis ein starker Wind ihn vertreibt.</p>
          `,
      description_en: `
          <p>Choose an area with a nonmagical flame that you can see and that fits within a 5-foot cube within range. You can extinguish the fire in that area, and you create either fireworks or smoke when you do so.</p>
          <p>Fireworks: The target explodes with a dazzling display of colors. Each creature within 10 feet of the target must succeed on a Constitution saving throw or become blinded until the end of your next turn.</p>
          <p>Smoke: Thick black smoke spreads out from the target in a 20-foot radius, moving around corners. The area of the smoke is heavily obscured. The smoke lasts for 1 minute or until a strong wind disperses it.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 2019,
      german: 'Flammenklinge',
      english: 'Flame Blade',
      image: 'flameBlade',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Du beschwörst eine feurige Klinge in deiner freien Hand. Die Klinge ähnelt in Größe und Form einem Krummsäbel und hält für die Dauer des Zaubers an. Wenn du die Klinge loslässt, verschwindet sie, aber du kannst die Klinge als Bonusaktion erneut beschwören.</p>
          <p>Du kannst deine Aktion nutzen, um einen Nahkampf-Zauberangriff mit der feurigen Klinge auszuführen. Bei einem Treffer erleidet das Ziel 3W6 Feuerschaden.</p>
          <p>Die flammende Klinge verströmt helles Licht in einem Radius von 3 Metern und schummriges Licht für weitere 3 Meter.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeweils zwei Zauberplatz-Grade über dem 2.</p>
          `,
      description_en: `
          <p>You evoke a fiery blade in your free hand. The blade is similar in size and shape to a scimitar, and it lasts for the duration. If you let go of the blade, it disappears, but you can evoke the blade again as a bonus action.</p>
          <p>You can use your action to make a melee spell attack with the fiery blade. On a hit, the target takes 3d6 fire damage.</p>
          <p>The flaming blade sheds bright light in a 10-foot radius and dim light for an additional 10 feet.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d6 for every two slot levels above 2nd.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2020,
      german: 'Flammenkugel',
      english: 'Flaming Sphere',
      image: 'flamingSphere',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Eine Feuerkugel mit einem Durchmesser von 1,5 Metern erscheint in einem unbesetzten Raum deiner Wahl innerhalb der Reichweite und hält für die Dauer der Aktion an. Jede Kreatur, die ihren Zug in einem Umkreis von 1,5 Metern um die Kugel beendet, muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 2W6 Feuerschaden, bei einem erfolgreichen Rettungswurf halb so viel Schaden.</p>
          <p>Als Bonusaktion kannst du die Kugel bis zu 9 Meter weit bewegen. Wenn du die Kugel in eine Kreatur rammst, muss diese Kreatur einen Rettungswurf gegen den Schaden der Kugel machen, und die Kugel hört in diesem Zug auf, sich zu bewegen.</p>
          <p>Wenn du die Kugel bewegst, kannst du sie über bis zu 1,5 Meter hohe Barrieren lenken und über bis zu 3 Meter breite Gruben springen. Die Kugel entzündet entflammbare Gegenstände, die du nicht trägst, und spendet helles Licht in einem Radius von 6 Metern und schwaches Licht für weitere 6 Meter.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>A 5-foot-diameter sphere of fire appears in an unoccupied space of your choice within range and lasts for the duration. Any creature that ends its turn within 5 feet of the sphere must make a Dexterity saving throw. The creature takes 2d6 fire damage on a failed save, or half as much damage on a successful one.</p>
          <p>As a bonus action, you can move the sphere up to 30 feet. If you ram the sphere into a creature, that creature must make the saving throw against the sphere’s damage, and the sphere stops moving this turn.</p>
          <p>When you move the sphere, you can direct it over barriers up to 5 feet high and jump it across pits up to 10 feet wide. The sphere ignites flammable objects not being worn or carried, and it sheds bright light in a 20-foot radius and dim light for an additional 20 feet.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 2021,
      german: 'Gebete der Heilung',
      english: 'Prayer of Healing',
      image: 'prayerOfHealing',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 10,
      cost: 'minutes',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 2, diceType: 8, additionalHeal: 0 },
      description_de: `
          <p>Bis zu sechs Kreaturen deiner Wahl in Reichweite, die du sehen kannst, erhalten jeweils 2W8 + deinen Zaubermodifikator Trefferpunkte zurück. Der Zauber hat keine Auswirkungen auf Untote Oder Konstrukte.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich die Heilung um 1W8 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>Up to six creatures of your choice that you can see within range each regain hit points equal to 2d8 + your spellcasting ability modifier. This spell has no effect on undead or constructs.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the healing increases by 1d8 for each slot level above 2.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2022,
      german: 'Gedanken wahrnehmen',
      english: 'Detect Thoughts',
      image: 'detectThoughts',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Für die Dauer des Zaubers kannst du die Gedanken von bestimmten Kreaturen lesen. Wenn du den Zauber sprichst und als deine Aktion in jeder Runde bis zum Ende des Zaubers, kannst du deine Gedanken auf eine beliebige Kreatur richten, die du im Umkreis von 9 Metern sehen kannst. Wenn die Kreatur, die du auswählst, eine Intelligenz von 3 oder weniger hat oder keine Sprache spricht, ist sie nicht betroffen.</p>
          <p>Anfänglich erfasst du die oberflächlichen Gedanken der Kreatur - also das, was ihr in diesem Moment am meisten durch den Kopf geht. Als Aktion kannst du entweder deine Aufmerksamkeit auf die Gedanken einer anderen Kreatur lenken oder versuchen, tiefer in die Gedanken derselben Kreatur einzudringen. Wenn du tiefer eindringst, muss das Ziel einen Rettungswurf auf Weisheit machen. Wenn er misslingt, erhältst du Einblick in seinen Verstand (falls vorhanden), seinen emotionalen Zustand und etwas, das in seinen Gedanken eine große Rolle spielt (z.B. etwas, worüber es sich Sorgen macht, das es liebt oder hasst). Wenn er erfolgreich ist, endet der Zauber. In jedem Fall weiß das Ziel, dass du in seine Gedanken eindringst, und solange du deine Aufmerksamkeit nicht auf die Gedanken einer anderen Kreatur lenkst, kann die Kreatur ihre Aktion in ihrem Zug nutzen, um eine Intelligenzprobe zu machen, die von deiner Intelligenzprobe angefochten wird; wenn sie erfolgreich ist, endet der Zauber.</p>
          <p>Fragen, die verbal an die Zielkreatur gerichtet werden, prägen natürlich den Verlauf ihrer Gedanken, daher ist dieser Zauber besonders effektiv als Teil eines Verhörs. Du kannst diesen Zauber auch benutzen, um die Anwesenheit von denkenden Kreaturen zu entdecken, die du nicht sehen kannst. Wenn du den Zauber sprichst oder als deine Aktion während der Dauer des Zaubers, kannst du im Umkreis von 9 Metern um dich herum nach Gedanken suchen. Der Zauber kann Barrieren durchdringen, aber 2 Fuß Fels, 2 Zoll Metall (außer Blei) oder eine dünne Bleiplatte blockieren dich. Du kannst keine Kreaturen mit einer Intelligenz von 3 oder weniger oder Kreaturen, die keine Sprache sprechen, aufspüren.</p>
          <p>Wenn du die Anwesenheit einer Kreatur auf diese Weise aufspürst, kannst du für den Rest der Dauer des Zaubers wie oben beschrieben ihre Gedanken lesen, auch wenn du sie nicht sehen kannst, aber sie muss sich noch in Reichweite befinden.</p>
          `,
      description_en: `
          <p>For the duration, you can read the thoughts of certain creatures. When you cast the spell and as your action in each round until the spell ends, you can focus your mind on any one creature that you can see within 30 feet of you. If the creature you choose has an Intelligence of 3 or lower or doesn’t speak any language, the creature is unaffected.</p>
          <p>You initially learn the surface thoughts of the creature - what is most on its mind in that moment. As an action, you can either shift your attention to another creature’s thoughts or attempt to probe deeper into the same creature’s mind. If you probe deeper, the target must make a Wisdom saving throw. If it fails, you gain insight into its reasoning (if any), its emotional state, and something that looms large in its mind (such as something it worries over, loves, or hates). If it succeeds, the spell ends. Either way, the target knows that you are probing into its mind, and unless you shift your attention to another creature’s thoughts, the creature can use its action on its turn to make an Intelligence check contested by your Intelligence check; if it succeeds, the spell ends.</p>
          <p>Questions verbally directed at the target creature naturally shape the course of its thoughts, so this spell is particularly effective as part of an interrogation. You can also use this spell to detect the presence of thinking creatures you can’t see. When you cast the spell or as your action during the duration, you can search for thoughts within 30 feet of you. The spell can penetrate barriers, but 2 feet of rock, 2 inches of any metal other than lead, or a thin sheet of lead blocks you. You can’t detect a creature with an Intelligence of 3 or lower or one that doesn’t speak any language.</p>
          <p>If you detect the presence of a creature in this way, you can read its thoughts for the rest of the duration as described above, even if you can’t see it, but it must still be within range.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2023,
      german: 'Gedankendorn',
      english: 'Mind Spike',
      image: 'mindSpike',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
          <p>Du greifst in den Verstand einer Kreatur, die du innerhalb der Reichweite sehen kannst. Das Ziel muss einen Weisheitsrettungswurf ablegen. Bei einem Fehlschlag erleidet es 3W8 psychischen Schaden, die Hälfte davon bei einem Erfolg. Bei einem Fehlschlag kennst du jederzeit den Aufenthaltsort des Ziels, bis der Zauber endet, aber nur während du und das Ziel auf derselben Ebene der Existenz sind. Während du dieses Wissen hast, kann sich das Ziel nicht vor dir verstecken, und wenn es unsichtbar ist, erhält es dir gegenüber keine Vorteile aus diesem Zustand.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>You reach into the mind of a creature you can see within range. The target must make a Wisdom saving throw. On a failed save, the target takes 3d8 psychic damage, and you know the target’s location until the spell ends, but only while the two of you are on the same plane of existence. While you have this knowledge, the target can’t become hidden from you, and if it’s invisible, it gains no benefit from that condition against you.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for each slot level above 2.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2024,
      german: 'Gefühle besänftigen',
      english: 'Calm Emotions',
      image: 'calmEmotions',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'charisma',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du versuchst, die starken Empfindungen einer Personengruppe zu unterdrücken. Alle Humanoiden in einer Sphäre mit einem Radius von 6 Metern, zentriert um einen Punkt deiner Wahl in Reichweite, müssen einen Charisma-RW ablegen: die Kreaturen können diesen Rettungswurf freiwillig nicht bestehen, wenn sie es wollen. Wenn eine Kreatur ihren Rettungswurf nicht schafft, wähle einen der folgenden zwei Effekte.</p>
          <p>Du kannst jeden Effekt unterdrücken, der dafür sorgt, dass das Ziel bezaubert oder verängstigt ist. Wenn dieser Zauber endet, tritt der unterdrückte Effekt wieder in Kraft, vorausgesetzt, seine Wirkung ist in der Zwischenzeit nicht beendet.</p>
          <p>Alternativ kannst du ein Ziel gleichgültig gegenüber Kreaturen deiner Wahl machen, denen es feindlich gesonnen ist. Diese Gleichgültigkeit endet, wenn das Ziel angegriffen oder von einem Zauber verletzt wird oder sieht, wie einer seiner Freunde verletzt wird. Wenn der Zauber endet, wird die Kreatur wieder feindselig, wenn der Spielleiter nichts anderes festlegt.</p>
          `,
      description_en: `
          <p>You attempt to suppress strong emotions in a group of people. Each humanoid in a 20-foot-radius sphere centered on a point you choose within range must make a Charisma saving throw; a creature can choose to fail this saving throw if it wishes. If a creature fails its saving throw, choose one of the following two effects.</p>
          <p>You can suppress any effect causing a target to be charmed or frightened. When this spell ends, any suppressed effect resumes, provided that its duration has not expired in the meantime.</p>
          <p>Alternatively, you can make a target indifferent about creatures of your choice that it is hostile toward. This indifference ends if the target is attacked or harmed by a spell or if it witnesses any of its friends being harmed. When the spell ends, the creature becomes hostile again, unless the DM rules otherwise.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 2025,
      german: 'Gegenstand aufspüren',
      english: 'Locate Object',
      image: 'locateObject',
      level: 2,
      isCustom: false,
      classes: [
        'test',
        'bard',
        'druid',
        'cleric',
        'wizard',
        'paladin',
        'ranger',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Beschreibe oder benenne ein Objekt, das dir bekannt ist. Du spürst die Richtung, in der sich das Objekt befindet, solange es sich nicht weiter als 300 Meter von dir entfernt befindet. Wenn das Objekt in Bewegung ist, kennst du die Richtung seiner Bewegung.</p>
          <p>Der Zauber kann ein bestimmtes, dir bekanntes Objekt lokalisieren, wenn du es mindestens einmal aus der Nähe - innerhalb von 9 Metern - gesehen hast. Alternativ kann der Zauber den nächstgelegenen Gegenstand einer bestimmten Art lokalisieren, z.B. eine bestimmte Art von Kleidung, Schmuck, Möbeln, Werkzeug oder Waffen.</p>
          <p>Dieser Zauber kann einen Gegenstand nicht lokalisieren, wenn eine dünne Schicht Blei den direkten Weg zwischen dir und dem Gegenstand versperrt, selbst wenn es sich um ein dünnes Blatt handelt.</p>
        `,
      description_en: `
          <p>Describe or name an object that is familiar to you. You sense the direction to the object’s location, as long as that object is within 1,000 feet of you. If the object is in motion, you know the direction of its movement.</p>
          <p>The spell can locate a specific object known to you, as long as you have seen it up close – within 30 feet – at least once. Alternatively, the spell can locate the nearest object of a particular kind, such as a certain kind of apparel, jewelry, furniture, tool, or weapon.</p>
          <p>This spell can’t locate an object if any thickness of lead, even a thin sheet, blocks a direct path between you and the object.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 1000,
      hasAreaOfEffect: false,
    },
    {
      id: 2026,
      german: 'Gestalt verändern',
      english: 'Alter Self',
      image: 'alterSelf',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du nimmst eine andere Gestalt an. Wenn du den Zauber sprichst, wählst du eine der folgenden Optionen, deren Wirkung für die Dauer des Zaubers anhält. Solange der Zauber andauert, kannst du eine Option als Aktion beenden, um die Vorteile einer anderen zu erhalten.</p>
          <p>Aquatische Anpassung: Du passt deinen Körper an eine Wasserumgebung an, lässt Kiemen sprießen und Schwimmhäute zwischen deinen Fingern wachsen. Du kannst unter Wasser atmen und erhältst eine Schwimmgeschwindigkeit, die deiner Gehgeschwindigkeit entspricht.</p>
          <p>Aussehen verändern: Du verwandelst dein Aussehen. Du entscheidest, wie du aussiehst, einschließlich deiner Größe, deines Gewichts, deiner Gesichtszüge, des Klangs deiner Stimme, deiner Haarlänge, deiner Haarfarbe und deiner besonderen Merkmale, falls vorhanden. Du kannst dich in ein Mitglied einer anderen Rasse verwandeln, obwohl sich keine deiner Statistiken ändern. Du kannst auch nicht als eine Kreatur erscheinen, die eine andere Größe hat als du, und deine Grundform bleibt gleich; wenn du z.B. zweibeinig bist, kannst du diesen Zauber nicht benutzen, um vierbeinig zu werden. Während der Dauer des Zaubers kannst du jederzeit deine Aktion nutzen, um dein Aussehen erneut auf diese Weise zu verändern.</p>
          <p>Natürliche Waffen: Dir wachsen Klauen, Reißzähne, Stacheln, Hörner oder eine andere natürliche Waffe deiner Wahl. Deine waffenlosen Schläge fügen 1W6 Hieb-, Stich- oder Schnittschaden zu, je nachdem, welche natürliche Waffe du gewählt hast, und du bist geübt im Umgang mit deinen waffenlosen Schlägen. Außerdem ist die natürliche Waffe magisch und du erhältst einen Bonus von +1 auf Angriffs- und Schadenswürfe, die du mit ihr ausführst.</p>
          `,
      description_en: `
          <p>You assume a different form. When you cast the spell, choose one of the following options, the effects of which last for the duration of the spell. While the spell lasts, you can end one option as an action to gain the benefits of a different one.</p>
          <p>Aquatic Adaptation: You adapt your body to an aquatic environment, sprouting gills and growing webbing between your fingers. You can breathe underwater and gain a swimming speed equal to your walking speed.</p>
          <p>Change Appearance. You transform your appearance. You decide what you look like, including your height, weight, facial features, sound of your voice, hair length, coloration, and distinguishing characteristics, if any. You can make yourself appear as a member of another race, though none of your statistics change. You also don't appear as a creature of a different size than you, and your basic shape stays the same, if you're bipedal, you can't use this spell to become quadrupedal, for instance. At any time for the duration of the spell, you can use your action to change your appearance in this way again.</p>
          <p>Natural Weapons. You grow claws, fangs, spines, horns, or a different natural weapon of your choice. Your unarmed strikes deal 1d6 bludgeoning, piercing, or slashing damage, as appropriate to the natural weapon you chose, and you are proficient with your unarmed strikes. Your natural weapon is magical, and you have a +1 bonus to the attack and damage rolls you make using it.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2027,
      german: 'Heilungsgeist',
      english: 'Healing Spirit',
      image: 'healingSpirit',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 1, diceType: 6, additionalHeal: 0 },
      description_de: `
          <p>Du rufst einen Naturgeist herbei, um Verwundete zu versorgen. Der körperlose Geist erscheint in einem Würfel von 1,50 m Kantenlänge in einem Bereich, den du innerhalb der Reichweite sehen kannst. Der Geist erscheint als durchsichtiges Tier oder Feenwesen (deine Wahl).</p>
          <p>Bis zum Ende des Zaubers wird der Geist bei einer Kreatur, die ihren Zug im Bereich des Geistes beginnt oder ihn zum ersten Mal in ihrem Zug betritt, 1W6 Trefferpunkte bei dieser Kreatur wiederherstellen (erfordert keine Aktion). Der Geist kann weder Konstrukte noch Untote heilen.</p>
          <p>Als Bonusaktion in deinem Zug kannst du den Geist um bis zu 9 m an einen Ort bewegen, den du sehen kannst.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich die Heilung um 1W6 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>You call forth a nature spirit to soothe the wounded. The intangible spirit appears in a space that is a 5-foot cube you can see within range. The spirit looks like a transparent beast or fey (your choice).</p>
          <p>Until the spell ends, whenever you or a creature you can see moves into the spirit’s space for the first time on a turn or starts its turn there, you can cause the spirit to restore 1d6 hit points to that creature (no action required). The spirit can’t heal constructs or undead.</p>
          <p>As a bonus action on your turn, you can move the spirit up to 30 feet to a space you can see.</p>
          <p><b>At Higher Levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the healing increases by 1d6 for each slot level above 2.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 2028,
      german: 'Himmelschreiben',
      english: 'Skywrite',
      image: 'skywrite',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'days',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du lässt bis zu zehn Worte in einem Teil des Himmels entstehen, den du sehen kannst. Die Worte scheinen aus Wolken zu bestehen und bleiben für die Dauer des Zaubers an ihrem Platz. Die Worte lösen sich auf, wenn der Zauber endet. Ein starker Wind kann die Wolken zerstreuen und den Zauber vorzeitig beenden.</p>
          `,
      description_en: `
          <p>You cause up to ten words to form in a part of the sky you can see. The words appear to be made of cloud and remain in place for the spell’s duration. The words dissipate when the spell ends. A light wind can disperse the clouds and end the spell early.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 99999,
      hasAreaOfEffect: false,
    },
    {
      id: 2029,
      german: 'Klopfen',
      english: 'Knock',
      image: 'knock',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wähle ein Objekt in Reichweite, das du sehen kannst. Das Objekt kann eine Tür, eine Kiste, eine Truhe, eine Fessel, ein Vorhängeschloss oder ein anderes Objekt sein, das ein weltliches oder magisches Mittel enthält, das den Zugang verhindert.</p>
          <p>Ein Ziel, das durch ein weltliches Schloss verschlossen oder verriegelt ist, wird entriegelt, gelöst oder entsperrt. Wenn das Objekt mehrere Schlösser hat, wird nur eines davon entriegelt.</p>
          <p>Wenn du ein Ziel auswählst, das mit "Arkanes Schloss (Lvl 2)" verschlossen gehalten wird, wird dieser Zauber für 10 Minuten unterdrückt, während dieser Zeit kann das Ziel normal geöffnet und verschlossen werden.</p>
          <p>Wenn du den Zauber sprichst, geht ein lautes Klopfen von dem Zielobjekt aus, das bis zu 100 Meter weit zu hören ist.</p>
          `,
      description_en: `
          <p>Choose an object that you can see within range. The object can be a door, a box, a chest, a set of manacles, a padlock, or another object that contains a mundane or magical means that prevents access.</p>
          <p>A target that is held shut by a mundane lock or that is stuck or barred becomes unlocked, unstuck, or unbarred. If the object has multiple locks, only one of them is unlocked.</p>
          <p>If you choose a target that is held shut with "Arcane Lock (Lvl 2)", this spell is suppressed for 10 minutes, during which time the target can be opened and shut normally.</p>
          <p>When you cast the spell, a loud knock, audible from as far away as 300 feet, emanates from the target object.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2030,
      german: 'Krone des Wahnsinns',
      english: 'Crown of Madness',
      image: 'crownOfMadness',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein Humanoid deiner Wahl in Reichweite, den du sehen kannst, muss einen Rettungswurf auf Weisheit bestehen oder wird für die Dauer des Zaubers von dir verzaubert.</p>
          <p>Während das Ziel auf diese Weise verzaubert wird, erscheint eine verdrehte Krone aus gezacktem Eisen auf seinem Kopf, und in seinen Augen leuchtet ein Wahnsinn.</p>
          <p>Das verzauberte Ziel muss seine Aktion nutzen, bevor es sich in jeder seiner Runden bewegt, um einen Nahkampfangriff gegen eine andere Kreatur als sich selbst auszuführen, die du im Geiste wählst. Das Ziel kann in seinem Zug normal agieren, wenn du keine Kreatur auswählst oder wenn sich keine in seiner Reichweite befindet.</p>
          <p>In deinen folgenden Zügen musst du deine Aktion nutzen, um die Kontrolle über das Ziel zu behalten, oder der Zauber endet. Außerdem kann das Ziel am Ende jedes seiner Züge einen Rettungswurf auf Weisheit machen. Bei einem Erfolg endet der Zauber.</p>
          `,
      description_en: `
          <p>One humanoid of your choice that you can see within range must succeed on a Wisdom saving throw or become charmed by you for the duration. While the target is charmed in this way, a twisted crown of jagged iron appears on its head, and a madness glows in its eyes.</p>
          <p>The charmed target must use its action before moving on each of its turns to make a melee attack against a creature other than itself that you mentally choose. The target can act normally on its turn if you choose no creature or if none are within its reach.</p>
          <p>On your subsequent turns, you must use your action to maintain control over the target, or the spell ends. Also, the target can make a Wisdom saving throw at the end of each of its turns. On a success, the spell ends.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 2031,
      german: 'Macht der Vorstellungskraft',
      english: 'Phantasmal Force',
      image: 'phantasmalForce',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'intelligence',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du stellst eine Illusion her, die sich im Geist einer Kreatur in Reichweite festsetzt, die du sehen kannst. Das Ziel muss einen Rettungswurf auf Intelligenz machen. Bei einem misslungenen Rettungswurf erschaffst du ein Fantasieobjekt, eine Kreatur oder ein anderes sichtbares Phänomen deiner Wahl, das nicht größer als ein 1O-Fuß-Würfel ist und das für die Dauer des Zaubers nur für das Ziel wahrnehmbar ist. Dieser Zauber hat keine Wirkung auf Untote oder Konstrukte.</p>
          <p>Die Fantasie umfasst Geräusche, Temperatur und andere Reize, die ebenfalls nur für die Kreatur wahrnehmbar sind.</p>
          <p>Das Ziel kann seine Aktion nutzen, um die Fantasie mit einer Intelligenzprobe (Nachforschungen) gegen deinen Zauberrettungswurf SG zu untersuchen. Wenn die Probe erfolgreich ist, erkennt das Ziel, dass die Fantasie eine Illusion ist, und der Zauber endet.</p>
          <p>Während das Ziel von dem Zauber betroffen ist, behandelt es die Fantasie so, als wäre sie real. Die Zielperson rationalisiert alle unlogischen Ergebnisse, die sich aus der Interaktion mit der Fantasie ergeben. Ein Beispiel: Ein Ziel, das versucht, über eine imaginäre Brücke zu gehen, die einen Abgrund überspannt, fällt, sobald es die Brücke betritt. Wenn das Ziel den Sturz überlebt, glaubt es immer noch, dass die Brücke existiert, und denkt sich eine andere Erklärung für den Sturz aus - es wurde geschubst, ist ausgerutscht oder ein starker Wind hat es heruntergestoßen.</p>
          `,
      description_en: `
          <p>You craft an illusion that takes root in the mind of a creature that you can see within range. The target must make an Intelligence saving throw. On a failed save, you create a phantasmal object, creature, or other visible phenomenon of your choice that is no larger than a 10-foot cube and that is perceivable only to the target for the duration. This spell has no effect on undead or constructs.</p>
          <p>The phantasm includes sound, temperature, and other stimuli, also evident only to the creature.</p>
          <p>The target can use its action to examine the phantasm with an Intelligence (Investigation) check against your spell save DC. If the check succeeds, the target realizes that the phantasm is an illusion, and the spell ends.</p>
          <p>While the target is affected by the spell, the target treats the phantasm as if it were real. The target rationalizes any illogical outcomes from interacting with the phantasm. For example, a target attempting to walk across an illusory bridge that spans a chasm falls once it steps onto the bridge. If the target survives the fall, it still believes that the bridge exists and comes up with some other explanation for its fall - it was pushed, it slipped, or a strong wind might have knocked it off.</p>
        `,
      school: 'illusion',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2032,
      german: 'Magische Waffe',
      english: 'Magic Weapon',
      image: 'magicWeapon',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard', 'paladin', 'artificer'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine nichtmagische Waffe. Bis der Zauber endet, wird diese Waffe zu einer magischen Waffe mit einem Bonus von +1 auf Angriffs- und Schadenswürfe.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Bonus auf +2. Wenn du einen Zauberplatz des 6. Grades oder höher verwendest, erhöht sich der Bonus auf +3.</p>
        `,
      description_en: `
          <p>You touch a nonmagical weapon. Until the spell ends, that weapon becomes a magic weapon with a +1 bonus to attack rolls and damage rolls.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the bonus increases to +2. When you use a spell slot of 6th level or higher, the bonus increases to +3.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2033,
      german: 'Magischer Mund',
      english: 'Magic Mouth',
      image: 'magicMouth',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 1,
      durationType: 'permanent',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du pflanzt eine Nachricht in einen Gegenstand in Reichweite ein, die ausgesprochen wird, wenn eine Auslösebedingung erfüllt ist. Wähle einen Gegenstand, den du sehen kannst und der nicht von einer anderen Kreatur getragen wird, und sprich dann die Nachricht, die höchstens 25 Wörter lang sein darf, aber bis zu zehn Minuten dauern kann. Bestimme schließlich den Umstand, der den Zauber auslöst, um deine Nachricht zu überbringen.</p>
          <p>Wenn dieser Umstand eintritt, erscheint ein magischer Mund auf dem Gegenstand und trägt die Nachricht mit deiner Stimme in der gleichen Lautstärke vor, wie du sie gesprochen hast. Wenn der von dir gewählte Gegenstand einen Mund hat oder etwas, das wie ein Mund aussieht (zum Beispiel der Mund einer Statue), erscheint der magische Mund dort, so dass die Worte aus dem Mund des Gegenstands zu kommen scheinen. Wenn du diesen Zauber sprichst, kannst du den Zauber beenden, nachdem er seine Botschaft überbracht hat, oder er kann bestehen bleiben und seine Botschaft wiederholen, sobald der Auslöser eintritt.</p>
          <p>Der auslösende Umstand kann so allgemein oder so detailliert sein, wie du willst, er muss jedoch auf visuellen oder hörbaren Bedingungen basieren, die innerhalb von 9 Metern um das Objekt auftreten. Du könntest den Mund zum Beispiel anweisen, zu sprechen, wenn sich eine Kreatur im Umkreis von 9 Metern um das Objekt bewegt oder wenn eine silberne Glocke im Umkreis von 9 Metern läutet.</p>
          `,
      description_en: `
          <p>You implant a message within an object in range, which is spoken when a trigger condition is met. Choose an object you can see and that isn’t being worn or carried by another creature, then speak the message, which must be 25 words or less, but it can last up to 10 minutes. Finally, determine the circumstance that will trigger the spell to deliver your message.</p>
          <p>When that circumstance occurs, a magical mouth appears on the object and recites the message in your voice and at the same volume you spoke. If the chosen object has a mouth or something that looks like a mouth (for example, the mouth of a statue), the magical mouth appears there so that the words appear to come from the object’s mouth. When you cast this spell, you can have the spell end after it delivers its message, or it can remain and repeat its message whenever the trigger occurs.</p>
          <p>The triggering circumstance can be as general or as detailed as you like, though it must be based on visual or audible conditions that occur within 30 feet of the object. For example, you could instruct the mouth to speak when any creature moves within 30 feet of the object or when a silver bell rings within 30 feet of it.</p>
        `,
      school: 'illusion',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2034,
      german: 'Maximilians Erdgriff',
      english: "Maximilian's Earthen Grasp",
      image: 'maximiliansEarthenGrasp',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      doesHeal: false,
      damage: [{ diceNumber: 2, diceType: 6, damageType: 'bludgeoning' }],
      description_de: `
          <p>Du wählst ein nicht besetztes Quadrat mit 1,50m Seitenlänge auf dem Untergrund, das du innerhalb der Reichweite sehen kannst. Eine mittelgroße Hand aus verdichtetem Erdreich erhebt sich dort und greift nach einer Kreatur, die du innerhalb von 1,50m davon sehen kannst. Das Ziel muss einen Stärkerettungswurf durchführen. Bei einem Fehlschlag erleidet das Ziel 2W6 Wuchtschaden und ist bis zum Ende des Zaubers festgesetzt.</p>
          <p>Als eine Aktion kannst du die Hand veranlassen, das gefesselte Ziel zu zermalmen, welches dann einen Stärkerettungswurf ablegen muss. Das Ziel erleidet 2W6 Wuchtschaden bei einem Fehlschlag, die Hälfte davon bei einem Erfolg.</p>
          <p>Um auszubrechen, kann das festgesetzte Ziel seine Aktion verwenden, um einen Wurf auf Stärke gegen deinen Zauberrettungswurf-SG abzulegen. Bei einem Erfolg entkommt das Ziel und ist nicht länger von der Hand festgesetzt.</p>
          <p>Als eine Aktion kannst du die Hand entweder nach einer anderen Kreatur greifen lassen oder sie zu einer anderen unbesetzten Stelle innerhalb der Reichweite bewegen. Die Hand lässt ein festgesetztes Ziel frei, wenn du eines von beidem tust.</p>
          `,
      description_en: `
          <p>You choose a 5-foot-square unoccupied space on the ground that you can see within range. A Medium hand made from compacted soil rises there and reaches for one creature you can see within 5 feet of it. The target must make a Strength saving throw. On a failed save, the target takes 2d6 bludgeoning damage and is restrained for the spell’s duration.</p>
          <p>As an action, you can cause the hand to crush the restrained target, which must make a Strength saving throw. The target takes 2d6 bludgeoning damage on a failed save, or half as much damage on a successful one.</p>
          <p>To break out, the restrained target can use its action to make a Strength check against your spell save DC. On a success, the target escapes and is no longer restrained by the hand.</p>
          <p>As an action, you can cause the hand to reach for a different creature or to move to a different unoccupied space within range. The hand releases a restrained target if you do either.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: true,
      areaOfEffectType: 'square',
      length: 5,
    },
    {
      id: 2035,
      german: 'Melfs Säurepfeil',
      english: "Melf's Acid Arrow",
      image: 'melfsAcidArrow',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      doesHeal: false,
      damage: [{ diceNumber: 4, diceType: 4, damageType: 'acid' }],
      description_de: `
          <p>Ein schimmernder grüner Pfeil schießt auf ein Ziel in Reichweite und explodiert in einem Sprühregen aus Säure. Führe einen Fernkampf-Zauberangriff gegen das Ziel aus. Bei einem Treffer erleidet es sofort 4W4 Säureschaden und 2W4 Säureschaden am Ende seines nächsten Zuges. Verfehlst du, spritzt genug Säure auf das Ziel, um die Hälfte des anfänglichen Schadens zu verursachen, ohne jedoch Schaden am Ende des nächsten Zuges zuzufügen.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der anfängliche Schaden um 1W4 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>A shimmering green arrow streaks toward a target within range and bursts in a spray of acid. Make a ranged spell attack against the target. On a hit, it takes 4d4 acid damage immediately and 2d4 acid damage at the end of its next turn. On a miss, the arrow splashes the target with acid for half as much of the initial damage and no damage at the end of its next turn.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the initial damage increases by 1d4 for each slot level above 2.</p>
          `,
      school: 'evocation',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 2036,
      german: 'Metall erhitzen',
      english: 'Heat Metal',
      image: 'heatMetal',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'bard', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      doesHeal: false,
      damage: [{ diceNumber: 2, diceType: 8, damageType: 'fire' }],
      description_de: `
          <p>Wähle einen hergestellten Metallgegenstand, z.B. eine Metallwaffe oder eine schwere oder mittelschwere Metallrüstung in Reichweite, den du sehen kannst. Du bringst den Gegenstand dazu, rötlich zu glühen. Jede Kreatur, die in physischen Kontakt mit dem Gegenstand kommt, erleidet 2W8 Feuerschaden, wenn du den Zauber sprichst. Bis der Zauber endet, kannst du in jedem deiner folgenden Züge eine Bonusaktion einsetzen, um diesen Schaden erneut zu verursachen.</p>
          <p>Wenn eine Kreatur den Gegenstand hält oder trägt und den Schaden davon erleidet, muss sie einen Rettungswurf auf Konstitution bestehen oder den Gegenstand fallen lassen, wenn sie kann. Wenn sie den Gegenstand nicht fallen lässt, hat sie bis zum Beginn deines nächsten Zuges Nachteil bei Angriffswürfen und Fähigkeitsproben.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>Choose a manufactured metal object, such as a metal weapon or a suit of heavy or medium metal armor, that you can see within range. You cause the object to glow red-hot. Any creature in physical contact with the object takes 2d8 fire damage when you cast the spell. Until the spell ends, you can use a bonus action on each of your subsequent turns to cause this damage again.</p>
          <p>If a creature is holding or wearing the object and takes the damage from it, the creature must succeed on a Constitution saving throw or drop the object if it can. If it doesn’t drop the object, it has disadvantage on attack rolls and ability checks until the start of your next turn.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for each slot level above 2.</p>
          `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2037,
      german: 'Mondstrahl',
      english: 'Moonbeam',
      image: 'moonbeam',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      doesHeal: false,
      damage: [{ diceNumber: 2, diceType: 10, damageType: 'radiant' }],
      description_de: `
          <p>Ein silbriger Strahl aus blassem Licht scheint in einem Zylinder mit einem Radius von 1,5 Metern und einer Höhe von 12 Metern auf einen Punkt in Reichweite. Bis zum Ende des Zaubers füllt ein schwaches Licht den Zylinder.</p>
          <p>Wenn eine Kreatur den Bereich des Zaubers zum ersten Mal in ihrem Zug betritt oder ihren Zug dort beginnt, wird sie von geisterhaften Flammen verschlungen, die brennende Schmerzen verursachen, und sie muss einen Rettungswurf auf Konstitution machen. Bei einem misslungenen Rettungswurf erleidet sie 2W10 gleißender Schaden, bei einem erfolgreichen Rettungswurf halb so viel.</p>
          <p>Ein Gestaltwandler macht seinen Rettungswurf mit Nachteil. Wenn er scheitert, kehrt er sofort in seine ursprüngliche Form zurück und kann erst wieder eine andere Form annehmen, wenn er das Licht des Zaubers verlässt.</p>
          <p>In jedem deiner Züge, nachdem du diesen Zauber gewirkt hast, kannst du eine Aktion nutzen, um den Strahl 18 Meter in eine beliebige Richtung zu bewegen.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W10 für jeden Zauberplatz-Grad über dem 2.</p>
          `,
      description_en: `
          <p>A silvery beam of pale light shines down in a 5-foot radius, 40-foot-high cylinder centered on a point within range. Until the spell ends, dim light fills the cylinder.</p>
          <p>When a creature enters the spell’s area for the first time on a turn or starts its turn there, it is engulfed in ghostly flames that cause searing pain, and it must make a Constitution saving throw. It takes 2d10 radiant damage on a failed save, or half as much damage on a successful one.</p>
          <p>A shapechanger makes its saving throw with disadvantage. If it fails, it also instantly reverts to its original form and can’t assume a different form until it leaves the spell’s light.</p>
          <p>On each of your turns after you cast this spell, you can use an action to move the beam 60 feet in any direction.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d10 for each slot level above 2.</p>
          `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cylinder',
      length: 5,
    },
    {
      id: 2038,
      german: 'Nebelschritt',
      english: 'Misty Step',
      image: 'mistyStep',
      level: 2,
      isCustom: false,
      classes: ['test', 'warlock', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du bist kurz von silbrigem Nebel umgeben und teleportierst dich bis zu 9 Meter weit an einen freien Platz, den du sehen kannst.</p>
        `,
      description_en: `
          <p>Briefly surrounded by silvery mist, you teleport up to 30 feet to an unoccupied space that you can see.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2039,
      german: 'Nystuls magische Aura',
      english: "Nystul's Magic Aura",
      image: 'nystulsMagicAura',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 24,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur oder einen Gegenstand und legst eine Illusion auf das Ziel, sodass Erkenntniszauber falsche Informationen über es vermitteln. Das Ziel kann eine bereitwillige Kreatur oder ein Gegenstand sein, der nicht von einer anderen Kreatur getragen oder in der Hand gehalten wird.</p>
          <p>Wähle beim wirken des Zaubers einen oder beide der Folgenden effekte, die für die Wirkungsdauer anhalten. Wenn du diesen Zauber für 30 Tage jeden Tag auf dieselbe Kreatur oder denselben Gegenstand wirkst und jedes mal den gleichen Effekt nutzt, hält die Illusion an, bis sie gebannt wird.</p>
          <p>Falsche Aura: Du veränderst die Art, wie das Ziel für Zauber und magische Effekte erscheint, die magische Auren aufspüren (etwa "Magie entdecken] Stufe 1"). Du kannst einen nicht-magischen Gegenstand als einen magischen erscheinen lassen, einen magischen als nicht-magischen oder die Aura eines magischen Gegenstandsändern, sodass er einer Magieschule deiner Wahl anzugehören scheint. Wenn du diesen Effekt auf einen Gegenstand wirkst, kannst du entscheiden, dass die Illusion von jeder Kreatur erkannt wird, die den Gegenstand in die Hand nimmt.</p>
          <p>Markierung: Du beeinflusst, wie das Ziel von Zaubern und magischen Effekten wahrgenommen wird, die Kreaturentypen aufspüren (etwa das "Göttliche Gespür" des Paladins oder der auslöser des Zaubers "Symbol" Stufe 7). Bestimme einen Kreaturentyp, und Zauber und andere magische Effekte behandeln das Ziel, als währe es eine entsprechende Kreatur oder von dieser Gesinnung.</p>
          `,
      description_en: `
          <p>You touch a creature or an object and place an illusion on it that lasts for the duration. You can make the false aura appear to belong to a creature of a different creature type that you choose. The change will be detected by spells or other abilities that detect magical auras.</p>
          <p>If the object is a magic item or some other magic-imbued object, you learn its properties and how to use them, whether it requires attunement to use, and how many charges it has, if any. You learn whether any spells are affecting the item and what they are. If the item was created by a spell, you learn which spell created it.</p>
          `,
      school: 'illusion',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2040,
      german: 'Person festhalten',
      english: 'Hold Person',
      image: 'holdPerson',
      level: 2,
      isCustom: false,
      classes: [
        'test',
        'bard',
        'druid',
        'warlock',
        'cleric',
        'wizard',
        'sorcerer',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wähle einen Humanoiden in Reichweite, den du sehen kannst. Das Ziel muss einen Rettungswurf auf Weisheit machen, um nicht für die Wirkungsdauer gelähmt zu werden. Zu Beginn eines jeden seiner Züge, kann das Ziel einen weiteren Rettungswurf auf Weisheit machen. Bei einem Erfolg endet der Zauber für das Ziel.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, dann kannst du den Zauber für jeden Zauberplatz-Grad über dem zweiten auf eine zusätzliche humanoide Kreatur wirken. Die Kreaturen dürfen nicht weiter als 9 Meter voneinander entfernt sein, wenn du den Zauber auf sie wirkst.</p>
        `,
      description_en: `
          <p>Choose a humanoid that you can see within range. The target must succeed on a Wisdom saving throw or be paralyzed for the duration. At the end of each of its turns, the target can make another Wisdom saving throw. On a success, the spell ends on the target.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, you can target one additional humanoid for each slot level above 2nd. The humanoids must be within 30 feet of each other when you target them.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2041,
      german: 'Pfeilspalier',
      english: 'Cordon of Arrows',
      image: 'cordonOfArrows',
      level: 2,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du pflanzt vier Stück nichtmagische Munition - Pfeile oder Armbrustbolzen - in den Boden in Reichweite und legst Magie darauf, um ein Gebiet zu schützen. Bis zum Ende des Zaubers fliegt jedes Mal, wenn eine andere Kreatur als du zum ersten Mal in einem Zug innerhalb von 9 Metern in die Nähe der Munition kommt oder ihren Zug dort beendet, ein Stück Munition hoch und trifft sie. Die Kreatur muss einen Rettungswurf auf Geschicklichkeit bestehen oder erleidet 1W6 Stichschaden. Das Munitionsstück wird dann zerstört. Der Zauber endet, wenn keine Munition mehr vorhanden ist.</p>
          <p>Wenn du diesen Zauber sprichst, kannst du beliebige Kreaturen bestimmen, die der Zauber ignoriert.</p>
          <p><b>Auf höheren Stufen:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Zauberplatz-Grad über dem 2.</p>
        `,
      description_en: `
          <p>You plant four pieces of nonmagical ammunition - arrows or crossbow bolts - in the ground within range and lay magic upon them to protect an area. Until the spell ends, whenever a creature other than you comes within 30 feet of the ammunition for the first time on a turn or ends its turn there, one piece of ammunition flies up to strike it. The creature must succeed on a Dexterity saving throw or take 1d6 piercing damage. The piece of ammunition is then destroyed. The spell ends when no ammunition remains.</p>
          <p>When you cast this spell, you can designate any creatures you choose to be unaffected by it.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2042,
      german: 'Reittier finden',
      english: 'Find Steed',
      image: 'findSteed',
      level: 2,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 10,
      cost: 'minutes',
      duration: 0,
      durationType: 'permanent',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du beschwörst einen Geist, der die Form eines ungewöhnlich intelligenten, starken und treuen Reittiers annimmt und eine dauerhafte Bindung mit ihm eingeht. Das Reittier erscheint an einem unbesetzten Ort in Reichweite und nimmt die Form eines Pferdes deiner Wahl an, z. B. eines Haudegens, eines Ponys, eines Kamels, eines Elchs oder eines Mastiffs. (Dein Spielleiter kann auch erlauben, andere Tiere als Reittiere zu beschwören.) Das Reittier hat die gleichen Eigenschaften wie die gewählte Form, ist aber ein Himmels-, Feen- oder Ungeheuerwesen (nach deiner Wahl) statt seines normalen Typs. Wenn dein Reittier eine Intelligenz von 5 oder weniger hat, wird seine Intelligenz auf 6 erhöht und es erhält die Fähigkeit, eine Sprache deiner Wahl zu verstehen, die du sprichst.</p>
          <p>Dein Reittier dient dir, sowohl im Kampf als auch in der Freizeit, und du hast eine instinktive Verbindung zu ihm, die es dir ermöglicht, als nahtlose Einheit zu kämpfen. Während du auf deinem Reittier sitzt, kannst du jeden Zauber, der nur auf dich zielt, auch auf dein Reittier wirken.</p>
          <p>Wenn das Reittier auf 0 Trefferpunkte sinkt, verschwindet es und hinterlässt keine physische Form. Du kannst dein Reittier auch jederzeit mit einer Aktion entlassen, die es verschwinden lässt. In beiden Fällen beschwörst du mit diesem Zauber dasselbe Reittier erneut, wobei es seine maximale Anzahl an Trefferpunkten wiedererlangt.</p>
          <p>Solange sich dein Reittier in einem Umkreis von 1,6 Kilometern befindet, kannst du telepathisch mit ihm kommunizieren.</p>
          <p>Du kannst nicht mehr als ein durch diesen Zauber gebundenes Reittier gleichzeitig haben. Als Aktion kannst du das Reittier jederzeit aus der Bindung lösen, so dass es verschwindet.</p>
        `,
      description_en: `
          <p>You summon a spirit that assumes the form of an unusually intelligent, strong, and loyal steed, creating a long-lasting bond with it. Appearing in an unoccupied space within range, the steed takes on a form that you choose: a warhorse, a pony, a camel, an elk, or a mastiff. (Your DM might allow other animals to be summoned as steeds.) The steed has the statistics of the chosen form, though it is a celestial, fey, or fiend (your choice) instead of its normal type. Additionally, if your steed has an Intelligence of 5 or less, its Intelligence becomes 6, and it gains the ability to understand one language of your choice that you speak.</p>
          <p>Your steed serves you as a mount, both in combat and out, and you have an instinctive bond with it that allows you to fight as a seamless unit. While mounted on your steed, you can make any spell you cast that targets only you also target your steed.</p>
          <p>When the steed drops to 0 hit points, it disappears, leaving behind no physical form. You can also dismiss your steed at any time as an action, causing it to disappear. In either case, casting this spell again summons the same steed, restored to its hit point maximum.</p>
          <p>While your steed is within 1 mile of you, you can communicate with each other telepathically. You can’t have more than one steed bonded by this spell at a time. As an action, you can release the steed from its bond at any time, causing it to disappear.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2043,
      german: 'Rindenhaut',
      english: 'Barkskin',
      image: 'barkskin',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine bereitwillige Kreatur. Bis der Zauber endet, hat die Haut des Ziels ein raues, rindenartiges Aussehen und die RK des Ziels kann nicht kleiner als 16 sein, unabhängig davon, welche Art von Rüstung es trägt.</p>
        `,
      description_en: `
          <p>You touch a willing creature. Until the spell ends, the target’s skin has a rough, bark-like appearance, and the target’s AC can’t be less than 16, regardless of what kind of armor it is wearing.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2044,
      german: 'Sanfte Ruhe',
      english: 'Gentle Repose',
      image: 'gentleRepose',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'days',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst einen Leichnam oder andere Überreste. Für die Dauer des Zaubers ist das Ziel vor Verwesung geschützt und kann nicht untot werden.</p>
          <p>Der Zauber verlängert auch effektiv die Frist für die Auferweckung des Ziels von den Toten, da Tage, die unter dem Einfluss dieses Zaubers verbracht werden, nicht auf die Frist von Zaubern wie "Tote erwecken (Stufe 5)" angerechnet werden.</p>
        `,
      description_en: `
          <p>You touch a corpse or other remains. For the duration, the target is protected from decay and can’t become undead.</p>
          <p>The spell also effectively extends the time limit on raising the target from the dead, since days spent under the influence of this spell don’t count against the time limit of spells such as "Raise Dead (Level 5)." </p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2045,
      german: 'Schattenklinge',
      english: 'Shadow Blade',
      image: 'shadowBlade',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wenn du diesen Zauber wirkst, erschaffst du eine halbphysische, halb-schattige Klinge in deiner freien Hand. Diese Klinge ist ähnlich wie eine normale Klinge und verursacht 2W8 psychischen Schaden bei einem Treffer. Die Klinge ist eine einfache Nahkampfwaffe die du als Finesse-Waffe verwenden kannst. Sie verursacht psychischen Schaden und hat die Eigeschaften Leicht, FInesse und Wurf(20/60)</p>
          <p>Wenn du die WWaffe fallen lässt oder wirfst, verschwindet sie am Ende deines Zuges. Solange der Zauber aktiv ist, kannst du das Schwert mit einer Bonusaktion in deine Hand zurückrufen.</p>
          <p><b>Auf höheren Graden:<b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder 4. Grad erhöht sich der Schaden auf 3W8. Bei einem Zauberplatz der Stufe 5 und 6 auf 4W8 und bei 7 oder höher auf 5W8</p>
        `,
      description_en: `
          <p>When you cast this spell, you create a shadowy blade in your free hand. This blade is similar to a normal blade and deals 2d8 psychic damage on a hit. The blade is a simple melee weapon that you can use as a finesse weapon. It deals psychic damage and has the properties light, finesse, and thrown (20/60).</p>
          <p>If you drop the weapon or throw it, it disappears at the end of your turn. While the spell is active, you can recall the sword to your hand with a bonus action.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or 4th level, the damage increases to 3d8. For a 5th- or 6th-level spell slot, the damage increases to 4d8. For a 7th-level spell slot or higher, the damage increases to 5d8.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2046,
      german: 'Schutz vor Gift',
      english: 'Protection from Poison',
      image: 'protectionFromPoison',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'cleric', 'paladin', 'ranger', 'artificer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur. Wenn sie vergiftet ist, neutralisierst du das Gift. Wenn das Ziel von mehr als einem Gift befallen ist, neutralisierst du ein Gift, von dem du weißt, dass es vorhanden ist, oder du neutralisierst ein zufälliges. Während der Dauer des Zaubers hat das Ziel Vorteil bei Rettungswürfen gegen Vergiftungen und ist resistent gegen Giftschaden.</p>
        `,
      description_en: `
          <p>You touch a creature. If it is poisoned, you neutralize the poison. If more than one poison afflicts the target, you neutralize one poison that you know is present, or you neutralize one at random.</p>
          <p>For the duration, the target has advantage on saving throws against being poisoned, and it has resistance to poison damage.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2047,
      german: 'Schutzwind',
      english: 'Warding Wind',
      image: 'wardingWind',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein starker Wind (32 Kilometer pro Stunde) weht in einem Radius von 3 Metern um dich herum, bewegt sich mit dir und bleibt auf dich zentriert. Der Wind hält für die Dauer des Zaubers an.</p>
          <ul>
          <li>Er macht dich und andere Kreaturen in seinem Bereich taub.</li>
          <li>Er löscht ungeschützte Flammen von der Größe einer Fackel oder kleiner in seinem Bereich.</li>
          <li>Das Gebiet ist schwieriges Terrain für andere Kreaturen als dich.</li>
          <li>Die Angriffswürfe von Angriffen mit Fernkampfwaffen haben Nachteil, wenn sie durch den Wind hindurchgehen.</li>
          <li>Sie vertreibt Dampf, Gas und Nebel, die durch starken Wind zerstreut werden können.</li>
          </ul>
          `,
      description_en: `
          <p>A strong wind (20 miles per hour) blows around you in a 10-foot radius and moves with you, remaining centered on you. The wind lasts for the duration.</p>
          <ul>
          <li>It deafens you and other creatures in its area.</li>
          <li>It extinguishes unprotected flames in its area that are torch-sized or smaller.</li>
          <li>It hedges out vapor, gas, and fog that can be dispersed by strong wind.</li>
          <li>The area is difficult terrain for creatures other than you.</li>
          <li>The attack rolls of ranged weapon attacks have disadvantage if they pass in or out of the wind.</li>
          </ul>
          `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 2048,
      german: 'Schwache Genesung',
      english: 'Lesser Restoration',
      image: 'lesserRestoration',
      level: 2,
      isCustom: false,
      classes: [
        'test',
        'bard',
        'druid',
        'cleric',
        'paladin',
        'ranger',
        'artificer',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      savingThrowAttribute: '',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur und kannst entweder eine Krankheit beenden, an der sie leidet, oder einen Zustand, den sie erlitten hat. Dieser Zustand kann blind, gelähmt, taub oder vergiftet sein.</p>
        `,
      description_en: `
          <p>You touch a creature and can end either one disease or one condition afflicting it. The condition can be blinded, deafened, paralyzed, or poisoned.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2049,
      german: 'Schweben',
      english: 'Levitate',
      image: 'levitate',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine Kreatur oder ein Gegenstand deiner Wahl in Reichweite, den du sehen kannst, steigt senkrecht bis zu 6 Meter in die Höhe und bleibt dort für die Dauer des Zaubers hängen. Der Zauber kann ein Ziel schweben lassen, das bis zu 225 Kilogramm wiegt. Eine unwillige Kreatur, die einen Rettungswurf auf Konstitution besteht, bleibt unberührt.</p>
          <p>Das Ziel kann sich nur bewegen, indem es gegen ein festes Objekt oder eine Oberfläche in Reichweite drückt oder zieht (z.B. eine Wand oder eine Decke), wodurch es sich so bewegen kann, als würde es klettern. Du kannst die Höhe des Ziels in deinem Zug um bis zu 6 Meter in jede Richtung ändern. Wenn du das Ziel bist, kannst du dich als Teil deiner Bewegung nach oben oder unten bewegen. Andernfalls kannst du deine Aktion nutzen, um das Ziel zu bewegen, das innerhalb der Reichweite des Zaubers bleiben muss.</p>
          <p>Wenn der Zauber endet, schwebt das Ziel sanft zu Boden, falls es sich noch in der Luft befindet.</p>
        `,
      description_en: `
          <p>One creature or object of your choice that you can see within range rises vertically, up to 20 feet, and remains suspended there for the duration. The spell can levitate a target that weighs up to 500 pounds. An unwilling creature that succeeds on a Constitution saving throw is unaffected.</p>
          <p>The target can move only by pushing or pulling against a fixed object or surface within reach (such as a wall or a ceiling), which allows it to move as if it were climbing. You can change the target’s altitude in your turn by up to 20 feet in either direction. If you are the target, you can move up or down as part of your move. Otherwise, you can use your action to move the target, which must remain within the spell’s range.</p>
          <p>When the spell ends, the target floats gently to the ground if it is still aloft.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2050,
      german: 'Schwächestrahl',
      english: 'Ray of Enfeeblement',
      image: 'rayOfEnfeeblement',
      level: 2,
      isCustom: false,
      classes: ['test', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Ein schwarzer Strahl aus entnervender Energie entspringt aus deinem Finger auf eine Kreatur in Reichweite. Führe einen Fernkampfzauberangriff gegen das Ziel aus. Bei einem Treffer verursacht das Ziel mit Waffenangriffen, die Stärke verwenden, nur halben Schaden, bis der Zauber endet.</p>
          <p>Am Ende jeder Runde des Ziels kann es einen Rettungswurf gegen den Zauber machen. Bei einem Erfolg endet der Zauber.</p>
        `,
      description_en: `
          <p>A black beam of enervating energy springs from your finger toward a creature within range. Make a ranged spell attack against the target. On a hit, the target deals only half damage with weapon attacks that use Strength until the spell ends.</p>
          <p>At the end of each of the target’s turns, it can make a Constitution saving throw against the spell. On a success, the spell ends.</p>
        `,
      school: 'necromancy',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2051,
      german: 'Schützenden Band',
      english: 'Warding Bond',
      image: 'wardingBond',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber schützt eine bereitwillige Kreatur, die du berührst und erschafft eine magische Verbindung zwischen dir und dem Ziel bis der Zauber endet. Solange sich das Ziel innerhalb von 18 Metern um dich befindet, erhält es einen Bonus von + 1 auf RK und Rettungswürfe und hat Resistenz gegen alle Schadensarten. Außerdem erleidest du jedes Mal, wenn das Ziel Schaden erleidet die gleiche Menge an Schaden. Der Zauber endet, wenn du auf 0 Trefferpunkte fällst oder wenn du und das Ziel weiter als 18 Meter auseinander sind.</p>
          <p>Er endet auch, wenn der Zauber erneut auf einen der beiden Beteiligten gewirkt wird. Du kannst den Zauber als Aktion aufheben.</p>
        `,
      description_en: `
          <p>This spell wards a willing creature you touch and creates a mystic connection between you and the target until the spell ends. While the target is within 60 feet of you, it gains a +1 bonus to AC and saving throws, and it has resistance to all damage. Also, each time it takes damage, you take the same amount of damage. The spell ends if you drop to 0 hit points or if you and the target become separated by more than 60 feet. It also ends if the spell is cast again on either of the connected creatures. You can also dismiss the spell as an action.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2052,
      german: 'Seiltrick',
      english: 'Rope Trick',
      image: 'ropeTrick',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst ein Seil, das bis zu 18 Meter lang ist. Ein Ende des Seils erhebt sich dann in die Luft, bis das ganze Seil senkrecht zum Boden hängt. Am oberen Ende des Seils öffnet sich ein unsichtbarer Eingang zu einem extradimensionalen Raum, der bis zum Ende des Zaubers bestehen bleibt.</p>
          <p>Der extradimensionale Raum kann erreicht werden, indem du auf die Spitze des Seils kletterst. Der Raum kann bis zu acht mittlere oder kleinere Kreaturen aufnehmen. Das Seil kann in den Raum hineingezogen werden, so dass es außerhalb des Raums nicht mehr zu sehen ist.</p>
          <p>Angriffe und Zauber können den Eingang des extradimensionalen Raums nicht durchqueren, aber diejenigen, die sich darin befinden, können wie durch ein 1 x 1,5 Meter großes Fenster in der Mitte des Seils hinaussehen.</p>
          <p>Alles, was sich im extradimensionalen Raum befindet, fällt heraus, wenn der Zauber endet.</p>
        `,
      description_en: `
          <p>You touch a length of rope that is up to 60 feet long. One end of the rope then rises into the air until the whole rope hangs perpendicular to the ground. At the upper end of the rope, an invisible entrance opens to an extradimensional space that lasts until the spell ends.</p>
          <p>The extradimensional space can be reached by climbing to the top of the rope. The space can hold as many as eight Medium or smaller creatures. The rope can be pulled into the space, making the rope disappear from view outside the space.</p>
          <p>Attacks and spells can’t cross through the entrance into or out of the extradimensional space, but those inside can see out of it as if through a 3-foot-by-5-foot window centered on the rope.</p>
          <p>Anything inside the extradimensional space drops out when the spell ends.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2053,
      german: 'Sengender Strahl',
      english: 'Scorching Ray',
      image: 'scorchingRay',
      level: 2,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 6, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Du erschaffst drei Strahlen aus Feuer und schießt sie auf Kreaturen in Reichweite. Du kannst sie auf ein Ziel oder mehrere Ziele aufteilen, solange du genug Strahlen hast. Führe für jeden Strahl einen Fernkampfangriff durch. Bei einem Treffer erleidet das Ziel 2W6 Feuerschaden.</p>
          <p><b>Auf höheren Graden:</b> Für jeden Zauberplatz über dem 2. erhältst du einen zusätzlichen Strahl.</p>
        `,
      description_en: `
          <p>You create three rays of fire and hurl them at targets within range. You can hurl them at one target or several. Make a ranged spell attack for each ray. On a hit, the target takes 2d6 fire damage.</p>
          <p><b>At higher levels:</b> For each spell slot above 2nd, you get one additional ray.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 2054,
      german: 'Snillocs Schneeballschwarm',
      english: "Snilloc's Snowball Swarm",
      image: 'snillocsSnowballSwarm',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 6, damageType: 'cold' }],
      doesHeal: false,
      description_de: `
          <p>Ein Hagel von Schneebällen fällt auf Kreaturen in einem 5 Fuß großen Radius um einen Punkt innerhalb der Reichweite. Jede Kreatur in diesem Bereich muss einen Rettungswurf auf Geschicklichkeit ablegen. Ein Ziel erleidet 3W6 Kälteschaden bei einem misslungenen Rettungswurf oder die Hälfte bei einem erfolgreichen.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden zweiten Grad über dem 2.</p>
        `,
      description_en: `
          <p>A flurry of snowballs pummels creatures in a 5-foot-radius sphere centered on a point within range. Each creature in that area must make a Dexterity saving throw. A creature takes 3d6 cold damage on a failed save, or half as much damage on a successful one.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2nd.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 2055,
      german: 'Spiegelbilder',
      english: 'Mirror Image',
      image: 'mirrorImage',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard', 'sorcerer', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Drei illusorische Duplikate von dir erscheinen in deinem Raum. Bis zum Ende des Zaubers bewegen sich die Duplikate mit dir und ahmen deine Handlungen nach, wobei sie ihre Position so verändern, dass es unmöglich ist, zu erkennen, welches Bild echt ist. Du kannst deine Aktion nutzen, um die illusorischen Duplikate abzuschalten.</p>
          <p>Würfe jedes Mal, wenn eine Kreatur während der Dauer des Zaubers einen Angriff auf dich richtet, einen W20, um zu bestimmen, ob der Angriff stattdessen auf eines deiner Duplikate zielt.</p>
          <p>Wenn du drei Duplikate hast, musst du eine 6 oder höher würfeln, um das Ziel des Angriffs auf ein Duplikat zu ändern. Bei zwei Duplikaten musst du eine 8 oder höher würfeln. Bei einem Duplikat musst du eine 11 oder höher würfeln.</p>
          <p>Die RK eines Duplikats ist gleich 10 + dein Geschicklichkeitsmodifikator. Wenn ein Angriff ein Duplikat trifft, wird das Duplikat zerstört. Ein Duplikat kann nur durch einen Angriff zerstört werden, der es trifft. Alle anderen Schäden und Effekte werden ignoriert. Der Zauber endet, wenn alle drei Duplikate zerstört sind.</p>
          <p>Eine Kreatur ist von diesem Zauber nicht betroffen, wenn sie nicht sehen kann, wenn sie sich auf andere Sinne als das Sehen verlässt, wie z.B. Blindsicht, oder wenn sie Illusionen als falsch wahrnehmen kann, wie bei "Wahrer Blick".</p>
        `,
      description_en: `
          <p>Three illusory duplicates of yourself appear in your space. Until the spell ends, the duplicates move with you and mimic your actions, shifting position so it’s impossible to track which image is real. You can use your action to dismiss the illusory duplicates.</p>
          <p>Each time a creature targets you with an attack during the spell’s duration, roll a d20 to determine whether the attack instead targets one of your duplicates.</p>
          <p>If you have three duplicates, you must roll a 6 or higher to change the attack’s target to a duplicate. With two duplicates, you must roll an 8 or higher. With one duplicate, you must roll an 11 or higher.</p>
          <p>A duplicate’s AC equals 10 + your Dexterity modifier. If an attack hits a duplicate, the duplicate is destroyed. A duplicate can be destroyed only by an attack that hits it. It ignores all other damage and effects. The spell ends when all three duplicates are destroyed.</p>
          <p>A creature is unaffected by this spell if it can’t see, if it relies on senses other than sight, such as blindsight, or if it can perceive illusions as false, as with true sight.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2056,
      german: 'Spinnenklettern',
      english: 'Spider Climb',
      image: 'spiderClimb',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard', 'sorcerer', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Bis zum Ende des Zaubers erhält eine bereitwillige Kreatur, die du berührst, die Fähigkeit, sich aufwärts, abwärts, über vertikale Flächen und kopfüber an der Decke entlang zu bewegen, wobei sie die Hände frei hat. Das Ziel erhält außerdem eine Klettergeschwindigkeit in Höhe seiner Gehgeschwindigkeit.</p>
        `,
      description_en: `
          <p>Until the spell ends, one willing creature you touch gains the ability to move up, down, and across vertical surfaces and upside down along ceilings, while leaving its hands free. The target also gains a climbing speed equal to its walking speed.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2057,
      german: 'Spinnennetz',
      english: 'Web',
      image: 'web',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du beschwörst eine Masse aus dicken, klebrigen Spinnweben an einem Punkt deiner Wahl in Reichweite. Die Spinnweben füllen für die Dauer des Zaubers einen 6 Meter großen Würfel von diesem Punkt aus. Die Spinnweben sind schwieriges Terrain und verdunkeln ihren Bereich leicht.</p>
          <p>Wenn die Spinnweben nicht zwischen zwei festen Massen (wie Mauern oder Bäumen) verankert oder über einen Boden, eine Wand oder eine Decke gelegt sind, fällt das beschworene Netz in sich zusammen und der Zauber endet zu Beginn deines nächsten Zuges. Netze, die über eine flache Oberfläche gelegt werden, haben eine Tiefe von 1,5 Metern.</p>
          <p>Jede Kreatur, die ihren Zug in den Netzen beginnt oder sie während ihres Zuges betritt, muss einen Rettungswurf auf Geschicklichkeit machen.Bei einem misslungenen Rettungswurf ist die Kreatur so lange gefesselt, wie sie sich in den Spinnweben befindet oder bis sie ausbricht.</p>
          <p>Eine Kreatur, die von den Spinnweben gefesselt ist, kann ihre Aktion nutzen, um eine Stärkeprobe gegen deinen Zauberrettungswurf SG zu machen. Gelingt ihr das, ist sie nicht mehr gefesselt.</p>
          <p>Die Spinnweben sind entflammbar. Jeder 1,5 mal 1,5 Meter große Würfel aus Spinnweben, der Feuer ausgesetzt ist, brennt in einer Runde ab und fügt jeder Kreatur, die ihren Zug im Feuer beginnt, 2W4 Feuerschaden zu.</p>
        `,
      description_en: `
          <p>You conjure a mass of thick, sticky webbing at a point of your choice within range. The webs fill a 20-foot cube from that point for the duration. The webs are difficult terrain and lightly obscure their area.</p>
          <p>If the webs aren’t anchored between two solid masses (such as walls or trees) or layered across a floor, wall, or ceiling, the conjured web collapses on itself, and the spell ends at the start of your next turn. Webs layered over a flat surface have a depth of 5 feet.</p>
          <p>Each creature that starts its turn in the webs or that enters them during its turn must make a Dexterity saving throw. On a failed save, the creature is restrained as long as it remains in the webs or until it breaks free.</p>
          <p>A creature restrained by the webs can use its action to make a Strength check against your spell save DC. If it succeeds, it is no longer restrained.</p>
          <p>The webs are flammable. Any 5-foot cube of webs exposed to fire burns away in 1 round, dealing 2d4 fire damage to any creature that starts its turn in the fire.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 2058,
      german: 'Spurloses Gehen',
      english: 'Pass without Trace',
      image: 'passWithoutTrace',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Von dir geht ein Schleier aus Schatten und Stille aus, der dich und deine Gefährten vor Entdeckung schützt. Für die Dauer des Schleiers erhält jede Kreatur deiner Wahl im Umkreis von 9 Metern um dich (dich eingeschlossen) einen Bonus von +10 auf Geschicklichkeitsproben (Heimlichkeit) und kann nur durch magische Mittel aufgespürt werden. Eine Kreatur, die diesen Bonus erhält, hinterlässt keine Spuren oder andere Hinweise auf ihren Weg.</p>
        `,
      description_en: `
          <p>A veil of shadows and silence radiates from you, masking you and your companions from detection. For the duration, each creature you choose within 30 feet of you (including you) has a +10 bonus to Dexterity (Stealth) checks and can’t be tracked except by magical means. A creature that receives this bonus leaves behind no tracks or other traces of its passage.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 2059,
      german: 'Staubteufel',
      english: 'Dust Devil',
      image: 'dustDevil',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 8, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
          <p>Wähle einen unbesetzten Würfel mit einer Kantenlänge von 1,5 Metern aus Luft in Reichweite, den du sehen kannst. Eine elementare Kraft, die einem Staubteufel ähnelt, erscheint in dem Würfel und hält für die Dauer des Zaubers an.</p>
          <p>Jede Kreatur, die ihren Zug innerhalb von 1,5 Metern um den Staubteufel beendet, muss einen Rettungswurf auf Stärke machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 1W8 Wuchtschaden und wird 3 Meter weit weggestoßen. Bei einem erfolgreichen Rettungswurf erleidet die Kreatur halb so viel Schaden und wird nicht weggestoßen.</p>
          <p>Als Bonusaktion kannst du den Staubteufel bis zu 9 Meter in eine beliebige Richtung bewegen. Bewegt sich der Staubteufel über Sand, Staub, losen Schmutz oder kleinen Kies, saugt er das Material auf und bildet um sich herum eine Trümmerwolke mit einem Radius von 3 Metern, die bis zum Beginn deines nächsten Zuges anhält. Die Wolke verdunkelt ihr Gebiet stark.</p>
          <p><b>Auf höheren Graden:</b> Der Schaden erhöht sich um 1W8 für jeden zweiten Grad über dem 2.</p>
        `,
      description_en: `
          <p>Choose an unoccupied 5-foot cube of air that you can see within range. An elemental force that resembles a dust devil appears in the cube and lasts for the spell’s duration.</p>
          <p>Any creature that ends its turn within 5 feet of the dust devil must make a Strength saving throw. On a failed save, the creature takes 1d8 bludgeoning damage and is pushed 10 feet away. On a successful save, the creature takes half as much damage and isn’t pushed.</p>
          <p>As a bonus action, you can move the dust devil up to 30 feet in any direction. If the dust devil moves over sand, dust, loose dirt, or small gravel, it sucks up the material and forms a 10-foot-radius cloud of debris around itself that lasts until the start of your next turn. The cloud heavily obscures its area.</p>
          <p><b>At higher levels:</b> The damage increases by 1d8 for each slot level above 2nd.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 5,
    },
    {
      id: 2060,
      german: 'Stille',
      english: 'Silence',
      image: 'silence',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Für die Dauer des Zaubers kann innerhalb einer Sphäre mit einem Radius von 6 Metern, die auf einen von dir gewählten Punkt innerhalb der Reichweite zentriert ist, kein Geräusch erzeugt werden oder diese passieren. Jede Kreatur oder jedes Objekt, das sich vollständig innerhalb der Sphäre befindet, ist immun gegen Donnerschaden, und Kreaturen sind taub, solange sie sich vollständig innerhalb der Sphäre befinden.</p>
        `,
      description_en: `
          <p>For the duration, no sound can be created within or pass through a 20-foot-radius sphere centered on a point you choose within range. Any creature or object entirely inside the sphere is immune to thunder damage, and creatures are deafened while entirely inside it.</p>
        `,
      school: 'illusion',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 2061,
      german: 'Tashas Gedankenpeitsche',
      english: "Tasha's Mind Whip",
      image: 'tashasMindWhip',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'intelligence',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 6, damageType: 'psychic' }],
      doesHeal: false,
      description_de: `
          <p>Du peitschst eine andere Kreatur psychisch, die du in Reichweite sehen kannst. Das Ziel muss einen Rettungswurf auf Intelligenz ausführen. Bei Misserfolg nimmt es 3W6 psychischen Schaden und kann bis zum Ende seines nächsten Zuges keine Reaktion ausführen. Außerdem kann es in seinem nächsten Zug nur entweder eine Bewegung, eine Aktion oder eine Bonusaktion ausführen. Bei einem erfolgreichen Rettungswurf nimmt das Ziel die Hälfte des erwürfelten Schadens und erleidet keinen der anderen Effekte.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden zweiten Grad über dem 2.</p>
        `,
      description_en: `
          <p>You psychically lash out at one creature you can see within range. The target must succeed on an Intelligence saving throw or take 3d6 psychic damage and can’t take a reaction until the end of its next turn. Moreover, on its next turn, it must choose whether it gets a move, an action, or a bonus action; it gets only one of the three. On a successful save, the target takes half as much damage and doesn’t have to choose as described above.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d6 for each slot level above 2nd.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 2062,
      german: 'Tiere beschwören',
      english: 'Conjure Animals',
      image: 'conjureAnimals',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du beschwörst Feengeister herauf, die die Form von Tieren annehmen und an freien Stellen in Reichweite, die du sehen kannst, erscheinen. Wähle eine der folgenden Optionen:</p>
          <ul>
            <li>Ein Tier von Herausforderungsstufe 2 oder niedriger</li>
            <li>Zwei Tiere von Herausforderungsstufe 1 oder niedriger</li>
            <li>Vier Tiere von Herausforderungsstufe 1/2 oder niedriger</li>
            <li>Acht Tiere von Herausforderungsstufe 1/4 oder niedriger</li>
          </ul>
          <p>Jedes Tier wird auch als Feengeist betrachtet und verschwindet, wenn es 0 Trefferpunkte erreicht oder wenn der Zauber endet.</p>
          <p>Die beschworenen Tiere sind freundlich gegenüber dir und deinen Gefährten. Rolle für jedes Tier die Initiative separat, aber du bestimmst, ob die Tiere in deinem Zug handeln, und du kannst ihre Aktionen kontrollieren. Ein beschworenes Tier verschwindet, wenn es stirbt, wenn du den Zauber beendest oder wenn die Dauer endet.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich die Herausforderungsstufe um 1 für jeden zweiten Grad über dem 2.</p>
          `,
      description_en: `
          <p>You summon fey spirits that take the form of beasts and appear in unoccupied spaces that you can see within range. Choose one of the following options for what appears:</p>
          <ul>
            <li>One beast of challenge rating 2 or lower</li>
            <li>Two beasts of challenge rating 1 or lower</li>
            <li>Four beasts of challenge rating 1/2 or lower</li>
            <li>Eight beasts of challenge rating 1/4 or lower</li>
          </ul>
          <p>Each beast is also considered fey, and it disappears when it drops to 0 hit points or when the spell ends.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2063,
      german: 'Tier oder Pflanze aufspüren',
      english: 'Locate Animals or Plants',
      image: 'locateAnimalsOrPlants',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'instant',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Beschreibe oder benenne eine bestimmte Art von Tier oder Pflanze. Indem du dich auf die Stimme der Natur in deiner Umgebung konzentrierst, erfährst du die Richtung und Entfernung zur nächsten Kreatur oder Pflanze dieser Art im Umkreis von 8 Kilometern, falls es welche gibt.</p>
        `,
      description_en: `
          <p>Describe or name a specific kind of beast or plant. Concentrating on the voice of nature in your surroundings, you learn the direction and distance to the closest creature or plant of that kind within 5 miles, if any are present.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 26400,
      hasAreaOfEffect: false,
    },
    {
      id: 2064,
      german: 'Tierbote',
      english: 'Animal Messenger',
      image: 'animalMessenger',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 24,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du benutzt ein Tier, um eine Nachricht zuzustellen. Wähle ein winziges Tier in Reichweite, dass du sehen kannst, beispielsweise ein Eichhörnchen, ein Rotkehlchen oder eine Fledermaus. Du gibst einen Ort an, den du bereits besucht hast und einen Empfänger, der einer allgemeinen Beschreibung entspricht (z.B. ein Mann oder eine Frau in der Uniform der Stadtwache oder ein rothaariger Zwerg mit einem spitzen Hut). Du kannst eine Nachricht von bis zu 25 Worten aufsagen. Für die Zauberdauer reist das Tier in Richtung des angegeben Ortes. Fliegende Botschafter schaffen 80km in 24 Std, Landtiere 40km.</p>
          <p>Wenn der Botschafter ankommt, überbringt er deine Nachricht der Kreatur, die du beschrieben hast, indem es deine Stimme repliziert. Der Botschafter spricht nur zu einer Kreatur, die deiner Beschreibung entspricht. Wenn die Kreatur ihr Ziel vor dem Zauberende nicht erreicht, ist die Botschaft verloren und sie macht sich auf den Heimweg.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich die Reichweite um 40km für jeden zweiten Grad über dem 2.</p>
        `,
      description_en: `
          <p>By means of this spell, you use an animal to deliver a message. Choose a Tiny beast you can see within range, such as a squirrel, a blue ray, or a bird. You specify a location, which you must have visited, and a recipient who matches a general description, such as a man or woman dressed in the uniform of the town guard or a red-haired dwarf wearing a pointed hat. You also speak a message of up to twenty-five words. The target beast travels for the duration of the spell towards the specified location, covering about 50 miles per 24 hours for a flying messenger or 25 miles for other animals.</p>
          <p>When the messenger arrives, it delivers your message to the creature that you described, replicating the sound of your voice. The messenger speaks only to a creature matching the description you gave. If the messenger doesn't reach its destination before the spell ends, the message is lost, and the beast makes its way back to where you cast this spell.</p>
          <p><b>At higher levels:</b> If you cast this spell using a spell slot of 3rd level or higher, the duration of the spell increases by 48 hours for each slot level above 2nd.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2065,
      german: 'Tiersinn',
      english: 'Beast Sense',
      image: 'beastSense',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: true,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine willige Kreatur. Für die Dauer des Zaubers kannst du mit deiner Aktion durch die Augen des Tieres sehen und hören, was es hört, und zwar so lange, bis du mit deiner Aktion zu deinen normalen Sinnen zurückkehrst.</p>
          <p>Während du durch die Sinne des Tieres wahrnimmst, profitierst du von allen besonderen Sinnen, die diese Kreatur besitzt, bist aber für deine eigene Umgebung geblendet und taub.</p>
        `,
      description_en: `
          <p>You touch a willing beast. For the duration of the spell, you can use your action to see through the beast’s eyes and hear what it hears, and continue doing so until you use your action to return to your normal senses.</p>
          <p>While perceiving through the beast’s senses, you benefit from any special senses possessed by that creature, but you are blinded and deafened to your own surroundings.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2066,
      german: 'Unsichtbares Sehen',
      english: 'See Invisibility',
      image: 'seeInvisibility',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Für die Dauer des Zaubers siehst du unsichtbare Kreaturen und Gegenstände, als wären sie sichtbar, und du kannst in die ätherische Ebene sehen. Ätherische Kreaturen und Gegenstände erscheinen geisterhaft und durchsichtig.</p>
        `,
      description_en: `
          <p>For the duration, you see invisible creatures and objects as if they were visible, and you can see into the Ethereal Plane. Ethereal creatures and objects appear ghostly and translucent.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2067,
      german: 'Unsichtbarkeit',
      english: 'Invisibility',
      image: 'invisibility',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'wizard', 'sorcerer', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine Kreatur, die du berührst, wird unsichtbar, bis der Zauber endet. Alles, was das Ziel trägt, ist unsichtbar, solange es sich an der Person des Ziels befindet. Der Zauber endet für ein Ziel, das angreift oder einen Zauber wirkt.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, kannst du eine zusätzliche Kreatur für jeden zweiten Grad über dem 2 unsichtbar machen.</p>
          `,
      description_en: `
          <p>A creature you touch becomes invisible until the spell ends. Anything the target is wearing or carrying is invisible as long as it is on the target’s person. The spell ends for a target that attacks or casts a spell.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, you can target one additional creature for each slot level above 2nd.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 2068,
      german: 'Vergrößern/Verkleinern',
      english: 'Enlarge/Reduce',
      image: 'enlargeReduce',
      level: 2,
      isCustom: false,
      classes: ['test', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du lässt eine Kreatur oder einen Gegenstand in Reichweite, den du siehst, für die Dauer der Aktion größer oder kleiner werden. Wähle entweder eine Kreatur oder einen Gegenstand, der weder getragen noch mitgeführt wird. Wenn das Ziel unwillig ist, kann es einen Rettungswurf auf Konstitution machen. Bei einem Erfolg hat der Zauber keine Wirkung.</p>
          <p>Wenn das Ziel eine Kreatur ist, ändert alles, was sie trägt und mit sich führt, seine Größe. Jeder Gegenstand, der von einer betroffenen Kreatur fallen gelassen wird, nimmt sofort wieder seine normale Größe an.</p>
          <p>Vergrößern: Die Größe des Ziels verdoppelt sich in allen Dimensionen, und sein Gewicht wird mit acht multipliziert. Dieses Wachstum erhöht seine Größe um eine Kategorie, zum Beispiel von Mittel auf Groß. Wenn der Platz für das Ziel nicht ausreicht, um seine Größe zu verdoppeln, erreicht die Kreatur oder der Gegenstand die maximal mögliche Größe in dem verfügbaren Raum. Bis zum Ende des Zaubers hat das Ziel außerdem Vorteile bei Stärkeüberprüfungen und Rettungswürfen auf Stärke. Auch die Waffen des Ziels wachsen auf seine neue Größe an. Während diese Waffen vergrößert sind, verursachen die Angriffe des Ziels mit ihnen 1W4 zusätzlichen Schaden.</p>
          <p>Verkleinern: Die Größe des Ziels wird in allen Dimensionen halbiert und sein Gewicht auf ein Achtel der normalen Größe reduziert. Diese Verkleinerung verringert seine Größe um eine Kategorie, z.B. von Mittel auf Klein. Bis zum Ende des Zaubers hat das Ziel außerdem Nachteil bei Stärkeprüfungen und Rettungswürfen. Auch die Waffen des Ziels schrumpfen auf seine neue Größe. Während diese Waffen verkleinert sind, verursachen die Angriffe des Ziels mit ihnen 1W4 weniger Schaden (der Schaden kann dadurch nicht unter 1 sinken).</p>
          `,
      description_en: `
          <p>You cause a creature or an object you can see within range to grow larger or smaller for the duration. Choose either a creature or an object that is neither worn nor carried. If the target is unwilling, it can make a Constitution saving throw. On a success, the spell has no effect.</p>
          <p>If the target is a creature, everything it is wearing and carrying changes size with it. Any item dropped by an affected creature returns to normal size at once.</p>
          <p>Enlarge: The target’s size doubles in all dimensions, and its weight is multiplied by eight. This growth increases its size by one category - from Medium to Large, for example. If there isn’t enough room for the target to double its size, the creature or object attains the maximum possible size in the space available. Until the spell ends, the target also has advantage on Strength checks and Strength saving throws. The target’s weapons also grow to match its new size. While these weapons are enlarged, the target’s attacks with them deal 1d4 extra damage.</p>
          <p>Reduce: The target’s size is halved in all dimensions, and its weight is reduced to one-eighth of normal. This reduction decreases its size by one category - from Medium to Small, for example. Until the spell ends, the target also has disadvantage on Strength checks and Strength saving throws. The target’s weapons also shrink to match its new size. While these weapons are reduced, the target’s attacks with them deal 1d4 less damage (this can’t reduce the damage below 1).</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 2069,
      german: 'Verschwimmen',
      english: 'Blur',
      image: 'blur',
      level: 2,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dein Körper wird unscharf, bewegt sich und schwankt für alle, die dich sehen können. Während der Dauer des Zaubers hat jede Kreatur Nachteil bei Angriffswürfen gegen dich. Ein Angreifer ist gegen diesen Effekt immun, wenn er nicht auf Sicht angewiesen ist (wie bei Blindsicht) oder durch Illusionen hindurch sehen kann (wie bei Wahre Sicht).</p>
          `,
      description_en: `
          <p>Your body becomes blurred, shifting and wavering to all who can see you. For the duration, any creature has disadvantage on attack rolls against you. An attacker is immune to this effect if it doesn’t rely on sight, as with blindsight, or can see through illusions, as with truesight.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2070,
      german: 'Vorahnung',
      english: 'Augury',
      image: 'augury',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 0,
      durationType: 'instant',
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Indem du Stöckchen mit Juwelenintarsien wirfst, mit Drachenwürfeln würfelst, reich verzierte Karten auslegst oder ein anderes Weissagungswerkzeug verwendest, erhältst du ein Omen von einer außerweltlichen Wesenheit über die Folgen einer bestimmten Handlungsweise, die du für die nächsten 30 Minuten planst. Der Spielleiter wählt eines der folgenden möglichen Omen:</p>
          <ul>
          <li>Wohl, für gute Folgen.</li>
          <li>Wehe, für schlechte Folgen.</li>
          <li>Wohl und Wehe, für gute und schlechte Folgen.</li>
          <li>Kein Omen, für weder gut noch schlecht.</li>
          </ul>
          <p>Der Zauber nimmt keine Rücksicht auf mögliche Umstände, die das Ergebnis verändern könnten, wie das Wirken zusätzlicher Zauber oder den Verlust eines Gefährten.</p>
          <p>Wenn du den Zauber zweimal oder öfter wirkst, ehe du deine nächste lange Rast abschließt, besteht eine kumulative Chance von 25 Prozent für jeden Zauber nach dem ersten, dass du ein zufälliges Ergebnis bekommst. Der Spielleiter macht diesen Wurf verdeckt.</p>
        `,
      description_en: `
          <p>By casting gem-inlaid sticks, rolling dragon bones, laying ornate cards, or employing some other divining tool, you receive an omen from an otherworldly entity about the results of a specific course of action that you plan to take within the next 30 minutes. The DM chooses from the following possible omens:</p>
          <ul>
          <li>Weal, for good results.</li>
          <li>Woe, for bad results.</li>
          <li>Weal and woe, for both good and bad results.</li>
          <li>Nothing, for results that aren’t especially good or bad.</li>
          </ul>
          <p>The spell doesn’t take into account any possible circumstances that might change the outcome, such as the casting of additional spells or the loss or gain of a companion.</p>
          <p>If you cast the spell two or more times before completing your next long rest, there is a cumulative 25 percent chance for each casting after the first that you get a random reading. The DM makes this roll in secret.</p>
        `,
      school: 'divination',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 2071,
      german: 'Waffe des Glaubens',
      english: 'Spiritual Weapon',
      image: 'spiritualWeapon',
      level: 2,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceNumber: 1, diceType: 8, damageType: 'force' }],
      doesHeal: false,
      description_de: `
        <p>Du erschaffst eine Schwebende, gespenstische Waffe in Reichweite, die für die Wirkungsdauer anhält oder bis du den Zauber erneut wirkst. Wenn du den Zauber wirkst, kannst du einen Nahkampfangriff gegen eine Kreatur innerhalb von 5 Fuß zur Waffe durchführen. Bei einem Treffer erleidet das Ziel 1W8 + deinen Zaubermodifikator als Energieschaden.</p>
        <p>Als Bonusaktion in deinem Zug kannst du die Waffe bis zu 20 Fuß weit bewegen und den Angriff gegen eine Kreatur innerhalb von 5 Fuß um sie wiederholen.</p>
        <p>Die Waffe kann jede Form annehmen, die du wählst. Kleriker von Gottheiten, die mit einer bestimmten Waffe assoziiert sind (wie St. Cuthbert für seine Keule und Thor für seinen Hammer), lassen den Effekt dieses Zaubers wie diese Waffe aussehen.</p>
        <p><b>Auf höheren Graden:</b> Wirkst du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher, erhöht sich der Schaden um 1W8 für jeden zweiten Grad über dem 2.</p>
      `,
      description_en: `
        <p>You create a floating, spectral weapon within range that lasts for the duration or until you cast this spell again. When you cast the spell, you can make a melee spell attack against a creature within 5 feet of the weapon. On a hit, the target takes force damage equal to 1d8 + your spellcasting ability modifier. As a bonus action on your turn, you can move the weapon up to 20 feet and repeat the attack against a creature within 5 feet of it. The weapon can take whatever form you choose. clerics of deities who are associated with a particular weapon (as St. Cuthbert is known for his mace and Thor for his hammer) make this spell’s effect resemble that weapon.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for every two slot levels above 2nd.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 2072,
      german: 'Windstoß',
      english: 'Gust of Wind',
      image: 'gustOfWind',
      level: 2,
      isCustom: false,
      classes: ['test', 'druid', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine Linie starken Windes, 20 Meter lang und 3 Meter breit, bläst während der Dauer des Zaubers von dir aus in eine von dir gewählte Richtung. Jede Kreatur, die ihren Zug in der Linie beginnt, muss einen Rettungswurf auf Stärke bestehen oder wird 9 Meter von dir weg in eine Richtung gestoßen, die der Linie folgt.</p>
          <p>Jede Kreatur in der Linie muss für jeden 1 Meter, den sie sich bewegt, 2 Meter Bewegung aufwenden, wenn sie sich dir nähert.</p>
          <p>Die Böe zerstreut Gas oder Dampf und löscht Kerzen, Fackeln und ähnliche ungeschützte Flammen in der Umgebung. Sie lässt geschützte Flammen, wie die von Laternen, wild tanzen und hat eine 50-prozentige Chance, sie zu löschen.</p>
          <p>Als Bonusaktion kannst du in jedem deiner Züge, bevor der Zauber endet, die Richtung ändern, in die die Linie von dir wegfliegt.</p>
        `,
      description_en: `
          <p>A line of strong wind 60 feet long and 10 feet wide blasts from you in a direction you choose for the spell’s duration. Each creature that starts its turn in the line must succeed on a Strength saving throw or be pushed 30 feet away from you in a direction following the line.</p>
          <p>Any creature in the line must spend 2 feet of movement for every 1 foot it moves when moving closer to you.</p>
          <p>The gust disperses gas or vapor, and it extinguishes candles, torches, and similar unprotected flames in the area. It causes protected flames, such as those of lanterns, to dance wildly and has a 50 percent chance to extinguish them.</p>
          <p>As a bonus action on each of your turns before the spell ends, you can change the direction in which the line blasts from you.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 60,
    },
    {
      id: 2073,
      german: 'Zerbersten',
      english: 'Shatter',
      image: 'shatter',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'wizard', 'sorcerer', 'warlock'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'thunder' }],
      doesHeal: false,
      description_de: `
          <p>Ein plötzliches, lautes, schmerzhaft intensives Geräusch ertönt an einem Punkt deiner Wahl in Reichweite. Jede Kreatur, die sich in einem Umkreis von drei Metern um diesen Punkt befindet, muss einen Rettungswurf auf Konstitution machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 3W8 Donnerschaden, bei einem erfolgreichen Rettungswurf halb so viel Schaden. Eine Kreatur aus anorganischem Material wie Stein, Kristall oder Metall hat bei diesem Rettungswurf Nachteil.</p>
          <p>Ein nichtmagischer Gegenstand, der nicht getragen wird, erleidet ebenfalls Schaden, wenn er sich im Wirkungsbereich des Zaubers befindet.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden zweiten Grad über dem 2.</p>
        `,
      description_en: `
          <p>A sudden loud ringing noise, painfully intense, erupts from a point of your choice within range. Each creature in a 10-foot-radius sphere centered on that point must make a Constitution saving throw. A creature takes 3d8 thunder damage on a failed save, or half as much damage on a successful one. A creature made of inorganic material such as stone, crystal, or metal has disadvantage on this saving throw.</p>
          <p>A nonmagical object that isn’t being worn or carried also takes damage if it’s in the spell’s area.</p>
          <p><b>At higher levels:</b> When you cast this spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for every two slot levels above 2nd.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 2074,
      german: 'Zone der Wahrheit',
      english: 'Zone of Truth',
      image: 'zoneOfTruth',
      level: 2,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsConcentration: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'charisma',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst eine magische Zone, die in einem Radius von 4,5 Metern, zentriert auf einen Punkt deiner Wahl innerhalb der Reichweite, vor Täuschung schützt. Bis zum Ende des Zaubers muss eine Kreatur, die den Zauberbereich zum ersten Mal in ihrem Zug betritt oder ihren Zug dort beginnt, einen Rettungswurf auf Charisma ablegen. Bei einem misslungenen Rettungswurf kann die Kreatur keine absichtliche Lüge sprechen, solange sie sich im Radius befindet. Du weißt, ob der Rettungswurf jeder Kreatur gelingt oder misslingt.</p>
          <p>Eine betroffene Kreatur ist sich des Zaubers bewusst und kann daher vermeiden, Fragen zu beantworten, auf die sie normalerweise mit einer Lüge antworten würde. Eine solche Kreatur kann in ihren Antworten ausweichen, solange sie innerhalb der Grenzen der Wahrheit bleibt.</p>
        `,
      description_en: `
          <p>You create a magical zone that guards against deception in a 15-foot-radius sphere centered on a point of your choice within range. Until the spell ends, a creature that enters the spell’s area for the first time on a turn or starts its turn there must make a Charisma saving throw. On a failed save, a creature can’t speak a deliberate lie while in the radius. You know whether each creature succeeds or fails on its saving throw.</p>
          <p>An affected creature is aware of the spell and can thus avoid answering questions to which it would normally respond with a lie. Such creatures can be evasive in its answers as long as it remains within the boundaries of the truth.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 15,
    },

    // #endregion

    // #region LEVEL 3
    {
      id: 3000,
      german: 'Ashardalons Schreiten',
      english: 'Ashardalon’s Stride',
      image: 'ashardalonsStride',
      level: 3,
      isCustom: false,
      classes: ['test', 'artificer', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Deine Schritte hinterlassen eine Spur von Flammen, die sich sofort entzünden, wenn sie mit brennbaren Materialien in Berührung kommen. Deine Bewegungsgeschwindigkeit erhöht sich um 9 Meter, und du kannst über Lava und brennende Oberflächen laufen, ohne Schaden zu nehmen. Wenn du über eine brennende Oberfläche läufst, wird sie nicht gelöscht, aber du erleidest keinen Schaden durch das Feuer.</p>
          <p>Wenn du einen Nahkampfangriff durchführst, fügt er zusätzlich 1W6 Feuerschaden zu.</p>
        `,
      description_en: `
        <p>The billowing flames of a dragon blast from your feet, granting you explosive speed. For the duration, your speed increases by 20 feet and moving doesn’t provoke opportunity attacks.</p>
        <p>When you move within 5 feet of a creature or an object that isn’t being worn or carried, it takes 1d6 fire damage from your trail of heat. A creature or object can take this damage only once during a turn.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, increase your speed by 5 feet for each spell slot level above 3rd. The spell deals an additional 1d6 fire damage for each slot level above 3rd.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3001,
      german: 'Auf Wasser gehen',
      english: 'Water Walk',
      image: 'waterWalk',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber verleiht die Fähigkeit, sich über jede flüssige Oberfläche - wie Wasser, Säure, Schlamm, Schnee, Treibsand oder Lava - zu bewegen, als wäre es harmloser fester Boden (Kreaturen, die geschmolzene Lava durchqueren, können trotzdem Schaden durch die Hitze nehmen). Bis zu zehn bereitwillige Kreaturen, die du in Reichweite siehst, erhalten diese Fähigkeit für die Dauer des Zaubers.</p>
          <p>Wenn du eine Kreatur anvisierst, die in einer Flüssigkeit untergetaucht ist, trägt der Zauber das Ziel mit einer Geschwindigkeit von 18 Metern pro Runde an die Oberfläche der Flüssigkeit.</p>
          `,
      description_en: `
        <p>This spell grants the ability to move across any liquid surface—such as water, acid, mud, snow, quicksand, or lava—as if it were harmless solid ground (creatures crossing molten lava can still take damage from the heat). Up to ten willing creatures you can see within range gain this ability for the duration.</p>
        <p>If you target a creature submerged in a liquid, the spell carries the target to the surface of the liquid at a rate of 60 feet per round.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3002,
      german: 'Aura der Gesundheit',
      english: 'Aura of Vitality',
      image: 'auraOfVitality',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 2, diceType: 6 },
      description_de: `
          <p>Heilende Energie strahlt von dir in einer Aura mit einem Radius von 9 Meter aus. Bis zum Ende des Zaubers bewegt sich die Aura mit dir, auf dich zentriert. Mit einer Bonusaktion kannst du eine Kreatur in der Aura (dich eingeschlossen) dazu bringen, 2W6 Trefferpunkte zurückzugewinnen.</p>
        `,
      description_en: `
        <p>Healing energy radiates from you in an aura with a 30-foot radius. Until the spell ends, the aura moves with you, centered on you. You can use a bonus action to cause one creature in the aura (including you) to regain 2d6 hit points.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3003,
      german: 'Ausbrechende Erde',
      english: 'Erupting Earth',
      image: 'eruptingEarth',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 12, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
          <p>Wähle einen Punkt, den du auf dem Boden in Reichweite sehen kannst. Eine Fontäne aus aufgewühlter Erde und Stein bricht in einem 6-Meter-Würfel aus, der auf diesen Punkt zentriert ist. Jede Kreatur in diesem Bereich muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 3W12 Wuchtschaden, bei einem erfolgreichen Rettungswurf halb so viel Schaden. Außerdem wird der Boden in diesem Bereich zu schwierigem Gelände, bis er geräumt ist. Jeder Teil des Gebiets, der 1,5 Meter groß ist, braucht mindestens 1 Minute, um von Hand geräumt zu werden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 3. Grades oder höher wirkst, erhöht sich der Schaden um 1W12 für jeden Grad des Zauberplatzes über dem 2.</p>
          `,
      description_en: `
        <p>Choose a point you can see on the ground within range. A fountain of churned earth and stone erupts in a 20-foot cube centered on that point. Each creature in that area must make a Dexterity saving throw. A creature takes 3d12 bludgeoning damage on a failed save, or half as much damage on a successful one. Additionally, the ground in that area becomes difficult terrain until cleared. Each 5-foot-square portion of the area requires at least 1 minute to clear by hand.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d12 for each slot level above 3rd.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 3004,
      german: 'Beschwörung schwacher Dämonen',
      english: 'Summon Lesser Demons',
      image: 'summonLesserDemons',
      level: 3,
      isCustom: false,
      classes: ['test', 'warlock', 'wizard', 'sorcerer'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du sprichst schändliche Worte aus, um einen Dämon aus dem Chaos des Abyss zu beschwören. Würfle auf die folgende Tabelle, um festzulegen was passiert.</p>
          <table>
          <tr><th>W6</th><th>Ergebnis</th></tr>
          <tr><td>1-2</td><td>Zwei Dämonen mit HG 1 oder geringer</td></tr>
          <tr><td>3-4</td><td>Vier Dämonen mit HG 1/2 oder geringer</td></tr>
          <tr><td>5-6</td><td>Acht Dämonen mit HG 1/4 oder geringer</td></tr>
          </table>
          <p>Der SL wählt die Dämonen aus, wie zum Beispiel Manes oder Dretchs, und du wählst den unbesetzten Bereich, den du innerhalb der Reichweite sehen kannst, wo sie erscheinen werden. Ein beschworener Dämon verschwindet, wenn seine Trefferpunkte auf 0 Trefferpunkte reduziert werden oder wenn der Zauber endet.</p>
          <p>Die Dämonen sind allen Kreaturen gegenüber feindlich gesinnt, einschließlich dir. Würfle die Initiative für die Damenengruppe als Ganzes, welche ihre eigenen Züge hat. Die Dämonen verfolgen und attackieren nach bestem Vermögen die nächsten Nicht-Dämonen.</p>
          <p>Als Bestandteil des Wirkens des Zaubers kannst du auf dem Grund einen Kreis mit dem Blut bilden, das als Materialkomponente verwendet wird. Der Kreis ist groß genug um deinen Bereich zu umfassen. Während der Zauber wirkt, können die beschworenen Dämonen den Kreis weder überschreiten noch ihm Schaden zufügen. Zudem können sie niemanden innerhalb des Kreises als Ziel wählen. Die Materialkomponente in dieser Weise zu verwenden, zehrt sie auf, wenn der Zauber endet.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 6. oder 7. Grades verwendest, so beschwörst du doppelt so viele Dämonen. Wenn du diesen Zauber mit einem Zauberplatz des 8. oder 9. Grades verwendest, so beschwörst du dreimal so viele Dämonen.</p>
          `,
      description_en: `
        <p>You utter foul words, summoning demons from the chaos of the Abyss. Roll on the following table to determine what appears.</p>
        <table>
        <tr><th>D6</th><th>Result</th></tr>
        <tr><td>1-2</td><td>Two demons of challenge rating 1 or lower</td></tr>
        <tr><td>3-4</td><td>Four demons of challenge rating 1/2 or lower</td></tr>
        <tr><td>5-6</td><td>Eight demons of challenge rating 1/4 or lower</td></tr>
        </table>
        <p>The DM chooses the demons, such as manes or dretches, and you choose the unoccupied spaces you can see within range where they appear. A summoned demon disappears when it drops to 0 hit points or when the spell ends.</p>
        <p>The demons are hostile to all creatures, including you. Roll initiative for the summoned demons as a group, which has its own turns. The demons pursue and attack the nearest non-demons to the best of their ability.</p>
        <p>As part of casting the spell, you can form a circle on the ground with the blood used as a material component. The circle is large enough to encompass your space. While the spell is in effect, the summoned demons cannot cross the circle or harm it, and they can’t target anyone within it. Using the material component in this manner consumes it when the spell ends.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 6th or 7th level, you summon twice as many demons. If you cast it using an 8th or 9th-level spell slot, you summon three times as many demons.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 3005,
      german: 'Blendendes Niederstrecken',
      english: 'Blinding Smite',
      image: 'blindingSmite',
      level: 3,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 0,
      durationType: 'instant',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
          <p>Das nächste Mal, wenn du innerhalb der Wirkungsdauer des Zaubers eine Kreatur mit einer Nahkampfwaffe triffst, lodert deine Waffe in hellem Licht auf und verursacht beim Ziel zusätzlich 3W8 gleissenden Schaden. Ausserdem muss ihm ein Rettungswurf auf Konstitution gelingen, um nicht für die Wirkungsdauer zu erblinden.</p>
          <p>Eine Kreatur, die durch diesen Zauber geblendet wurde, darf am Ende jedes ihrer Zuge einen weiteren Rettungswurf auf Konstitution machen. Bei einem Erfolg verliert sie den Zustand blind.</p>
          `,
      description_en: `
        <p>The next time you hit a creature with a melee weapon attack during this spell’s duration, your weapon flares with bright light, and the attack deals an extra 3d8 radiant damage to the target. Additionally, the target must succeed on a Constitution saving throw or be blinded for the spell’s duration.</p>
        <p>A creature blinded by this spell makes another Constitution saving throw at the end of each of its turns. On a successful save, it is no longer blinded.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3006,
      german: 'Blitz',
      english: 'Lightning Bolt',
      image: 'lightningBolt',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 8, diceType: 6, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
          <p>Ein Blitzschlag, der eine 30 Meter lange und 1,5 Meter breite Linie bildet, schießt von dir aus in eine von dir gewählte Richtung. Jede Kreatur in dieser Linie muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet eine Kreatur 8W6 Blitzschaden, bei einem erfolgreichen Rettungswurf halb so viel Schaden.</p>
          <p>Der Blitz entzündet brennbare Gegenstände in der Umgebung, die nicht getragen werden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Grad über dem 3.</p>
          `,
      description_en: `
        <p>A stroke of lightning forming a line 100 feet long and 5 feet wide blasts out from you in a direction you choose. Each creature in the line must make a Dexterity saving throw. A creature takes 8d6 lightning damage on a failed save, or half as much damage on a successful one.</p>
        <p>The lightning ignites flammable objects in the area that aren’t being worn or carried.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d6 for each slot level above 3rd.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 100,
    },
    {
      id: 3007,
      german: 'Blitze herbeirufen',
      english: 'Call Lightning',
      image: 'callLightning',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 10, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
          <p>Eine Sturmwolke erscheint in der Gestalt eines Zylinders, der 3 Meter hoch ist, mit einem Radius von 18 Metern, zentriert um einen Punkt, den du sehen kannst, 30 Meter direkt über dir. Der Zauber misslingt, wenn du keinen Punkt in der Luft sehen kannst, an dem die Sturmwolke erscheinen könnte (wenn du beispielsweise in einem Raum bist, der die Wolke nicht aufnehmen kann).</p>
          <p>Wenn du den Zauber wirkst, wähle einen Punkt in Reichweite, den du sehen kannst. Ein Blitz fährt von der Wolke herab und Schlägt in diesen Punkt ein. Jede Kreatur in Umkreis von 1,5 Metern um den Punkt muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet das Ziel 3W10 Blitzschaden, halb so viel Schaden bei einem erfolgreichen Rettungswurf. In jedem deiner Züge, bis der Zauber endet, kannst du deine Aktion verwenden, um noch einen Blitz auf diese Weise zu beschwören, und er schlägt auf dem gleichen Punkt oder einem anderen ein.</p>
          <p>Wenn du dich draußen aufhältst und ein Sturm herrscht, wenn du diesen Zauber wirkst. Verleiht er dir die Kontrolle über den bestehenden Sturm und erschafft keinen eigenen. In diesem Fall steigt der Schaden des Zaubers um 1W10.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W10 für jeden Grad über dem 3.</p>
          `,
      description_en: `
        <p>A storm cloud appears in the shape of a cylinder that is 10 feet tall with a 60-foot radius, centered on a point you can see 100 feet directly above you. The spell fails if you can’t see a point in the air where the storm cloud could appear (for example, if you are in a room that can’t accommodate the cloud).</p>
        <p>When you cast the spell, choose a point you can see within range. A bolt of lightning flashes down from the cloud to that point. Each creature within 5 feet of that point must make a Dexterity saving throw. A creature takes 3d10 lightning damage on a failed save, or half as much damage on a successful one. On each of your turns until the spell ends, you can use your action to call down lightning in this way again, targeting the same point or a different one.</p>
        <p>If you are outdoors in stormy conditions when you cast this spell, the spell gives you control over the existing storm instead of creating a new one. Under such conditions, the spell’s damage increases by 1d10.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d10 for each slot level above 3rd.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3008,
      german: 'Blitzpfeil',
      english: 'Lightning Arrow',
      image: 'lightningArrow',
      level: 3,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 0,
      durationType: 'instant',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 4, diceType: 8, damageType: 'lightning' }],
      doesHeal: false,
      description_de: `
          <p>Wenn du das nächste Mal während der Dauer des Zaubers einen Angriff mit einer Fernkampfwaffe ausführst, verwandelt sich die Munition der Waffe oder die Waffe selbst, wenn es sich um eine Wurfwaffe handelt, in einen Blitz. Führe den Angriffswurf wie gewohnt durch. Das Ziel erleidet bei einem Treffer 4W8 Blitzschaden, bei einem Fehlschlag die Hälfte, anstatt des normalen Schadens der Waffe.</p>
          <p>Ob du triffst oder verfehlst, jede Kreatur im Umkreis von 3 Metern um das Ziel muss einen Rettungswurf auf Geschicklichkeit machen. Jede dieser Kreaturen erleidet bei einem misslungenen Rettungswurf 2W8 Blitzschaden, bei einem Erfolg die Hälfte des Schadens.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden für beide Effekte des Zaubers um 1W8 für jeden Slotgrad über dem 3.</p>
          `,
      description_en: `
        <p>The next time you make a ranged weapon attack during the spell’s duration, the weapon’s ammunition, or the weapon itself if it’s a thrown weapon, transforms into a bolt of lightning. Make the attack roll as normal. The target takes 4d8 lightning damage on a hit, or half as much damage on a miss, instead of the weapon’s normal damage.</p>
        <p>Whether you hit or miss, each creature within 10 feet of the target must make a Dexterity saving throw. Each of these creatures takes 2d8 lightning damage on a failed save, or half as much damage on a successful one.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage for both effects of the spell increases by 1d8 for each slot level above 3rd.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 3009,
      german: 'Donnerschritt',
      english: 'Thunder Step',
      image: 'thunderStep',
      level: 3,
      isCustom: false,
      classes: ['test', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 10, damageType: 'thunder' }],
      doesHeal: false,
      description_de: `
          <p>Du teleportierst dich selbst in einen nicht besetzten Bereich, den du innerhalb der Reichweite sehen kannst. Unmittelbar nachdem du verschwindest, erklingt ein gewaltiges Donnern, und jede Kreatur innerhalb von 3 m von dem Ort, den du verlassen hast, muss einen Konstitutionsrettungswurf ablegen. Bei einem Fehlschlag erleidet sie 3W10 Schallschaden, und die Hälfte davon bei einem Erfolg. Den Donner kann man bis zu 90 m weit hören.</p>
          <p>Du kannst auch Objekte mit dir nehmen, solange ihr Gewicht nicht das übersteigt, welches du tragen kannst. Du kannst auch eine bereitwillige Kreatur deiner Größe oder kleiner teleportieren, die Ausrüstung bis zu ihrer Traglast mit sich führt. Die Kreatur muss sich innerhalb von 1,50 m von dir befinden, wenn du den Zauber wirkst, und es muss an dem Ort wo du erscheinst einen unbesetzten Bereich innerhalb von 1,50 m geben, wo die Kreatur erscheinen kann. Andernfalls wird die Kreatur zurückgelassen.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W10 für jeden Grad über dem 3.</p>
          `,
      description_en: `
        <p>You teleport yourself to an unoccupied space you can see within range. Immediately after you disappear, a thunderous boom sounds, and each creature within 10 feet of the space you left must make a Constitution saving throw. On a failed save, the creature takes 3d10 thunder damage, and it is deafened for 1 minute. On a successful save, the creature takes half as much damage and isn’t deafened.</p>
        <p>You can also teleport one willing creature of your size or smaller who is carrying gear up to its carrying capacity. The creature must be within 5 feet of you when you cast this spell, and there must be an unoccupied space within 5 feet of your destination space for the creature to appear in; otherwise, the creature will be left behind.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d10 for each slot level above 3rd.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 3010,
      german: 'Elementare Waffe',
      english: 'Elemental Weapon',
      image: 'elementalWeapon',
      level: 3,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine nichtmagische Waffe, die du berührst, wird zu einer magischen Waffe. Wähle eine der folgenden Schadensarten: Säure, Kälte, Feuer, Blitz oder Donner. Für die Dauer hat die Waffe einen Bonus von +1 auf Angriffswürfe und verursacht bei einem Treffer zusätzlich 1W4 Schaden des gewählten Typs.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauberspruch mit einem Zauberplatz des 5. oder 6. Grades wirkst, erhöht sich der Bonus auf Angriffswürfe auf +2 und der zusätzliche Schaden auf 2W4. Wenn du einen Zauberplatz des 7. Grades oder höher verwendest, erhöht sich der Bonus auf +3 und der zusätzliche Schaden erhöht sich auf 3W4.</p>
          `,
      description_en: `
        <p>A nonmagical weapon you touch becomes a magic weapon. Choose one of the following damage types: acid, cold, fire, lightning, or thunder. For the duration, the weapon has a +1 bonus to attack rolls and deals an extra 1d4 damage of the chosen type when it hits.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 5th or 6th level, the bonus to attack rolls increases to +2 and the extra damage increases to 2d4. When you use a spell slot of 7th level or higher, the bonus increases to +3 and the extra damage increases to 3d4.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3011,
      german: 'Feenwesen beschwören',
      english: 'Conjure Fey',
      image: 'conjureFey',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du beschwörst ein Feenwesen des Herausforderungsgrades 6 oder niedriger oder einen Feengeist, der die Form eines Tieres des Herausforderungsgrades 6 oder niedriger annimmt. Es erscheint in einem unbesetzten Feld in Reichweite, das du sehen kannst. Das Feenwesen verschwindet, wenn es auf 0 Trefferpunkte fällt oder wenn der Zauber endet.</p>
          <p>Das Feenwesen ist dir und deinen Begleitern für die Dauer des Zaubers freundlich gesonnen. Würfle die Initiative für die Kreatur, die ihre eigenen Spielzüge hat. Sie gehorcht allen verbalen Befehlen, die du ihr gibst (ohne dass du etwas tun musst), solange sie nicht gegen ihre Ausrichtung verstoßen. Wenn du dem Feenwesen keine Befehle gibst, verteidigt es sich gegen feindliche Kreaturen, führt aber ansonsten keine Aktionen aus.</p>
          <p>Wenn deine Konzentration unterbrochen wird, verschwindet das Feenwesen nicht. Stattdessen verlierst du die Kontrolle über das Feenwesen, es wird dir und deinen Begleitern gegenüber feindselig und kann angreifen. Ein unkontrolliertes Feenwesen kann von dir nicht entlassen werden und verschwindet 1 Stunde, nachdem du es beschworen hast.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 7. Grades oder höher wirkst, erhöht sich der Herausforderungsgrad für jeden Zauberplatz-Grad über dem 6. um 1.</p>
          `,
      description_en: `
        <p>You summon a fey creature of challenge rating 6 or lower, or a fey spirit that takes the form of a beast of challenge rating 6 or lower. It appears in an unoccupied space that you can see within range. The fey creature disappears when it drops to 0 hit points or when the spell ends.</p>
        <p>The fey creature is friendly to you and your companions for the duration. Roll initiative for the creature, which has its own turns. It obeys any verbal commands that you issue to it (no action required by you), as long as they don’t violate its alignment. If you don’t issue any commands to the fey creature, it defends itself against hostile creatures but otherwise takes no actions.</p>
        <p>If your concentration is broken, the fey creature doesn’t disappear. Instead, you lose control of the fey creature, it becomes hostile toward you and your companions, and it might attack. An uncontrolled fey creature can’t be dismissed by you, and it disappears 1 hour after you summoned it.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 7th level or higher, the challenge rating increases by 1 for each slot level above 6.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 3012,
      german: 'Feindüberschuss',
      english: 'Enemies Abound',
      image: 'enemiesAbound',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'intelligence',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>You reach into the mind of one creature you can see and force it to make an Intelligence saving throw. A creature automatically succeeds if it is immune to being frightened. On a failed save, the target loses the ability to distinguish friend from foe, regarding all creatures it can see as enemies until the spell ends. Each time the target takes damage, it can repeat the saving throw, ending the effect on itself on a success.</p>
          <p>Whenever the affected creature chooses another creature as a target, it must choose the target at random from among the creatures it can see within range of the attack, spell, or other ability it’s using. If an enemy provokes an opportunity attack from the affected creature, the creature must make that attack if it is able to.</p>
          `,
      description_en: `
        <p>You reach into the mind of one creature you can see and force it to make an Intelligence saving throw. A creature automatically succeeds if it is immune to being frightened. On a failed save, the target loses the ability to distinguish friend from foe, regarding all creatures it can see as enemies until the spell ends. Each time the target takes damage, it can repeat the saving throw, ending the effect on itself on a success.</p>
        <p>Whenever the affected creature chooses another creature as a target, it must choose the target at random from among the creatures it can see within range of the attack, spell, or other ability it’s using. If an enemy provokes an opportunity attack from the affected creature, the creature must make that attack if it is able to.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 3013,
      german: 'Feuerball',
      english: 'Fireball',
      image: 'fireball',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 8, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Ein heller Strahl schießt von deinem Zeigefinger zu einem von dir gewählten Punkt in Reichweite und blüht dann mit einem tiefen Brüllen zu einer Flammenexplosion auf. Jede Kreatur in einem Umkreis von 6 Metern um diesen Punkt muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf erleidet das Ziel 8W6 Feuerschaden, bei einem erfolgreichen Wurf die Hälfte des Schadens.</p>
          <p>Das Feuer breitet sich um Ecken aus. Es entzündet brennbare Gegenstände in der Umgebung, die nicht getragen werden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Grad über dem 3.</p>
          `,
      description_en: `
        <p>A bright streak flashes from your pointing finger to a point you choose within range and then blossoms with a low roar into an explosion of flame. Each creature in a 20-foot-radius sphere centered on that point must make a Dexterity saving throw. A target takes 8d6 fire damage on a failed save, or half as much damage on a successful one.</p>
        <p>The fire spreads around corners. It ignites flammable objects in the area that aren’t being worn or carried.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d6 for each slot level above 3rd.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 150,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 3014,
      german: 'Flammenpfeile',
      english: 'Flame Arrows',
      image: 'flameArrows',
      level: 3,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst einen Köcher, der Pfeile oder Bolzen enthält. Wenn ein Ziel durch einen Angriff mit einer Fernkampfwaffe getroffen wird, bei dem ein Stück Munition aus dem Köcher gezogen wurde, erleidet das Ziel zusätzlich 1W6 Feuerschaden. Die Magie des Zaubers endet auf dem Munitionsstück, wenn es trifft oder verfehlt, und der Zauber endet, wenn zwölf Munitionsstücke aus dem Köcher gezogen worden sind.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich die Anzahl der Munitionsstücke, die du mit diesem Zauber beeinflussen kannst, um zwei für jeden Zauberplatz-Grad über dem 3.</p>
          `,
      description_en: ` 
        <p>You touch a quiver containing arrows or bolts. When a target is hit by a ranged weapon attack using a piece of ammunition drawn from the quiver, the target takes an extra 1d6 fire damage. The spell’s magic ends on the piece of ammunition when it hits or misses, and the spell ends when twelve pieces of ammunition have been drawn from the quiver.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the number of pieces of ammunition you can affect with this spell increases by two for each slot level above 3rd.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3015,
      german: 'Fliegen',
      english: 'Fly',
      image: 'fly',
      level: 3,
      isCustom: false,
      classes: ['test', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine bereitwillige Kreatur. Das Ziel erhält für die Dauer des Zaubers eine Fluggeschwindigkeit von 90 Metern. Wenn der Zauber endet, fällt das Ziel, falls es sich noch in der Luft befindet, es sei denn, es kann den Fall aufhalten.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, kannst du eine zusätzliche Kreatur für jeden Slot über dem 3 anvisieren.</p>
          `,
      description_en: `
        <p>You touch a willing creature. The target gains a flying speed of 60 feet for the duration. When the spell ends, the target falls if it is still aloft, unless it can stop the fall.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, you can target one additional creature for each slot level above 3.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3016,
      german: 'Flimmern',
      english: 'Blink',
      image: 'blink',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wirf einen W20 am Ende jedes deines Zuges für die Zauberdauer. Bei einer 11 oder höher verschwindest du von deiner momentanen Ebene der Existenz und tauchst in der Ätherischen Ebene wieder auf (der Zauber schlägt fehl und der Zauberplatz wird verschwendet, wenn du bereits auf dieser Ebene bist). Wenn du auf der Ätherischen Ebene bist, kannst du am Anfang deines nächsten Zuges, oder wenn der Zauber endet, an einen freien Punkt, den du sehen kannst und der nicht weiter als 3 Meter entfernt ist, zurückkehren. Wenn es keinen freien Punkt in Reichweite gibt, tauchst du an der nächst möglichen Stelle auf (gibt es mehrere, wird zufällig entschieden). Du kannst den Zauber mit einer Aktion beenden.</p>
          <p>Während du auf der Ätherischen Ebene bist, kannst du die Ursprungsebene sehen und hören, aber alles besteht aus grauen Schemen und ist gedämpft. Du kannst nicht weiter als 18 Meter sehen und Interaktion ist nur mit Kreaturen auf der Ätherischen Ebene möglich. Kreaturen, die sich nicht dort aufhalten, können dich weder wahrnehmen noch mit der inter-agieren, außer sie haben die Fähigkeit das zu tun.</p>
          `,
      description_en: `
        <p>Roll a d20 at the end of each of your turns for the duration of the spell. On a roll of 11 or higher, you vanish from your current plane of existence and appear in the Ethereal Plane (the spell fails and the spell slot is wasted if you were already on that plane). At the start of your next turn, and when the spell ends if you are on the Ethereal Plane, you return to an unoccupied space of your choice that you can see within 10 feet of the space you vanished from. If no unoccupied space is available within that range, you appear in the nearest unoccupied space (chosen at random if more than one space is equally near). You can dismiss this spell as an action.</p>
        <p>While on the Ethereal Plane, you can see and hear the plane you originated from, which is cast in shades of gray, and you can’t see anything there more than 60 feet away. You can only affect and be affected by other creatures on the Ethereal Plane. Creatures that aren’t there can’t perceive you or interact with you, unless they have the ability to do so.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3017,
      german: 'Fluch',
      english: 'Bestow Curse',
      image: 'bestowCurse',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du berührst eine Kreatur, die einen Rettungswurf auf Weisheit machen muss. Bei einem Fehlschlag wird sie für die Zauberdauer verflucht. Wähle die Natur des Fluchs aus den folgenden Optionen:</p>
          <ul>
            <li>Wähle ein Attribut. Solange der Fluch besteht, hat das Ziel Nachteil auf alle Rettungs- und Fähigkeitswürfe mit diesem Attribut.</li>
            <li>Solange der Fluch besteht, hat das Ziel Nach-teil auf alle Angriffswürfe gegen dich.</li>
            <li>Solange der Fluch besteht, muss das Ziel am Anfang jedes ihres Zuges einen Weisheitsrettungswurf bestehen.Bei einem Fehlschlag verschwendet es seine Aktion für diese Runde und tut nichts.</li>
            <li>Solange der Fluch besteht, verursachen alle deine Angriffe und Zauber zusätzlich 1W8 nekrotischen Schaden am Ziel.</li>
          </ul>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. oder eines höheren Grades sprichst, erhöht sich die Zauberdauer auf 10 Minuten Konzentration. Wenn du einen Zauberplatz des 5. oder eines höheren Grades benutzt, beträgt die Dauer 8 Stunden. Wenn du einen Zauberplatz des 7. oder eines höheren Grades benutzt, beträgt die Dauer 24 Stunden. Wenn du einen Zauberplatz des 9. Grades benutzt, dauert der Zauber an, bis er gebannt wird. Ein Zauberplatz des 5. oder eines höheren Grades erfordert keine Konzentration mehr.</p>
          `,
      description_en: `
        <p>You touch a creature, and that creature must succeed on a Wisdom saving throw or become cursed for the duration of the spell. When you cast this spell, choose the nature of the curse from the following options:</p>
        <ul>
          <li>Choose one attribute. While cursed, the target has disadvantage on ability checks and saving throws made with that attribute.</li>
          <li>While cursed, the target has disadvantage on attack rolls against you.</li>
          <li>While cursed, the target must make a Wisdom saving throw at the start of each of its turns. If it fails, it wastes its action that turn doing nothing.</li>
          <li>While the target is cursed, your attacks and spells deal an extra 1d8 necrotic damage to the target.</li>
        </ul>
        <p><b>At higher levels:</b> If you cast this spell using a spell slot of 4th level or higher, the duration is concentration, up to 10 minutes. If you use a spell slot of 5th level or higher, the duration is 8 hours. If you use a spell slot of 7th level or higher, the duration is 24 hours. If you use a 9th level spell slot, the spell lasts until it is dispelled. Using a spell slot of 5th level or higher doesn’t require concentration.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3018,
      german: 'Fluch brechen',
      english: 'Remove Curse',
      image: 'removeCurse',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Bei deiner Berührung enden alle Flüche, die eine Kreatur oder einen Gegenstand betreffen. Wenn es sich bei dem Gegenstand um einen verfluchten magischen Gegenstand handelt, bleibt der Fluch bestehen, aber der Zauber unterbricht die Einstimmung des Besitzers auf den Gegenstand, so dass dieser entfernt oder weggeworfen werden kann.</p>
          `,
      description_en: `
        <p>At your touch, all curses affecting one creature or object end. If the object is a cursed magic item, its curse remains, but the spell breaks its owner’s attunement to the object so it can be removed or discarded.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3019,
      german: 'Flutwelle',
      english: 'Tidal Wave',
      image: 'tidalWave',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 4, diceType: 8, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
          <p>Du beschwörst eine Wasserwelle herauf, die auf ein Gebiet in Reichweite niederprasselt. Das Gebiet kann bis zu 9 Meter lang, bis zu 3 Meter breit und bis zu 3 Meter hoch sein. Jede Kreatur in diesem Gebiet muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem Fehlschlag erleidet die Kreatur 4W8 Wuchtschaden und wird auf den Boden geworfen. Bei einem Erfolg erleidet die Kreatur halb so viel Schaden und wird nicht umgeworfen. Das Wasser breitet sich dann über den Boden in alle Richtungen aus und löscht ungeschützte Flammen in seinem Bereich und im Umkreis von 9 Metern.</p>
          `,
      description_en: `
        <p>You conjure up a wave of water that crashes down on an area within range. The area can be up to 30 feet long, up to 10 feet wide, and up to 10 feet tall. Each creature in that area must make a Dexterity saving throw. On a failed save, a creature takes 4d8 bludgeoning damage and is knocked prone. On a successful save, a creature takes half as much damage and isn’t knocked prone. The water then spreads out across the ground in all directions, extinguishing unprotected flames in its area and within 30 feet of it.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'square',
      length: 30,
    },
    {
      id: 3020,
      german: 'Furcht',
      english: 'Fear',
      image: 'fear',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du projizierst ein phantasmagorisches Bild der schlimmsten Ängste einer Kreatur. Jede Kreatur im Umkreis von 9 Metern muss einen Rettungswurf auf Weisheit machen oder alles, was sie in der Hand hält, fallen lassen und wird für die Dauer des Zaubers verängstigt.</p>
          <p>Während eine Kreatur durch diesen Zauber verängstigt ist, muss sie in jedem ihrer Züge die Sprintaktion ausführen und sich auf dem sichersten verfügbaren Weg von dir entfernen, es sei denn, sie kann sich nirgendwo hinbewegen. Wenn die Kreatur ihren Zug an einem Ort beendet, an dem sie keine Sichtlinie zu dir hat, kann sie einen Rettungswurf auf Weisheit machen. Bei einem erfolgreichen Rettungswurf endet der Zauber für diese Kreatur.</p>
          `,
      description_en: `
        <p>You project a phantasmal image of a creature’s worst fears. Each creature in a 30-foot cone must succeed on a Wisdom saving throw or drop whatever it is holding and become frightened for the duration of the spell.</p>
        <p>While a creature is frightened by this spell, it must take the Dash action and move away from you by the safest available route on each of its turns, unless there is nowhere to move. If the creature ends its turn in a location where it doesn’t have line of sight to you, the creature can make a Wisdom saving throw. On a successful save, the spell ends for that creature.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3021,
      german: 'Gasförmige Gestalt',
      english: 'Gaseous Form',
      image: 'gaseousForm',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du verwandelst eine bereitwillige Kreatur, die du berührst, zusammen mit allem, was sie trägt, für die Dauer des Zaubers in eine Nebelwolke. Der Zauber endet, wenn die Kreatur auf 0 Trefferpunkte fällt. Eine körperlose Kreatur ist davon nicht betroffen.</p>
          <p>Während das Ziel in dieser Form ist, kann es sich nur mit einer Fluggeschwindigkeit von 3 Metern bewegen. Das Ziel kann den Raum einer anderen Kreatur betreten und einnehmen. Das Ziel ist resistent gegen nichtmagischen Schaden und hat Vorteil bei Rettungswürfen auf Stärke, Geschicklichkeit und Konstitution. Das Ziel kann durch kleine Löcher, schmale Öffnungen und sogar durch einfache Risse hindurchgehen, wobei es Flüssigkeiten wie feste Oberflächen behandelt. Das Ziel kann nicht fallen und bleibt in der Luft schweben, selbst wenn es betäubt oder anderweitig außer Gefecht gesetzt ist.</p>
          <p>Während es sich in der Form einer Nebelwolke befindet, kann das Ziel nicht sprechen oder Objekte manipulieren, und alle Objekte, die es trägt oder hält, können nicht fallen gelassen, benutzt oder anderweitig beeinflusst werden. Das Ziel kann nicht angreifen oder zaubern.</p>
          `,
      description_en: `
        <p>You transform a willing creature you touch, along with everything it’s wearing and carrying, into a misty cloud for the duration. The spell ends if the creature drops to 0 hit points. An incorporeal creature isn’t affected.</p>
        <p>While in this form, the target’s only method of movement is a flying speed of 10 feet. The target can enter and occupy the space of another creature. The target has resistance to nonmagical damage, and it has advantage on Strength, Dexterity, and Constitution saving throws. The target can pass through small holes, narrow openings, and even mere cracks, though it treats liquids as though they were solid surfaces. The target can’t fall and remains hovering in the air even when stunned or otherwise incapacitated.</p>
        <p>While in the form of a misty cloud, the target can’t talk or manipulate objects, and any objects it was carrying or holding can’t be dropped, used, or otherwise interacted with. The target can’t attack or cast spells.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3022,
      german: 'Gegenzauber',
      english: 'Counterspell',
      image: 'counterspell',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'reaction',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du versuchst eine Kreatur zu unterbrechen, wenn sie einen Zauber wirkt. Wenn die Kreatur einen Zauber des 3. oder eines niederen Grades wirkt, dann schlägt der Zauber automatisch fehl. Wenn sie einen Zauber des 4. oder eines höheren Grades wirkt, musst du einen Attributswurf mit deinem Zauberattribut machen. Der SG ist 10 + der Grad des Zaubers. Bei einem Erfolg schlägt der Zauber fehl und hat keinen Effekt.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. oder eines höheren Grades sprichst, schlägt der unterbrochene Zauber fehl, sofern er dem Grad entspricht oder von einem geringeren Grad ist als der Grad, mit dem du Gegenzauber wirkst.</p>
          `,
      description_en: `
        <p>You attempt to interrupt a creature in the process of casting a spell. If the creature is casting a spell of 3rd level or lower, its spell fails and has no effect. If it is casting a spell of 4th level or higher, make an ability check using your spellcasting ability. The DC equals 10 + the spell’s level. On a success, the creature’s spell fails and has no effect.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the interrupted spell has no effect if its level is less than or equal to the level of the spell slot you used.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3023,
      german: 'Geisterross',
      english: 'Phantom Steed',
      image: 'phantomSteed',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine große, quasi-reale, pferdeähnliche Kreatur erscheint auf dem Boden in einem unbesetzten Raum deiner Wahl in Reichweite. Du bestimmst das Aussehen der Kreatur, aber sie ist mit einem Sattel, einer Trense und Zaumzeug ausgestattet. Alle durch den Zauber erzeugten Ausrüstungsgegenstände verschwinden in einer Rauchwolke, wenn sie mehr als 3 Meter vom Ross entfernt getragen werden.</p>
          <p>Für die Dauer des Zaubers kannst du oder eine Kreatur deiner Wahl das Ross reiten. Die Kreatur verwendet die Statistiken eines Reitpferdes, außer dass sie eine Geschwindigkeit von 100 Metern pro Runde hat und 5 Kilometer in der Stunde zurücklegen kann, bzw. 20 Kilometer im Galopp. Wenn der Zauber endet, verblasst das Ross allmählich und gibt dem Reiter 1 Minute Zeit, abzusteigen. Der Zauber endet, wenn du eine Aktion verwendest, um ihn zu beenden, oder wenn das Ross Schaden nimmt.</p>
          `,
      description_en: `
        <p>A Large quasi-real, horselike creature appears on the ground in an unoccupied space of your choice within range. You decide the creature’s appearance, but it is equipped with a saddle, bit, and bridle. Any of the equipment created by the spell vanishes in a puff of smoke if it is carried more than 10 feet away from the steed.</p>
        <p>For the duration, you or a creature you choose can ride the steed. The creature uses the statistics for a riding horse, except it has a speed of 100 feet and can travel 10 miles in an hour, or 13 miles at a fast pace. When the spell ends, the steed gradually fades, giving the rider 1 minute to dismount. The spell ends if you use an action to dismiss it or if the steed takes any damage.</p>
        `,
      school: 'illusion',
      isRanged: false,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3024,
      german: 'Geistwolke',
      english: 'Spirit Shroud',
      image: 'spiritShroud',
      level: 3,
      isCustom: false,
      classes: ['test', 'paladin', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du rufst Geister der Toten herbei, die für die Wirkungsdauer des Zaubers um dich herumflattern. Die Geister sind immateriell und unverwundbar.</p>
          <p>Bis der Zauber endet, fügst du Kreaturen, die du innerhalb von drei Metern mit einem Angriff triffst, 1W8 zusätzlichen Schaden zu. Die Schadensart ist gleißend, kalt oder nekrotisch (deine Wahl beim Wirken des Zaubers). Kreaturen, die diesen Schaden erleiden, können bis zu Beginn deines nächsten Zuges keine Trefferpunkte regenerieren.</p>
          <p>Außerdem kannst du die Bewegungsrate jeglicher Kreatur, die du sehen kannst und die ihren Zug innerhalb von drei Metern von dir beginnt, bis zum Beginn deines nächsten Zuges um drei Meter reduzieren.</p>
          `,
      description_en: `
        <p>You call forth spirits of the dead, which flit around you for the spell’s duration. The spirits are intangible and invulnerable.</p>
        <p>Until the spell ends, any attack you make deals 1d8 extra damage when you hit a creature within 10 feet of you. The damage type is radiant, necrotic, or cold (your choice when you cast the spell). Any creature that takes this damage can’t regain hit points until the start of your next turn.</p>
        <p>In addition, any creature that starts its turn within 10 feet of you has its speed reduced by 10 feet until the start of your next turn.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3025,
      german: 'Glyphe des Schutzes',
      english: 'Glyph of Warding',
      image: 'glyphOfWarding',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'hours',
      duration: 0,
      durationType: 'permanent',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wenn du diesen Zauber sprichst, ritzt du eine Glyphe ein, die später einen magischen Effekt auslöst. Die Glyphe wird entweder auf einer Oberfläche (z.B. einem Tisch, einem Stück Boden oder einer Wand) oder in einem verschließbaren Gegenstand (z.B. einem Buch, einer Schriftrolle oder einer Schatztruhe) eingraviert, um die Glyphe zu verbergen. Die Glyphe kann eine Fläche mit einem Durchmesser von maximal 3 Metern bedecken. Wird die Fläche oder der Gegenstand mehr als 3 Meter von der Stelle entfernt, an der du diesen Zauber gewirkt hast, wird die Glyphe zerbrochen und der Zauber endet, ohne dass er ausgelöst wird.</p>
          <p>Die Glyphe ist nahezu unsichtbar und erfordert eine erfolgreiche Intelligenzprobe (Erkundung) gegen deinen Zauberrettungswurf SG, um gefunden zu werden.</p>
          <p>Du entscheidest, was die Glyphe auslöst, wenn du den Zauber sprichst. Bei Glyphen, die auf einer Oberfläche eingraviert sind, sind die typischsten Auslöser das Berühren oder Stehen auf der Glyphe, das Entfernen eines anderen Objekts, das die Glyphe verdeckt, das Annähern an die Glyphe in einer bestimmten Entfernung oder das Manipulieren des Objekts, auf dem die Glyphe eingraviert ist. Bei Glyphen, die in ein Objekt eingraviert sind, sind die häufigsten Auslöser das Öffnen des Objekts, die Annäherung an das Objekt in einer bestimmten Entfernung oder das Sehen oder Lesen der Glyphe. Sobald eine Glyphe ausgelöst wird, endet dieser Zauber.</p>
          <p>Du kannst den Auslöser weiter verfeinern, so dass der Zauber nur unter bestimmten Umständen oder in Abhängigkeit von körperlichen Merkmalen (wie Größe oder Gewicht), der Art der Kreatur (zum Beispiel könnte der Schutzzauber so eingestellt werden, dass er Aberrationen oder Drow betrifft) oder der Ausrichtung aktiviert wird. Du kannst auch Bedingungen für Kreaturen festlegen, die die Glyphe nicht auslösen, wie zum Beispiel diejenigen, die ein bestimmtes Passwort sagen.</p>
          <p>Wenn du die Glyphe beschriftest, wähle "explosive Runen" oder eine "Zauberglyphe".</p>
          <p><b>Explosive Runen:</b> Wenn die Glyphe ausgelöst wird, bricht sie mit magischer Energie in einer Kugel mit einem Radius von 6 Metern aus, die sich um die Glyphe dreht. Die Kugel breitet sich um Ecken herum aus. Jede Kreatur in diesem Gebiet muss einen Rettungswurf auf Geschicklichkeit machen. Eine Kreatur erleidet 5W8 Säure-, Kälte-, Feuer-, Blitz- oder Donnerschaden bei einem misslungenen Rettungswurf (den du bei der Erschaffung der Glyphe wählst) oder halb so viel Schaden bei einem erfolgreichen Rettungswurf.</p>
          <p><b>Zauberglyphe:</b> Du kannst einen vorbereiteten Zauber des 3. Grades oder niedriger in der Glyphe speichern, indem du ihn als Teil der Erstellung der Glyphe wirkst. Der Zauber muss auf eine einzelne Kreatur oder ein Gebiet gerichtet sein. Der gespeicherte Zauber hat keine unmittelbare Wirkung, wenn er auf diese Weise gewirkt wird. Wenn die Glyphe ausgelöst wird, wird der gespeicherte Zauber gewirkt. Wenn der Zauber ein Ziel hat, zielt er auf die Kreatur, die die Glyphe ausgelöst hat. Wirkt der Zauber auf ein Gebiet, wird das Gebiet auf diese Kreatur zentriert. Wenn der Zauber feindliche Kreaturen herbeiruft oder schädliche Objekte oder Fallen erzeugt, erscheinen diese so nah wie möglich am Eindringling und greifen ihn an. Wenn der Zauber Konzentration erfordert, hält er bis zum Ende seiner vollen Dauer an.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden einer Glyphe "explosive Runen" um 1W8 für jeden Zauberplatz-Grad über dem 3. Wenn du eine "Zauberglyphe" erschaffst, kannst du einen beliebigen Zauber bis zum gleichen Grad wie der Slot, den du für die "Glyphe der Schutzes" benutzt, speichern.</p>
          `,
      description_en: `
        <p>When you cast this spell, you inscribe a glyph that later unleashes a magical effect. The glyph can be inscribed on a surface (such as a table, a section of floor, or a wall) or within an object that can be closed (such as a book, a scroll, or a treasure chest) to conceal the glyph. If you choose a surface, the glyph can cover an area of the surface no larger than 10 feet in diameter. If you choose an object, that object must remain in its place; if the object is moved more than 10 feet from where you cast this spell, the glyph is broken, and the spell ends without being triggered.</p>
        <p>The glyph is nearly invisible and requires a successful Intelligence (Investigation) check against your spell save DC to be found.</p>
        <p>You decide what triggers the glyph when you cast the spell. For glyphs inscribed on a surface, the most typical triggers include touching or standing on the glyph, removing another object covering the glyph, approaching within a certain distance of the glyph, or manipulating the object on which the glyph is inscribed. For glyphs inscribed within an object, the most common triggers include opening that object, approaching within a certain distance of the object, or seeing or reading the glyph. Once a glyph is triggered, this spell ends.</p>
        <p>You can further refine the trigger so the spell activates only under certain circumstances or according to physical characteristics (such as height or weight), creature kind (for example, the ward could be set to affect aberrations or drow), or alignment. You can also set conditions for creatures that don’t trigger the glyph, such as those who say a certain password.</p>
        <p>When you inscribe the glyph, choose "explosive runes" or a "spell glyph."</p>
        <p><b>Explosive Runes:</b> When the glyph is triggered, the glyph erupts with magical energy in a 20-foot-radius sphere centered on the glyph. The sphere spreads around corners. Each creature in the area must make a Dexterity saving throw. A creature takes 5d8 acid, cold, fire, lightning, or thunder damage on a failed saving throw (your choice when you create the glyph) or half as much damage on a successful one.</p>
        <p><b>Spell Glyph:</b> You can store a prepared spell of 3rd level or lower in the glyph by casting it as part of creating the glyph. The spell must target a single creature or an area. The spell being stored has no immediate effect when cast in this way. When the glyph is triggered, the stored spell is cast. If the spell has a target, it targets the creature that triggered the glyph. If the spell affects an area, the area is centered on that creature. If the spell summons hostile creatures or creates harmful objects or traps, they appear as close as possible to the intruder and attack it. If the spell requires concentration, it lasts until the end of its full duration.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage of an "explosive runes" glyph increases by 1d8 for each slot level above 3. If you create a "spell glyph," you can store any spell of up to the same level as the slot you use for the glyph of warding.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 10,
    },
    {
      id: 3026,
      german: 'Hast',
      english: 'Haste',
      image: 'haste',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wähle eine bereitwillige Kreatur, in Reichweite die du sehen kannst. Bis zum Ende des Zaubers wird die Geschwindigkeit des Ziels verdoppelt, es erhält einen Bonus von +2 auf RK, es hat Vorteil bei Rettungswürfen in Geschicklichkeit und es erhält in jeder seiner Runden eine zusätzliche Aktion. Diese Aktion kann nur für die Aktionen Angriff (nur ein Waffenangriff), Sprint, Ausweichen, Verstecken oder Gegenstand benutzen genutzt werden.</p>
          <p>Wenn der Zauber endet, kann sich das Ziel bis nach seinem nächsten Zug nicht mehr bewegen oder Aktionen ausführen, da eine Welle der Lethargie über es hinwegfegt.</p>
          `,
      description_en: `
        <p>Choose a willing creature that you can see within range. Until the spell ends, the target’s speed is doubled, it gains a +2 bonus to AC, it has advantage on Dexterity saving throws, and it gains an additional action on each of its turns. That action can be used only to take the Attack (one weapon attack only), Dash, Disengage, Hide, or Use an Object action.</p>
        <p>When the spell ends, the target can’t move or take actions until after its next turn, as a wave of lethargy sweeps over it.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3027,
      german: 'Hellsehen',
      english: 'Clairvoyance',
      image: 'clairvoyance',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'sorcerer', 'wizard'],
      timeToCast: 10,
      cost: 'minutes',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst einen unsichtbaren Sensor in Reichweite an einem dir bekannten Ort (ein Ort, den du bereits besucht oder gesehen hast) oder an einem offensichtlichen Ort, der dir unbekannt ist (z.B. hinter einer Tür oder hinter einer Ecke). Der Sensor verbleibt für die Zauberdauer an diesem Ort und kann weder angegriffen, noch kann mit ihm interagiert werden.</p>
          <p>Wenn du diesen Zauber wirkst, wählst du, ob du sehen oder hören willst. Deinen gewählten Sinn kannst du benutzen, als ob du an der Stelle des Sensor wärst. Du kannst mit einer Aktion zwischen sehen und hören wechseln.</p>
          <p>Eine Kreatur, die den Sensor sehen kann (eine Kreatur die Wahrsicht oder Unsichtbares Sehen besitzt), sieht eine leuchtende, immaterielle Kugel in der Größe deiner Faust.</p>
          `,
      description_en: `
        <p>You create an invisible sensor within range at a location familiar to you (a place you have visited or seen before) or in an obvious location that is unfamiliar to you (such as behind a door, around a corner, or in a grove of trees). The sensor remains in place for the duration and can’t be attacked or otherwise interacted with.</p>
        <p>When you cast the spell, you choose seeing or hearing. You can use the chosen sense through the sensor as if you were in its space. As your action, you can switch between seeing and hearing.</p>
        <p>A creature that can see the sensor (such as one that can see invisible objects or creatures) sees a luminous, intangible orb about the size of your fist.</p>
        `,
      school: 'divination',
      isRanged: true,
      range: 5280,
      hasAreaOfEffect: false,
    },
    {
      id: 3028,
      german: 'Hunger von Hadar',
      english: 'Hunger of Hadar',
      image: 'hungerOfHadar',
      level: 3,
      isCustom: false,
      classes: ['test', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [
        { diceNumber: 2, diceType: 6, damageType: 'cold' },
        { diceNumber: 2, diceType: 6, damageType: 'acid' },
      ],
      doesHeal: false,
      description_de: `
          <p>Du öffnest ein Tor zur Dunkelheit zwischen den Sternen, ein Gebiet, das von unbekannten Schrecken erfüllt ist. Eine Sphäre aus Schwärze und bitterer Kälte mit einem Radius von 6 m erscheint, zentriert an einem Punkt in Reichweite, und bleibt für die Wirkungsdauer bestehen. Diese Leere ist von einer Kakophonie leisen Flüsterns und schmatzender Geräusche erfüllt, die auf eine Entfernung von bis zu 9 m hörbar sind. Kein Licht, ob magischen oder natürlichen Ursprungs, vermag den Bereich der Sphäre zu erheilen, und Kreaturen, die sich vollständig in diesem aufhalten, sind blind.</p>
          <p>Die Leere erzeugt eine Verzerrung in der Struktur des Raums und der Bereich gilt als schwieriges Gelände. Alle Kreaturen, die ihren Zug in dem Bereich beginnen, erleiden 2W6 Kälteschaden. Alle Kreaturen, die ihren Zug in dem Bereich beenden, müssen einen erfolgreichen Geschicklichkeitsrettungswurf ablegen, um nicht 2W6 Säureschaden zu erleiden, da sie von milchigen, unweltlichen Tentakeln berührt werden.</p>
          `,
      description_en: `
        <p>You open a gateway to the dark between the stars, a region infested with unknown horrors. A 20-foot-radius sphere of blackness and bitter cold appears, centered on a point with range and lasting for the duration. This void is filled with a cacophony of soft whispers and slurping noises that can be heard up to 30 feet away. No light, magical or otherwise, can illuminate the area, and creatures fully within the area are blinded.</p>
        <p>The void creates a warp in the fabric of space, and the area is difficult terrain. Any creature that starts its turn in the area takes 2d6 cold damage. Any creature that ends its turn in the area must succeed on a Dexterity saving throw or take 2d6 acid damage as milky, otherworldly tentacles rub against it.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 150,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 3029,
      german: 'Hypnotisches Muster',
      english: 'Hypnotic Pattern',
      image: 'hypnoticPattern',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst ein wirbelndes Muster aus Farben, das sich innerhalb eines Würfels mit 9 Meter Kantenlänge in Reichweite durch die Luft schlängelt. Das Muster erscheint für einen Moment und verschwindet wieder. Jede Kreatur in diesem Bereich, die das Muster sieht, muss einen Rettungswurf auf Weisheit machen. Bei einem misslungenen Rettungswurf wird die Kreatur für die Dauer des Zaubers verzaubert. Während sie von diesem Zauber verzaubert ist, ist die Kreatur handlungsunfähig und hat eine Geschwindigkeit von 0.</p>
          <p>Der Zauber endet für eine betroffene Kreatur, wenn sie Schaden erleidet oder wenn jemand anderes eine Aktion einsetzt, um die Kreatur aus ihrer Betäubung zu schütteln.</p>
          `,
      description_en: `
        <p>You create a twisting pattern of colors that weaves through the air inside a 30-foot cube within range. The pattern appears for a moment and vanishes. Each creature in the area who sees the pattern must make a Wisdom saving throw. On a failed save, the creature becomes charmed for the duration. While charmed by this spell, the creature is incapacitated and has a speed of 0.</p>
        <p>The spell ends for an affected creature if it takes any damage or if someone else uses an action to shake the creature out of its stupor.</p>
        `,
      school: 'illusion',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 30,
    },
    {
      id: 3030,
      german: 'Intellektfestung',
      english: 'Intellect Fortress',
      image: 'intellectFortress',
      level: 3,
      isCustom: false,
      classes: ['test', 'artificer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du oder eine bereitwillige Kreatur, die du in Reichweite sehen kannst, ist für die Wirkungsdauer resistent gegen psychischen Schaden sowie bei Rettungswürfen auf Charisma, Intelligenz und Weisheit im Vorteil.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber wirkst, indem du einen Zauberplatz 4. Grades oder höher nutzt, kannst du für jeden Grad über dem 3. Grad eine zusätzliche Kreatur als Ziel wählen. Die Kreaturen müssen sich innerhalb von neun Metern voneinander befinden, wenn du sie als Ziel wählst.</p>
          `,
      description_en: `
        <p>You or a willing creature you can see within range gains resistance to psychic damage, and has advantage on saving throws against being charmed or frightened.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, you can target one additional creature for each slot level above 3rd. The creatures must be within 30 feet of each other when you target them.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3031,
      german: 'Kreuzfahrermantel',
      english: 'Crusader’s Mantle',
      image: 'crusadersMantle',
      level: 3,
      isCustom: false,
      classes: ['test', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Heilige Macht strahlt von dir in einer Aura mit einem Radius von 9 Metern aus und erweckt Kühnheit in befreundeten Kreaturen. Bis der Zauber endet, bewegt sich die Aura mit dir, zentriert auf dich. Solange du dich in der Aura befindest, fügt jede nicht feindliche Kreatur in der Aura (dich eingeschlossen) einen zusätzlichen gleißenden Schaden von 1W4 zu, wenn sie mit einem Waffenangriff trifft.</p>
          `,
      description_en: `
        <p>Holy power radiates from you in an aura with a 30-foot radius, awakening boldness in friendly creatures. Until the spell ends, the aura moves with you, centered on you. While in the aura, each nonhostile creature in the aura (including you) deals an extra 1d4 radiant damage when it hits with a weapon attack.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3032,
      german: 'Lebenstransfer',
      english: 'Life Transference',
      image: 'lifeTransference',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 4, diceType: 8 },
      description_de: `
          <p>Du opferst ein wenig deiner Gesundheit, um die Wunden einer anderen Kreatur zu heilen. Du erleidest 4W8 nekrotischen Schaden, und eine Kreatur deiner Wahl, die du innerhalb der Reichweite sehen kannst, erhält eine Anzahl an Trefferpunkten zurück, welche dem Doppelten des nekrotischen Schadens entspricht, den du erlitten hast.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Zauberplatz-Grad über dem dritten.</p>
          `,
      description_en: `
        <p>You sacrifice some of your health to mend the wounds of another creature. You take 4d8 necrotic damage, and a creature of your choice that you can see within range regains hit points equal to twice the necrotic damage you take.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d8 for each slot level above 3rd.</p>
        `,
      school: 'necromancy',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3033,
      german: 'Leomunds winzige Hütte',
      english: "Leomund's Tiny Hut",
      image: 'leomundsTinyHut',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 8,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Eine unbewegliche Kuppel aus Energie mit einem Radius von 3 m entsteht um dich herum und bleibt für die Wirkungsdauer an ihrem Standort. Der Zauber endet, wenn du den Bereich verlässt. Außer dir passen neun mittelgroße oder kleinere Kreaturen in die Kuppel. Der Zauber schlägt fehl, wenn der Bereich eine größere Kreatur oder mehr als neun beinhaltet. Kreaturen und Gegenstände, die sich beim Wirken des Zaubers in der Kuppel aufhalten, können sich frei durch sie bewegen. Allen anderen Kreaturen und Gegenständen ist es nicht möglich, sie zu durchschreiten. Zauber und andere magische Effekte können weder in die Kuppel eindringen noch durch sie hindurch gesprochen werden. Die Atmosphäre innerhalb des Bereichs ist angenehm und trocken, ungeachtet des Wetters außerhalb.</p>
          <p>Bis der Zauber endet, kannst du bestimmen, ob das Innere der Kuppel dämmrig beleuchtet oder dunkel sein soll. Die Kuppel ist von außen undurchsichtig und hat eine Farbe deiner Wahl, von innen aber ist sie transparent.</p>
          `,
      description_en: `
        <p>A 10-foot-radius immobile dome of force springs into existence around and above you and remains stationary for the duration. The spell ends if you leave its area.</p>
        <p>Nine creatures of Medium size or smaller can fit inside the dome with you. The spell fails if its area includes a larger creature or more than nine creatures. Creatures and objects⁠ within the dome when you cast this spell can move through it freely. All other creatures and objects⁠ are barred from passing through it. Spells and other magical effects can't extend through the dome or be cast through it. The atmosphere inside the space is comfortable and dry, regardless of the weather outside.</p>
        <p>Until the spell ends, you can command⁠ the interior to become dimly lit or dark. The dome is opaque from the outside, of any color you choose, but it is transparent from the inside.</p>
        `,
      school: 'evocation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3034,
      german: 'Leuchtfeuer der Hoffnung',
      english: 'Beacon of Hope',
      image: 'beaconOfHope',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber schenkt Hoffnung und Lebenskraft. Wähle eine beliebige Anzahl von Kreaturen in Reichweite. Für die Dauer des Zaubers hat jedes Ziel einen Vorteil bei Rettungswürfen auf Weisheit und vor dem Tod und erhält die maximal mögliche Anzahl an Trefferpunkten durch Heilung zurück.</p>
          `,
      description_en: `
        <p>This spell bestows hope and vitality. Choose any number of creatures within range. For the duration, each target has advantage on Wisdom saving throws and death saving throws, and regains the maximum number of hit points possible from any healing.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3035,
      german: 'Magie bannen',
      english: 'Dispel Magic',
      image: 'dispelMagic',
      level: 3,
      isCustom: false,
      classes: [
        'test',
        'bard',
        'cleric',
        'druid',
        'paladin',
        'sorcerer',
        'warlock',
        'wizard',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Wähle eine Kreatur, ein Objekt oder einen magischen Effekt in Reichweite. Jeder Zauber der 3. Grades oder niedriger auf dem Ziel endet. Für jeden Zauber der 4. Grades oder höher auf dem Ziel musst du eine Attributswurf mit deinem Zauberattribut durchführen. Der SG ist gleich 10 + Grad des Zaubers. Bei einem erfolgreichen Check endet der Zauber.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, beendest du automatisch die Effekte eines Zaubers auf das Ziel, wenn der Grad des Zaubers gleich oder niedriger ist als der Grad des verwendeten Zauberplatzes.</p>
          `,
      description_en: `
        <p>Choose one creature, object, or magical effect within range. Any spell of 3rd level or lower on the target ends. For each spell of 4th level or higher on the target, make an ability check using your spellcasting ability. The DC equals 10 + the spell’s level. On a successful check, the spell ends.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, you automatically end the effects of a spell on the target if the spell’s level is equal to or less than the level of the spell slot you used.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: false,
    },
    {
      id: 3036,
      german: 'Massen-Heilendes Wort',
      english: 'Mass Healing Word',
      image: 'massHealingWord',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric'],
      timeToCast: 1,
      cost: 'bonus',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: true,
      heal: { diceNumber: 1, diceType: 4 },
      description_de: `
          <p>Während du die Worte der Genesung rufst, gewinnen bis zu sechs Kreaturen deiner Wahl in Reichweite, die du sehen kannst, Trefferpunkte in Höhe von 1W4 + deinem Zaubermodifikator zurück. Dieser Zauber hat keine Wirkung auf Untote oder Konstrukte.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich die Heilung um 1W4 für jeden Zauberplatz-Grad über dem 3.</p>
          `,
      description_en: `
        <p>As you call out words of restoration, up to six creatures of your choice that you can see within range regain hit points equal to 1d4 + your spellcasting ability modifier. This spell has no effect on undead or constructs.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the healing increases by 1d4 for each slot level above 3.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 3037,
      german: 'Melfs präzise Meteore',
      english: "Melf's Minute Meteors",
      image: 'melfsMinuteMeteors',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 2, diceType: 6, damageType: 'fire' }],
      doesHeal: false,
      description_de: `
          <p>Du erzeugst sechs winzige Meteore in deinem Bereich. Für die Dauer des Zaubers schweben sie in der Luft und kreisen um dich herum. Wenn du den Zauber wirkst - und als eine Bonusaktion in jedem deiner Züge danach -, kannst du einen oder zwei der Meteore aufwenden, indem du sie wie der Blitz auf einen Punkt oder Punkte deiner Wahl innerhalb von 36 m von dir zurasen lässt. Sobald ein Meteor seinen Bestimmungsort erreicht oder auf einer soliden Oberfläche einschlägt, explodiert er. Jede Kreatur innerhalb von 1,50 m von dem Punkt, wo der Meteor explodiert, muss einen Geschicklichkeitsrettungswurf ablegen. Eine Kreatur erleidet 2W6 Feuerschaden bei einem Fehlschlag, die Hälfte davon bei einem Erfolg.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher verwendest, erhöht sich lie Anzahl der erzeugten Meteore um zwei für jeden Zauberjlatz-Grad über dem dritten.</p>
          `,
      description_en: `
        <p>You create six tiny meteors in your space. For the duration, they float in the air and orbit you, and as a bonus action on each of your turns before the spell ends, you can expend one or two of the meteors, sending them streaking toward a point or points you choose within 120 feet of you. Once a meteor reaches its destination or impacts against a solid surface, the meteor explodes. Each creature within 5 feet of the point where the meteor explodes must make a Dexterity saving throw. A creature takes 2d6 fire damage on a failed save, or half as much damage on a successful one.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the number of meteors created increases by two for each slot level above 3rd.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 5,
    },
    {
      id: 3038,
      german: 'Mit Pflanzen sprechen',
      english: 'Speak with Plants',
      image: 'speakWithPlants',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du gibst Pflanzen im Umkreis von 9 Metern eine begrenzte Empfindungsfähigkeit und Beweglichkeit, so dass sie mit dir kommunizieren und deinen einfachen Befehlen folgen können. Du kannst Pflanzen zu Ereignissen befragen, die sich innerhalb des letzten Tages in der Umgebung des Zaubers ereignet haben, und so Informationen über vorbeiziehende Kreaturen, das Wetter und andere Umstände erhalten.</p>
          <p>Du kannst auch schwieriges Gelände, das durch Pflanzenwuchs verursacht wird (z.B. Dickicht und Unterholz), für die Dauer des Zaubers in normales Gelände verwandeln. Oder du kannst normales Gelände, in dem Pflanzen vorhanden sind, in schwieriges Gelände verwandeln, das für die Dauer des Zaubers bestehen bleibt, so dass zum Beispiel Bäume und Äste Verfolger behindern.</p>
          <p>Pflanzen können nach dem Ermessen des Spielleiters auch andere Aufgaben für dich übernehmen. Der Zauber befähigt Pflanzen nicht, sich selbst zu entwurzeln und zu bewegen, aber sie können Äste, Ranken und Stängel frei bewegen.</p>
          <p>Wenn sich eine Pflanzenkreatur in der Nähe befindet, kannst du mit ihr kommunizieren, als hättet ihr eine gemeinsame Sprache, aber du erhältst keine magischen Fähigkeiten, um sie zu beeinflussen.</p>
          <p>Dieser Zauber kann bewirken, dass die durch den Verstrickungszauber erschaffenen Pflanzen eine gefesselte Kreatur loslassen.</p>
          `,
      description_en: `
        <p>You imbue plants within 30 feet of you with limited sentience and animation, giving them the ability to communicate with you and follow your simple commands. You can question plants about events in the spell’s area within the past day, gaining information about creatures that have passed, weather, and other circumstances.</p>
        <p>You can also turn difficult terrain caused by plant growth (such as thickets and undergrowth) into ordinary terrain that lasts for the duration. Or you can turn ordinary terrain where plants are present into difficult terrain that lasts for the duration, causing vines and branches to hinder pursuers, for example.</p>
        <p>Plants might be able to perform other tasks on your behalf at the DM’s discretion. The spell doesn’t enable plants to uproot themselves and move about, but they can freely move branches, tendrils, and stalks.</p>
        <p>If a plant creature is in the area, you can communicate with it as if you shared a common language, but you gain no magical ability to influence it.</p>
        <p>This spell can cause the plants created by the entangle spell to release a restrained creature.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3039,
      german: 'Mit Stein verschmelzen',
      english: 'Meld into Stone',
      image: 'meldIntoStone',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du trittst in ein steinernes Objekt oder eine steinerne Fläche, die groß genug ist, um deinen Körper vollständig aufzunehmen, und verschmilzt für die Dauer mit dir und deiner gesamten Ausrüstung mit dem Stein. Mit deiner Bewegungsaktion steigst du an einer Stelle, die du berühren kannst, in den Stein ein. Solange du mit dem Stein verschmolzen bist, kannst du nicht sehen, was außerhalb des Steins geschieht, und alle Würfe auf Weisheit (Wahrnehmung), mit denen du Geräusche außerhalb des Steins hörst, haben Nachteil. Du nimmst den Lauf der Zeit wahr und kannst Zauber auf dich wirken, während du mit dem Stein verschmolzen bist. Du kannst deine Bewegung nutzen, um den Stein dort zu verlassen, wo du ihn betreten hast, was den Zauber beendet. Ansonsten kannst du dich nicht bewegen.</p>
          <p>Kleiner körperlicher Schaden am Stein schadet dir nicht, aber seine teilweise Zerstörung oder eine Veränderung seiner Form (so weit, dass du nicht mehr in ihn hineinpasst) stößt dich aus und fügt dir 6W6 Wuchtschaden zu. Die vollständige Zerstörung des Steins (oder seine Verwandlung in eine andere Substanz) vertreibt dich und fügt dir 50 Wuchtschaden zu. Wenn du ausgestoßen wirst, tauchst du liegend an einem unbesetztem Platz wieder auf, die der Eintrittsstelle am nächsten ist.</p>
          `,
      description_en: `
        <p>You step into a stone object or surface large enough to fully contain your body, melding yourself and all the equipment you carry with the stone for the duration. Using your movement, you step into the stone at a point you can touch. While merged with the stone, you can’t see what occurs outside it, and any Wisdom (Perception) checks you make to hear sounds outside it are made with disadvantage. You remain aware of the passage of time and can cast spells on yourself while merged in the stone. You can use your movement to leave the stone where you entered it, which ends the spell. Otherwise, you can’t move.</p>
        <p>Minor physical damage to the stone doesn’t harm you, but its partial destruction or change in shape (to the extent that you no longer fit within it) expels you and deals 6d6 bludgeoning damage. The complete destruction of the stone (or transmutation of the stone into a different substance) expels you and deals 50 bludgeoning damage. If expelled, you fall prone in an unoccupied space closest to where you first entered.</p>
        `,
      school: 'transmutation',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3040,
      german: 'Mit Toten sprechen',
      english: 'Speak with Dead',
      image: 'speakWithDead',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du gewährst einem Leichnam deiner Wahl in Reichweite den Anschein von Leben und Intelligenz und erlaubst ihm, die von dir gestellten Fragen zu beantworten. Der Leichnam muss noch einen Mund haben und darf nicht untot sein. Der Zauber schlägt fehl, wenn der Leichnam innerhalb der letzten 10 Tage Ziel dieses Zaubers war.</p>
          <p>Bis zum Ende des Zaubers kannst du dem Leichnam bis zu fünf Fragen stellen. Der Leichnam weiß nur, was er zu Lebzeiten wusste, einschließlich der Sprachen, die er beherrschte. Die Antworten sind in der Regel kurz, kryptisch oder wiederholend, und der Leichnam ist nicht gezwungen, eine wahrheitsgemäße Antwort zu geben, wenn du ihm gegenüber feindselig bist oder er dich als Feind anerkennt. Dieser Zauber lässt die Seele der Kreatur nicht in ihren Körper zurückkehren, sondern nur ihren belebenden Geist. Daher kann der Leichnam keine neuen Informationen erfahren, er versteht nicht, was seit seinem Tod geschehen ist, und kann nicht über zukünftige Ereignisse spekulieren.</p>
          `,
      description_en: `
        <p>You grant the semblance of life and intelligence to a corpse of your choice within range, allowing it to answer the questions you pose. The corpse must still have a mouth and can’t be undead. The spell fails if the corpse was the target of this spell within the last 10 days.</p>
        <p>Until the spell ends, you can ask the corpse up to five questions. The corpse knows only what it knew in life, including the languages it knew. Answers are usually brief, cryptic, or repetitive, and the corpse is under no compulsion to offer a truthful answer if you are hostile to it or it recognizes you as an enemy. This spell doesn’t return the creature’s soul to its body, only its animating spirit. Thus, the corpse can’t learn new information, doesn’t comprehend anything that has happened since it died, and can’t speculate about future events.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 10,
      hasAreaOfEffect: false,
    },
    {
      id: 3041,
      german: 'Mächtiges Trugbild',
      english: 'Major Image',
      image: 'majorImage',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst das Trugbild eines Gegenstands, einer Kreatur oder eines anderen sichtbaren Phänomens, das nicht größer Sein kann als ein Würfel mit 6 Metern Seitenlänge. Das Bildnis erscheint an einem Ort in Reichweite, den du sehen kannst, und hält für die Wirkungsdauer an. Es erscheint vollständig real, inklusive Geräuschen, Gerüchen und Temperatur, die zu der abgebildeten Sache passt. Du kannst nicht ausreichend Hitze oder Kälte erschaffen, um Schaden zu verursachen, kein Geräusch, das laut genug ist, um Schallschaden zu verursachen, und auch keinen Geruch, der übelkeit auslösen könnte (wie der Gestank eines Troglodyten).</p>
          <p>Solange du dich in Reichweite der Illusion aufhältst, kannst du deine Aktion verwenden, um das Bildnis an einen anderen Punkt in Reichweite zu bewegen. Wenn das Bild seine Position wechselt, kannst du sein Aussehen verändern, so dass seine Bewegung natürlich für das Bildnis erscheint. Wenn du beispielsweise ein Bildnis einer Kreatur erschaffst und es bewegst, dann kannst du das Bild so verändern, dass es zu gehen Scheint. Gleichermaßen kannst du die Illusion zu verschiedenen Zeiten verschiedene Geräusche machen lassen. Du kannst sie beispielsweise sogar ein Gespräch führen lassen.</p>
          <p>Körperliche Interaktion mit dem Trugbild offenbart, dass es sich um eine Illusion handelt, weil Dinge es einfach durchdringen können. Eine Kreatur, die ihre Aktion verwendet, um das Bildnis zu untersuchen, kann erkennen, dass es sich um eine Illusion handelt, indem sie einen erfolgreichen Wurf auf Intelligenz (Nachforschungen) ablegt. Wenn eine Kreatur die Illusion als das erkennt, was sie ist, kann die Kreatur das Bildnis durchschauen, und die anderen Sensorischen Merkmale werden für die Kreatur schwach.</p>
          <p><b>Auf höheren Graden:<b> Wenn du diesen Zauber mit einem Zauberplatz des 6. Grades oder höher wirkst, dann hält der Zauber an, bis er gebannt wird und erfordert nicht mehr deine Konzentration.</p>
          `,
      description_en: `
        <p>You create the image of an object, a creature, or some other visible phenomenon that is no larger than a 20-foot cube. The image appears at a spot within range that you can see and lasts for the duration. It seems completely real, including sounds, smells, and temperature appropriate to the thing depicted. You can’t create sufficient heat or cold to cause damage, a sound loud enough to deal thunder damage or deafen a creature, or a smell that might sicken a creature (like a troglodyte’s stench).</p>
        <p>As long as you are within range of the illusion, you can use your action to cause the image to move to any other spot within range. As the image changes location, you can alter its appearance so that its movements appear natural for the image. For example, if you create an image of a creature and move it, you can alter the image so that it appears to be walking. Similarly, you can cause the illusion to make different sounds at different times, even making it carry on a conversation, for example.</p>
        <p>Physical interaction with the image reveals it to be an illusion because things can pass through it. A creature that uses its action to examine the image can determine that it is an illusion with a successful Intelligence (Investigation) check against your spell save DC. If a creature discerns the illusion for what it is, the creature can see through the image, and its other sensory qualities become faint to the creature.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 6th level or higher, the spell lasts until dispelled, without requiring your concentration.</p>
        `,
      school: 'illusion',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 20,
    },
    {
      id: 3042,
      german: 'Nahrung und Wasser erschaffen',
      english: 'Create Food and Water',
      image: 'createFoodAndWater',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erzeugst 45 Pfund Nahrung und 120 Liter Wasser auf dem Boden oder in Behältern in Reichweite, genug um bis zu fünfzehn Humanoide oder fünf Reittiere 24 Stunden lang zu versorgen. Das Essen ist fade, aber nahrhaft und verdirbt, wenn es nach 24 Stunden nicht gegessen wird. Das Wasser ist sauber und wird nicht schlecht.</p>
          `,
      description_en: `
        <p>You create 45 pounds of food and 30 gallons of water on the ground or in containers within range, enough to sustain up to fifteen humanoids or five steeds for 24 hours. The food is bland but nourishing, and spoils if uneaten after 24 hours. The water is clean and doesn’t go bad.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3043,
      german: 'Nickerchen',
      english: 'Catnap',
      image: 'catnap',
      level: 3,
      isCustom: false,
      classes: ['test', 'artificer', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: false,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>You make a calming gesture, and up to three willing creatures of your choice that you can see within range fall unconscious for the spell’s duration. The spell ends on a target early if it takes damage or someone uses an action to shake or slap it awake. If a target remains unconscious for the full duration, that target gains the benefit of a short rest, and it can’t be affected by this spell again until it finishes a long rest.</p>
          `,
      description_en: `
        <p>You make a calming gesture, and up to three willing creatures of your choice that you can see within range fall unconscious for the spell’s duration. The spell ends on a target early if it takes damage or someone uses an action to shake or slap it awake. If a target remains unconscious for the full duration, that target gains the benefit of a short rest, and it can’t be affected by this spell again until it finishes a long rest.</p>
        `,
      school: 'enchantment',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3044,
      german: 'Pflanzenwachstum',
      english: 'Plant Growth',
      image: 'plantGrowth',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Dieser Zauber kanalisiert Vitalität in Pflanzen innerhalb eines bestimmten Bereichs. Es gibt zwei Verwendungsmöglichkeiten für den Zauber, die entweder sofortigen oder langfristigen Nutzen bringen.</p>
          <p>Wenn du diesen Zauber mit 1 Aktion wirkst, wähle einen Punkt in Reichweite. Alle normalen Pflanzen in einem Radius von 30 Metern um diesen Punkt herum werden dicht und überwuchert. Eine Kreatur, die sich durch diesen Bereich bewegt, muss für jeweils 30 Zentimeter, die sie zurücklegen will, 1,20 Meter Bewegungsrate aufwenden.</p>
          <p>Du kannst ein oder mehrere Gebiete beliebiger Größe innerhalb des Wirkungsbereichs des Zaubers von der Wirkung ausschließen.</p>
          <p>Wenn du diesen Zauber über 8 Stunden hinweg wirkst, wird das Land angereichert. Alle Pflanzen in einem Radius von 750 Metern um einen Punkt innerhalb der Reichweite werden für 1 Jahr angereichert. Die Pflanzen liefern die doppelte Menge an Nahrung, wenn sie geerntet werden.</p>
          `,
      description_en: `
        <p>This spell channels vitality into plants within a specific area. There are two possible uses for the spell, granting either immediate or long-term benefits.</p>
        <p>If you cast this spell using 1 action, choose a point within range. All normal plants in a 100-foot radius centered on that point become thick and overgrown. A creature moving through the area must spend 4 feet of movement for every 1 foot it moves.</p>
        <p>You can exclude one or more areas of any size within the spell’s area from being affected.</p>
        <p>If you cast this spell over 8 hours, you enrich the land. All plants in a half-mile radius centered on a point within range become enriched for 1 year. The plants yield twice the normal amount of food when harvested.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 150,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 30,
    },
    {
      id: 3045,
      german: 'Sandwand',
      english: 'Wall of Sand',
      image: 'wallOfSand',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du beschwörst eine Wand aus aufgewirbeltem Sand auf dem Boden an einem Punkt, den du in Reichweite sehen kannst. Die Wand kann bis zu 9 Meter lang, 3 Meter hoch und 3 Meter dick sein und verschwindet, wenn der Zauber endet. Sie blockiert die Sichtlinie, aber nicht die Bewegung. Eine Kreatur ist geblendet, solange sie sich im Bereich der Wand befindet, und muss für jeden Meter, den sie sich dort bewegt, 3 Meter Bewegung aufwenden.</p>
          `,
      description_en: `
          <p>You conjure up a wall of swirling sand on the ground at a point you can see within range. You can make the wall up to 30 feet long, 10 feet high, and 10 feet thick, and it vanishes when the spell ends. It blocks line of sight but not movement. A creature is blinded while in the wall’s space and must spend 3 feet of movement for every 1 foot it moves there.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 20,
    },
    {
      id: 3046,
      german: 'Schattenbrut Beschwören',
      english: 'Summon Shadowspawn',
      image: 'summonShadowspawn',
      level: 3,
      isCustom: false,
      classes: ['test', 'sorcerer', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>You call forth a shadowy spirit. It manifests in an unoccupied space that you can see within range. This corporeal form uses the Shadow Spirit stat block. When you cast the spell, choose an emotion: Fury, Despair, or Fear. The creature resembles a misshapen biped marked by the chosen emotion, which determines certain traits in its stat block. The creature disappears when it drop to 0 hit points or when the spell ends.</p>
        <p>The creature is an ally to you and your companions. In combat, the creature shares your initiative count, but it takes its turn immediately after your. It obeys your verbal commands (no action required by you). If you don’t issue any, it takes the Dodge action and it uses its move to avoid danger.</p>
        <p><strong><em>At Higher Levels.</em></strong> When you cast the spell using a spell slot of 4th level or higher, use the higher level wherever the spell’s level appears on the stat block.</p>
        <table class="wiki-content-table">
        <tr>
        <th colspan="6">Shadow Spirit</th>
        </tr>
        <tr>
        <td colspan="6"><em>Medium monstrosity</em></td>
        </tr>
        <tr>
        <td colspan="6"><strong>Armor Class:</strong> 11 + the level of the spell (natural armor)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Hit Points:</strong> 35 + 15 for each level of the spell above 3rd</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Speed:</strong> 40 ft.</td>
        </tr>
        <tr>
        <th>STR</th>
        <th>DEX</th>
        <th>CON</th>
        <th>INT</th>
        <th>WIS</th>
        <th>CHA</th>
        </tr>
        <tr>
        <td>13 (+1)</td>
        <td>16 (+3)</td>
        <td>15 (+2)</td>
        <td>4 (−3)</td>
        <td>10 (+0)</td>
        <td>16 (+3)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Damage Resistances:</strong> necrotic</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Condition Immunities:</strong> frightened</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Senses:</strong> darkvision 120 ft., passive Perception 10</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Languages:</strong> understands the languages you speak</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Challenge:</strong> &#8212;</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Proficiency Bonus:</strong> equals your bonus</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Terror Frenzy (Fury Only).</em></strong> The spirit has advantage on attack rolls against frightened creatures.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Weight of Sorrow (Despair Only).</em></strong> Any creature, other than you, that starts its turn within 5 feet of the spirit has its speed reduced by 20 feet until the start of that creature’s next turn.</td>
        </tr>
        <tr>
        <th colspan="6">Actions</th>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Multiattack.</em></strong> The spirit makes a number of attacks equal to half this spell’s level (rounded down).</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Chilling Hand.</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one target. <em>Hit:</em> 1d12 + 3 + the spell’s level cold damage.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Dreadful Scream (1/Day).</em></strong> The spirit screams. Each creature within 30 feet of it must succeed on a Wisdom saving throw against your spell save DC or be frightened of the spirit for 1 minute. The frightened creature can repeat the saving throw at the end of each of its turns, ending the effect on itself on a success.</td>
        </tr>
        <tr>
        <th colspan="6">Bonus Action</th>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Shadow Stealth (Fear Only).</em></strong> While in dim light or darkness, the spirit takes the Hide action.</td>
        </tr>
        </table>
        `,
      description_en: `
        <p>You call forth a shadowy spirit. It manifests in an unoccupied space that you can see within range. This corporeal form uses the Shadow Spirit stat block. When you cast the spell, choose an emotion: Fury, Despair, or Fear. The creature resembles a misshapen biped marked by the chosen emotion, which determines certain traits in its stat block. The creature disappears when it drop to 0 hit points or when the spell ends.</p>
        <p>The creature is an ally to you and your companions. In combat, the creature shares your initiative count, but it takes its turn immediately after your. It obeys your verbal commands (no action required by you). If you don’t issue any, it takes the Dodge action and it uses its move to avoid danger.</p>
        <p><strong><em>At Higher Levels.</em></strong> When you cast the spell using a spell slot of 4th level or higher, use the higher level wherever the spell’s level appears on the stat block.</p>
        <table class="wiki-content-table">
        <tr>
        <th colspan="6">Shadow Spirit</th>
        </tr>
        <tr>
        <td colspan="6"><em>Medium monstrosity</em></td>
        </tr>
        <tr>
        <td colspan="6"><strong>Armor Class:</strong> 11 + the level of the spell (natural armor)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Hit Points:</strong> 35 + 15 for each level of the spell above 3rd</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Speed:</strong> 40 ft.</td>
        </tr>
        <tr>
        <th>STR</th>
        <th>DEX</th>
        <th>CON</th>
        <th>INT</th>
        <th>WIS</th>
        <th>CHA</th>
        </tr>
        <tr>
        <td>13 (+1)</td>
        <td>16 (+3)</td>
        <td>15 (+2)</td>
        <td>4 (−3)</td>
        <td>10 (+0)</td>
        <td>16 (+3)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Damage Resistances:</strong> necrotic</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Condition Immunities:</strong> frightened</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Senses:</strong> darkvision 120 ft., passive Perception 10</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Languages:</strong> understands the languages you speak</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Challenge:</strong> &#8212;</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Proficiency Bonus:</strong> equals your bonus</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Terror Frenzy (Fury Only).</em></strong> The spirit has advantage on attack rolls against frightened creatures.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Weight of Sorrow (Despair Only).</em></strong> Any creature, other than you, that starts its turn within 5 feet of the spirit has its speed reduced by 20 feet until the start of that creature’s next turn.</td>
        </tr>
        <tr>
        <th colspan="6">Actions</th>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Multiattack.</em></strong> The spirit makes a number of attacks equal to half this spell’s level (rounded down).</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Chilling Hand.</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one target. <em>Hit:</em> 1d12 + 3 + the spell’s level cold damage.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Dreadful Scream (1/Day).</em></strong> The spirit screams. Each creature within 30 feet of it must succeed on a Wisdom saving throw against your spell save DC or be frightened of the spirit for 1 minute. The frightened creature can repeat the saving throw at the end of each of its turns, ending the effect on itself on a success.</td>
        </tr>
        <tr>
        <th colspan="6">Bonus Action</th>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Shadow Stealth (Fear Only).</em></strong> While in dim light or darkness, the spirit takes the Hide action.</td>
        </tr>
        </table>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 3047,
      german: 'Schneesturm',
      english: 'Sleet Storm',
      image: 'sleetStorm',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Bis zum Ende des Zaubers fallen Eisregen und Graupel in einem 6 Meter hohen Zylinder mit einem Radius von 15 Metern, der auf einen von dir gewählten Punkt in Reichweite zentriert ist. Das Gebiet ist stark verdunkelt, und offene Flammen in dem Gebiet werden gelöscht.</p>
          <p>Der Boden in dem Gebiet ist mit glattem Eis bedeckt, was es zu schwierigen Terrain macht. Wenn eine Kreatur den Bereich des Zaubers zum ersten Mal in einem Zug betritt oder ihren Zug dort beginnt, muss sie einen Rettungswurf auf Geschicklichkeit machen. Bei einem misslungenen Rettungswurf fällt sie auf den Boden.</p>
          <p>Wenn sich eine Kreatur im Bereich des Zaubers konzentriert, muss sie einen erfolgreichen Rettungswurf auf Konstitution gegen deinen Zauberrettungswurf SG machen oder sie verliert die Konzentration.</p>
          `,
      description_en: `
        <p>Until the spell ends, freezing rain and sleet fall in a 20-foot-tall cylinder with a 40-foot radius centered on a point you choose within range. The area is heavily obscured, and exposed flames in the area are doused.</p>
        <p>The ground in the area is covered with slick ice, making it difficult terrain. When a creature enters the spell’s area for the first time on a turn or starts its turn there, it must make a Dexterity saving throw. On a failed save, it falls prone.</p>
        <p>If a creature is concentrating in the spell’s area, it must make a successful Constitution saving throw against your spell save DC or lose concentration.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 150,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cylinder',
      length: 50,
    },
    {
      id: 3048,
      german: 'Schusshagel beschwören',
      english: 'Conjure Barrage',
      image: 'conjureVolley',
      level: 3,
      isCustom: false,
      classes: ['test', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'dexterity',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'special' }],
      doesHeal: false,
      description_de: `
          <p>Du wirfst eine nicht-magische Waffe oder schießt ein Stück nicht-magischer Munition in die Luft und erschaffst einen Kegel aus identischen Waffen, die vorwärts schießen und dann verschwinden. Jede Kreatur in einem 18 Meter großen Kegel muss einen Rettungswurf auf Geschicklichkeit machen. Bei einem Fehlschlag erleidet sie 3W8 Schaden und die Hälfte bei einem Erfolg. Der Schadenstyp entspricht dem der Waffe oder Munition, die zum Beschwören benutzt wurde.</p>
          `,
      description_en: `
        <p>You throw a nonmagical weapon or fire a piece of nonmagical ammunition into the air to create a cone of identical weapons that shoot forward and then disappear. Each creature in a 60-foot cone must make a Dexterity saving throw. A creature takes 3d8 damage on a failed save, and half as much damage on a successful one. The damage type is the same as that of the weapon or ammunition used to make the attack.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cone',
      length: 60,
    },
    {
      id: 3049,
      german: 'Schutz vor Energie',
      english: 'Protection from Energy',
      image: 'protectionFromEnergy',
      level: 3,
      isCustom: false,
      classes: [
        'test',
        'artificer',
        'cleric',
        'druid',
        'ranger',
        'sorcerer',
        'wizard',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Die bereitwillige Kreatur, die du berührst, hat für die Dauer der Berührung Widerstand gegen eine Schadensart deiner Wahl: Säure, Kälte, Feuer, Blitz oder Donner.</p>
          `,
      description_en: `
        <p>For the duration, the willing creature you touch has resistance to one damage type of your choice: acid, cold, fire, lightning, or thunder.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 3050,
      german: 'Schutzgeister',
      english: 'Spirit Guardians',
      image: 'spiritGuardians',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: true,
      damage: [{ diceNumber: 3, diceType: 8, damageType: 'radiant' }],
      doesHeal: false,
      description_de: `
          <p>Du rufst Geister herbei, die dich beschützen. Sie umschwirren dich für die Dauer des Zaubers in einem Abstand von 4,5 Metern. Wenn du gut oder neutral bist, erscheint ihre Spektralform engelhaft oder feenhaft (nach deiner Wahl). Wenn du böse bist, erscheinen sie teuflisch.</p>
          <p>Wenn du diesen Zauber sprichst, kannst du eine beliebige Anzahl von Kreaturen bestimmen, die du sehen kannst und die davon nicht betroffen sind. Die Geschwindigkeit einer betroffenen Kreatur wird in dem Gebiet halbiert, und wenn die Kreatur das Gebiet zum ersten Mal in einem Zug betritt oder ihren Zug dort beginnt, muss sie einen Rettungswurf auf Weisheit ablegen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 3W8 gleißenden Schaden (wenn du gut oder neutral bist) oder 3W8 nekrotischen Schaden (wenn du böse bist). Bei einem erfolgreichen Schutzwurf erleidet die Kreatur nur halb so viel Schaden.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W8 für jeden Grad über dem Dritten.</p>
          `,
      description_en: `
        <p>You call forth spirits to protect you. They flit around you to a distance of 15 feet for the duration. If you are good or neutral, their spectral form appears angelic or fey (your choice). If you are evil, they appear fiendish.</p>
        <p>When you cast this spell, you can designate any number of creatures you can see to be unaffected by it. An affected creature’s speed is halved in the area, and when the creature enters the area for the first time on a turn or starts its turn there, it must make a Wisdom saving throw. On a failed save, the creature takes 3d8 radiant damage (if you are good or neutral) or 3d8 necrotic damage (if you are evil). On a successful save, the creature takes half as much damage.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d8 for each slot level above 3rd.</p>
        `,
      school: 'conjuration',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 15,
    },
    {
      id: 3051,
      german: 'Schutzkreis',
      english: 'Magic Circle',
      image: 'magicCircle',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
          <p>Du erschaffst einen Zylinder aus magischer Energie mit einem Radius von 3 Metern und einer Höhe von 6 Metern, der auf einen Punkt auf dem Boden in Reichweite zentriert ist, den du sehen kannst. Glühende Runen erscheinen überall dort, wo sich der Zylinder mit dem Boden oder einer anderen Oberfläche schneidet.</p>
          <ul>
          <li>Die Kreatur kann den Zylinder nicht willentlich mit nichtmagischen Mitteln betreten. Wenn die Kreatur versucht, Teleportation oder interplanare Reisen zu benutzen, um dies zu tun, muss sie zuerst einen Rettungswurf auf Charisma ablegen.</li>
          <li>Die Kreatur hat Nachteil bei Angriffswürfen gegen Ziele innerhalb des Zylinders.</li>
          <li>Ziele innerhalb des Zylinders können von der Kreatur nicht verzaubert, verängstigt oder besessen werden.</li>
          <ul>
          <p>Wenn du diesen Zauber sprichst, kannst du die Magie auch in die umgekehrte Richtung wirken lassen, so dass eine Kreatur des angegebenen Typs den Zylinder nicht verlassen kann und Ziele außerhalb des Zylinders schützt.</p>
          <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich die Dauer um 1 Stunde für jeden Zaubergrad über dem 3.</p>
          `,
      description_en: `
        <p>You create a 10-foot-radius, 20-foot-tall cylinder of magical energy centered on a point on the ground that you can see within range. Glowing runes appear wherever the cylinder intersects with the floor or other surface.</p>
        <ul>
        <li>The creature can’t willingly enter the cylinder by nonmagical means. If the creature tries to use teleportation or interplanar travel to do so, it must first succeed on a Charisma saving throw.</li>
        <li>The creature has disadvantage on attack rolls against targets within the cylinder.</li>
        <li>Targets within the cylinder can’t be charmed, frightened, or possessed by the creature.</li>
        </ul>
        <p>When you cast this spell, you can elect to cause its magic to operate in the reverse direction, preventing a creature of the specified type from leaving the cylinder and protecting targets outside it.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the duration increases by 1 hour for each slot level above 3.</p>
        `,
      school: 'abjuration',
      isRanged: true,
      range: 10,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cylinder',
      length: 10,
    },
    {
      id: 3052,
      german: 'Stinkende Wolke',
      english: 'Stinking Cloud',
      image: 'stinkingCloud',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'constitution',
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Du erschaffst eine gelbe, ekelerregende Gaswolke mit einem Radius von 6 Metern, die sich auf einen Punkt in Reichweite konzentriert. Die Wolke breitet sich um Ecken herum aus, und das Gebiet ist stark vernebelt. Jede Kreatur, die sich zu Beginn ihres Zuges vollständig innerhalb der Wolke befindet, muss einen Rettungswurf gegen Gift machen. Bei einem misslungenen Rettungswurf verbringt die Kreatur ihre Aktion in dieser Runde mit Würgen und Taumeln. Kreaturen, die nicht atmen müssen oder immun gegen Gift sind, schaffen diesen Rettungswurf automatisch.</p>
      <p>Ein mäßiger Wind (mindestens 15 Kilometer pro Stunde) löst die Wolke nach 4 Runden auf. Ein starker Wind (mindestens 30 Kilometer pro Stunde) zerstreut sie nach 1 Runde.</p>
      `,
      description_en: `
        <p>You create a 20-foot-radius sphere of yellow, nauseating gas centered on a point within range. The cloud spreads around corners, and its area is heavily obscured. The cloud lingers in the air for the duration. Each creature that is completely within the cloud at the start of its turn must make a Constitution saving throw against poison. On a failed save, the creature spends its action that turn retching and reeling. Creatures that don’t need to breathe or are immune to poison automatically succeed on the saving throw.</p>
        <p>A moderate wind (at least 10 miles per hour) disperses the cloud after 4 rounds. A strong wind (at least 20 miles per hour) disperses it after 1 round.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 20,
    },
    {
      id: 3053,
      german: 'Tageslicht',
      english: 'Daylight',
      image: 'daylight',
      level: 3,
      isCustom: false,
      classes: [
        'test',
        'cleric',
        'druid',
        'paladin',
        'ranger',
        'sorcerer',
        'wizard',
      ],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
      <p>Eine Lichtkugel mit einem Radius von 18 Metern breitet sich von einem von dir gewählten Punkt in Reichweite aus. Die Sphäre ist helles Licht und wirft für weitere 90 Meter schummriges Licht ab.</p>
      <p>Wenn du einen Punkt auf einem Objekt wählst, das du in der Hand hältst oder das nicht getragen wird, strahlt das Licht vom Objekt aus und bewegt sich mit ihm. Wenn du das betroffene Objekt vollständig mit einem undurchsichtigen Gegenstand, wie einer Schale oder einem Helm, abdeckst, wird das Licht blockiert.</p>
      <p>Wenn sich ein Bereich dieses Zaubers mit einem Bereich der Dunkelheit überschneidet, der durch einen Zauber der 3. Stufe oder niedriger erzeugt wurde, wird der Zauber, der die Dunkelheit erzeugt hat, gebannt.</p>
      `,
      description_en: `
        <p>A 60-foot-radius sphere of light spreads out from a point you choose within range. The sphere is bright light and sheds dim light for an additional 60 feet.</p>
        <p>If you choose a point on an object you are holding or one that isn’t being worn or carried, the light shines from the object and moves with it. Completely covering the affected object with an opaque object, such as a bowl or a helm, blocks the light.</p>
        <p>If any of this spell’s area overlaps with an area of darkness created by a spell of 3rd level or lower, the spell that created the darkness is dispelled.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'sphere',
      length: 60,
    },
    {
      id: 3054,
      german: 'Tiere beschwören',
      english: 'Conjure Animals',
      image: 'conjureAnimals',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'ranger'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du beschwörst Feengeister herauf, die die Form von Tieren annehmen und an freien Stellen in Reichweite, die du sehen kannst, erscheinen. Wähle eine der folgenden Optionen:</p>
        <ul>
        <li>Ein Tier von Herausforderungsstufe 2 oder niedriger</li>
        <li>Zwei Tiere von Herausforderungsstufe 1 oder niedriger</li>
        <li>Vier Tiere von Herausforderungsstufe 1/2 oder niedriger</li>
        <li>Acht Tiere von Herausforderungsstufe 1/4 oder niedriger</li>
        </ul>
        <p>Jedes Tier gilt als Feenkreatur und verschwindet, wenn es auf 0 TP sinkt oder der Zauber endet.</p>
        <p>Die beschworenen Kreaturen sind dir und deinen Gefährten gegenüber freundlich gesinnt. Würfel Initiative für die Gruppe der Kreaturen, die dann ihren eigenen Platz in der Kampfordnung bekommt. Sie folgen deinen verbalen Befehlen und verteidigen sich gegen feindliche Kreaturen, wenn sie keinen Befehl von dir erhalten, agieren ansonsten aber nicht selbstständig.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit bestimmten höheren Zauberplätzen sprichst, erscheinen mehr Kreaturen. Doppelt so viele für den 5. Grad, dreimal so viele für den 7. Grad und viermal so viele für den 9. Grad.</p>
        `,
      description_en: `
        <p>You summon fey spirits that take the form of beasts and appear in unoccupied spaces that you can see within range. Choose one of the following options:</p>
        <ul>
        <li>One beast of challenge rating 2 or lower</li>
        <li>Two beasts of challenge rating 1 or lower</li>
        <li>Four beasts of challenge rating 1/2 or lower</li>
        <li>Eight beasts of challenge rating 1/4 or lower</li>
        </ul>
        <p>Each beast is considered fey and disappears when it drops to 0 hit points or when the spell ends.</p>
        <p>The summoned creatures are friendly to you and your companions. Roll initiative for the summoned creatures as a group, which has its own turns. They obey any verbal commands that you issue to them (no action required by you). If you don’t issue any commands to them, they defend themselves from hostile creatures, but otherwise take no actions.</p>
        <p><b>At higher levels:</b> When you cast this spell using certain higher-level spell slots, you choose one of the summoning options above, and more creatures appear: twice as many with a 5th-level slot, three times as many with a 7th-level slot, and four times as many with a 9th-level slot.</p>
        `,
      school: 'conjuration',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 3055,
      german: 'Tote beleben',
      english: 'Animate Dead',
      image: 'animateDead',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'wizard'],
      timeToCast: 1,
      cost: 'minutes',
      duration: 24,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Der Zauber erschafft einen untoten Diener. Wähle einen Knochenhaufen oder einen Leichnam eines mittelgroßen oder kleinen Humanoiden in Reichweite. Dein Zauber erfüllt das Ziel mit einer verderbten Nachahmung des Lebens und lässt sie sich als untote Kreatur erheben. Das Ziel wird ein Skelett, wenn du Knochen ausgewählt hast, oder ein Zombie, wenn du dich für eine Leiche entschieden hast (der Spielleiter hat die Spielwerte der Kreatur).</p>
        <p>In jedem deiner Züge kannst du eine Bonusaktion verwenden, um geistig jegliche Kreaturen, die du mit diesem Zauber erschaffen hast, zu befehligen, wenn sie sich im Umkreis von 18 Metern um dich befinden (wenn du mehrere Kreaturen kontrollierst, kannst du sie entweder einzeln befehligen oder alle auf einmal, wobei du ihnen allen den gleichen Befehl geben musst). Du entscheidest, welche Aktion die Kreatur in ihrem nächsten Zug ausführt und wohin sie sich bewegt, oder du gibst einen allgemeinen Befehl, wie das Bewachen eines bestimmten Raumes oder Flurs. Wenn du keine Befehle erteilst, verteidigt sich die Kreatur nur gegen feindliche Kreaturen). Sobald die Kreatur einen Befehl erhalten hat, führt sie ihn weiter aus, bis die Aufgabe abgeschlossen ist.</p>
        <p>Die Kreatur steht für 24 Stunden unter deiner Kontrolle, danach hört sie auf den Befehlen zu folgen, die du ihr gegeben hast. Um die Kontrolle über die Kreatur für weitere 24 Stunden zu behalten, musst du diesen Zauber erneut auf die Kreatur wirken, ehe der aktuelle Zeitraum von 24 Stunden vorüber ist. Der Einsatz dieses Zaubers stellt deine Kontrolle über bis zu vier Kreaturen wieder her, die du mit diesem Zauber belebt hast, anstelle eine neue zu beleben.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, dann kannst du zwei zusätzliche untote Kreaturen für jeden Zauberplatz-Grad über dem dritten erheben oder wieder unter deine Kontrolle bringen. Jede Kreatur muss aus einer eigenen Leiche oder einem eigenen Knochenhaufen entstehen.</p>
        `,
      description_en: `
        <p>This spell creates an undead servant. Choose a pile of bones or a corpse of a Medium or Small humanoid within range. Your spell imbues the target with a foul mimicry of life, raising it as an undead creature. The target becomes a skeleton if you chose bones or a zombie if you chose a corpse (the DM has the creature’s game statistics).</p>
        <p>In each of your turns, you can use a bonus action to mentally command any creature you made with this spell if the creature is within 60 feet of you (if you control multiple creatures, you can command any or all of them at the same time, issuing the same command to each one). You decide what action the creature will take and where it will move during its next turn, or you can issue a general command, such as to guard a particular chamber or corridor. If you issue no commands, the creature only defends itself against hostile creatures). Once given an order, the creature continues to follow it until its task is complete.</p>
        <p>The creature is under your control for 24 hours, after which it stops obeying any command you’ve given it. To maintain control of the creature for another 24 hours, you must cast this spell on the creature again before the current 24-hour period ends. This use of the spell reasserts your control over up to four creatures you have animated with this spell, rather than animating a new one.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, you can animate or reassert control over two additional undead creatures for each slot level above 3rd. Each of the creatures must come from a different corpse or pile of bones.</p>
        `,
      school: 'necromancy',
      isRanged: true,
      range: 10,
      hasAreaOfEffect: false,
    },
    {
      id: 3056,
      german: 'Totstellen',
      english: 'Feign Death',
      image: 'feignDeath',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'druid', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du berührst eine bereitwillige Kreatur und versetzt sie in einen reglosen Zustand, der nicht vom Tod zu unterscheiden ist.</p>
        <p>Für die Dauer des Zaubers oder bis du eine Aktion verwendest, um das Ziel zu berühren und den Zauber aufzuheben, erscheint das Ziel für alle äußeren Blicke und für Zauber, die zur Bestimmung des Status des Ziels verwendet werden, tot. Das Ziel ist geblendet und handlungsunfähig und seine Geschwindigkeit sinkt auf 0. Das Ziel ist resistent gegen jeglichen Schaden außer psychischem Schaden. Wenn das Ziel krank oder vergiftet ist, wenn du den Zauber sprichst, oder während der Wirkung des Zaubers krank oder vergiftet wird, haben die Krankheit und das Gift keine Wirkung, bis der Zauber endet.</p>
        `,
      description_en: `
        <p>You touch a willing creature and put it into a cataleptic state that is indistinguishable from death.</p>
        <p>For the spell’s duration, or until you use an action to touch the target and dismiss the spell, the target appears dead to all outward inspection and to spells used to determine the target’s status. The target is blinded and incapacitated, and its speed drops to 0. The target has resistance to all damage except psychic damage. If the target is diseased or poisoned when you cast the spell or becomes diseased or poisoned while under the spell’s effect, the disease and poison have no effect until the spell ends.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 3057,
      german: 'Unauffindbarkeit',
      english: 'Nondetection',
      image: 'nondetection',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 8,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Für die Zauberdauer versteckst du ein Ziel vor Erkenntnismagie. Du berührst ein Ziel, das entweder eine bereitwillige Kreatur oder eine Stelle oder ein Objekt, nicht größer als 3 Meter in den drei Dimensionen, sein muss. Dein Ziel kann weder von Erkenntniszaubern anvisiert, noch durch Kristallomantie ausgespäht werden.</p>
        `,
      description_en: `
        <p>For the duration, you hide a target that you touch from divination magic. The target can be a willing creature or a place or an object no larger than 10 feet in any dimension. The target can’t be targeted by any divination magic or perceived through magical scrying sensors.</p>
        `,
      school: 'abjuration',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 3058,
      german: 'Untoten beschwören',
      english: 'Summon Undead',
      image: 'summonUndead',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du rufst einen untoten Geist herbei. Er erscheint in einem freien Bereich, den du in Reichweite sehen kannst. Seine physische Form nutzt die Spielwerte aus dem Wertekasten des untoten Geists. Wenn du diesen Zauber wirkst, wähle eine folgender Gestalten für die Kreatur: Geisterhaft, Skelett oder Verwesend. Der Geist ähnelt einem Untoten in der von dir gewählten Gestalt, wodurch einige seiner Merkmale bestimmt werden (siehe Wertekasten). Die Kreatur verschwindet, wenn der Zauber endet oder ihre Trefferpunkte auf 0 sinken.</p>
        <p>Die Kreatur ist mit dir und deinen Gefährten verbündet. Im Kampf nutzt die Kreatur deine Initiative und ist unmittelbar nach dir am Zug. Sie gehorcht deinen mündlichen Befehlen (erfordert keine Aktion). Befiehlst du ihr nichts, führt sie die Ausweichaktion aus und vermeidet mit ihrer Bewegung Gefahren.</p>
        <p><strong><em>Auf höheren Graden:</em></strong> Wenn du diesen Zauber mit einem Zauberplatz des 4. oder eines höheren Grades wirkst, verwende den höheren Zaubergrad, wo immer er im Wertekasten auftaucht.</p>
        <table class="wiki-content-table">
        <tr>
        <th colspan="6">Undead Spirit</th>
        </tr>
        <tr>
        <td colspan="6"><em>Medium undead</em></td>
        </tr>
        <tr>
        <td colspan="6"><strong>Armor Class:</strong> 11 + the level of the spell (natural armor)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Hit Points:</strong> 30 (Ghostly and Putrid only) or 20 (Skeletal only) + 10 for each level of the spell above 3rd</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Speed:</strong> 30 ft., fly 40 ft. (hover) (Ghostly only)</td>
        </tr>
        <tr>
        <th>STR</th>
        <th>DEX</th>
        <th>CON</th>
        <th>INT</th>
        <th>WIS</th>
        <th>CHA</th>
        </tr>
        <tr>
        <td>12 (+1)</td>
        <td>16 (+3)</td>
        <td>15 (+2)</td>
        <td>4 (−3)</td>
        <td>10 (+0)</td>
        <td>9 (−1)</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Damage Immunities:</strong> necrotic, poison</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Condition Immunities:</strong> exhaustion, frightened, paralyzed, poisoned</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Senses:</strong> darkvision 60 ft., passive Perception 10</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Languages:</strong> understands the languages you speak</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Challenge:</strong> &#8212;</td>
        </tr>
        <tr>
        <td colspan="6"><strong>Proficiency Bonus:</strong> equals your bonus</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Festering Aura (Putrid Only).</em></strong> Any creature, other than you, that starts its turn within 5 feet of the spirit must succeed on a Constitution saving throw against your spell save DC or be poisoned until the start of its next turn.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Incorporeal Passage (Ghostly Only).</em></strong> The spirit can move through other creatures and objects as if they were difficult terrain. If it ends its turn inside an objects, it is shunted to the nearest unoccupied space and takes 1d10 force damage for every 5 feet traveled.</td>
        </tr>
        <tr>
        <th colspan="6">Actions</th>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Multiattack.</em></strong> The spirit makes a number of attacks equal to half this spell’s level (rounded down).</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Deathly Touch (Ghostly Only).</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one creature. <em>Hit:</em> 1d8 + 3 + the spell’s level necrotic damage, and the creature must succeed on a Wisdom saving throw against your spell save DC or be frightened of the undead until the end of the target’s next turn.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Grave Bolt (Skeletal Only).</em></strong> <em>Ranged Spell Attack:</em> your spell attack modifier to hit, range 150 ft., one target. <em>Hit:</em> 2d4 + 3 + the spell’s level necrotic damage.</td>
        </tr>
        <tr>
        <td colspan="6"><strong><em>Rotting Claw (Putrid Only).</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one target. <em>Hit:</em> 1d6 + 3 + the spell’s level slashing damage. If the target is poisoned, it must succeed on a Constitution saving throw against your spell save DC or be paralyzed until the end of its next turn.</td>
        </tr>
        </table>
        `,
      description_en: `
      <p>You call forth an undead spirit. It manifests in an unoccupied space that you can see within range. This corporeal form uses the Undead Spirit stat block. When you cast the spell, choose the creature’s form: Ghostly, Putrid, or Skeletal. The spirit resembles an undead creature with the chosen form, which determines certain traits in its stat block. The creature disappears when it drops to 0 hit points or when the spell ends.</p>
      <p>The creature is an ally to you and your companions. In combat, the creature shares your initiative count, but it takes its turn immediately after yours. It obeys your verbal commands (no action required by you). If you don’t issue any, it takes the Dodge action and uses its move to avoid danger.</p>
      <p><strong><em>At Higher Levels.</em></strong> When you cast this spell using a spell slot of 4th level or higher, use the higher level wherever the spell’s level appears in the stat block.</p>
      <p><strong><em>Spell Lists.</em></strong> <a href="/spells:warlock">Warlock (Optional)</a>, <a href="/spells:wizard">Wizard (Optional)</a></p>
      <table class="wiki-content-table">
      <tr>
      <th colspan="6">Undead Spirit</th>
      </tr>
      <tr>
      <td colspan="6"><em>Medium undead</em></td>
      </tr>
      <tr>
      <td colspan="6"><strong>Armor Class:</strong> 11 + the level of the spell (natural armor)</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Hit Points:</strong> 30 (Ghostly and Putrid only) or 20 (Skeletal only) + 10 for each level of the spell above 3rd</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Speed:</strong> 30 ft., fly 40 ft. (hover) (Ghostly only)</td>
      </tr>
      <tr>
      <th>STR</th>
      <th>DEX</th>
      <th>CON</th>
      <th>INT</th>
      <th>WIS</th>
      <th>CHA</th>
      </tr>
      <tr>
      <td>12 (+1)</td>
      <td>16 (+3)</td>
      <td>15 (+2)</td>
      <td>4 (−3)</td>
      <td>10 (+0)</td>
      <td>9 (−1)</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Damage Immunities:</strong> necrotic, poison</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Condition Immunities:</strong> exhaustion, frightened, paralyzed, poisoned</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Senses:</strong> darkvision 60 ft., passive Perception 10</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Languages:</strong> understands the languages you speak</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Challenge:</strong> &#8212;</td>
      </tr>
      <tr>
      <td colspan="6"><strong>Proficiency Bonus:</strong> equals your bonus</td>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Festering Aura (Putrid Only).</em></strong> Any creature, other than you, that starts its turn within 5 feet of the spirit must succeed on a Constitution saving throw against your spell save DC or be poisoned until the start of its next turn.</td>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Incorporeal Passage (Ghostly Only).</em></strong> The spirit can move through other creatures and objects as if they were difficult terrain. If it ends its turn inside an objects, it is shunted to the nearest unoccupied space and takes 1d10 force damage for every 5 feet traveled.</td>
      </tr>
      <tr>
      <th colspan="6">Actions</th>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Multiattack.</em></strong> The spirit makes a number of attacks equal to half this spell’s level (rounded down).</td>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Deathly Touch (Ghostly Only).</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one creature. <em>Hit:</em> 1d8 + 3 + the spell’s level necrotic damage, and the creature must succeed on a Wisdom saving throw against your spell save DC or be frightened of the undead until the end of the target’s next turn.</td>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Grave Bolt (Skeletal Only).</em></strong> <em>Ranged Spell Attack:</em> your spell attack modifier to hit, range 150 ft., one target. <em>Hit:</em> 2d4 + 3 + the spell’s level necrotic damage.</td>
      </tr>
      <tr>
      <td colspan="6"><strong><em>Rotting Claw (Putrid Only).</em></strong> <em>Melee Weapon Attack:</em> your spell attack modifier to hit, reach 5 ft., one target. <em>Hit:</em> 1d6 + 3 + the spell’s level slashing damage. If the target is poisoned, it must succeed on a Constitution saving throw against your spell save DC or be paralyzed until the end of its next turn.</td>
      </tr>
      </table>
      `,
      school: 'necromancy',
      isRanged: true,
      range: 90,
      hasAreaOfEffect: false,
    },
    {
      id: 3059,
      german: 'Vampirgriff',
      english: 'Vampiric Touch',
      image: 'vampiricTouch',
      level: 3,
      isCustom: false,
      classes: ['test', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: false,
      needsAttackRoll: true,
      needsSavingThrow: false,
      doesDamage: true,
      damage: [{ diceType: 6, diceNumber: 3, damageType: 'necrotic' }],
      doesHeal: false,
      description_de: `
        <p>Die Berührung deiner schattenumrankten Hand kann anderen die Lebenskraft entziehen, um deine Wunden zu heilen. Führe einen Nahkampf-Zauberangriff gegen eine Kreatur in deiner Reichweite aus. Bei einem Treffer erleidet das Ziel 3W6 nekrotischen Schaden, und du erhältst Trefferpunkte in Höhe der Hälfte des zugefügten nekrotischen Schadens zurück. Bis der Zauber endet, kannst du den Angriff in jeder deiner Runden als Aktion wiederholen.</p>
        <p><b>Auf höheren Graden:</b> Wenn du diesen Zauber mit einem Zauberplatz des 4. Grades oder höher wirkst, erhöht sich der Schaden um 1W6 für jeden Grad über dem 3.</p>
        `,
      description_en: `
        <p>Your touch inflicts necrotic damage. The target must make a Constitution saving throw. The target takes 3d6 necrotic damage on a failed save, or half as much damage on a successful one. You regain hit points equal to half the amount of necrotic damage dealt. Until the spell ends, you can make the attack again on each of your turns as an action.</p>
        <p><b>At higher levels:</b> When you cast this spell using a spell slot of 4th level or higher, the damage increases by 1d6 for each slot level above 3.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 0,
      hasAreaOfEffect: false,
    },
    {
      id: 3060,
      german: 'Verlangsamen',
      english: 'Slow',
      image: 'slow',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'wisdom',
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du veränderst die Zeit um bis zu sechs Kreaturen deiner Wahl in einem Würfel mit einer Kantenlänge von 15 Metern in Reichweite.Jedes Ziel muss einen Rettungswurf in Weisheit bestehen oder wird für die Dauer des Zaubers von diesem beeinflusst.</p>
        <p>Die Geschwindigkeit eines betroffenen Ziels wird halbiert, es erleidet einen Malus von -2 auf die Rüstungsklasse und Rettungswürfe auf Geschicklichkeit und kann keine Reaktionen verwenden. In seinem Zug kann es lediglich entweder eine Aktion oder eine Bonusaktion einsetzen, nicht beides. Unabhängig von den Fähigkeiten oder magischen Gegenständen der Kreatur kann sie während ihres Zuges nicht mehr als einen Nah- oder Fernkampfangriff ausführen.</p>
        <p>Wenn die Kreatur versucht, einen Zauber mit einer Wirkzeit von 1 Aktion zu wirken, wirf einen W20. Bei einer 11 oder höher wird der Zauber erst in der nächsten Runde der Kreatur wirksam, und die Kreatur muss ihre Aktion in dieser Runde nutzen, um den Zauber zu vollenden. Gelingt ihr das nicht, ist der Zauber vergeudet.</p>
        <p>Eine Kreatur, die von diesem Zauber betroffen ist, macht am Ende ihres Zuges einen weiteren Rettungswurf in Weisheit. Bei einem erfolgreichen Schutz endet der Effekt für sie.</p>
        `,
      description_en: `
        <p>You alter time around up to six creatures of your choice in a 40-foot cube within range. Each target must succeed on a Wisdom saving throw or be affected by this spell for the duration.</p>
        <p>An affected target’s speed is halved, it takes a −2 penalty to AC and Dexterity saving throws, and it can’t use reactions. On its turn, it can use either an action or a bonus action, not both. Regardless of the creature’s abilities or magic items, it can’t make more than one melee or ranged attack during its turn.</p>
        <p>If the creature attempts to cast a spell with a casting time of 1 action, roll a d20. On an 11 or higher, the spell doesn’t take effect until the creature’s next turn, and the creature must use its action on that turn to complete the spell. If it can’t, the spell is wasted.</p>
        <p>A creature affected by this spell makes another Wisdom saving throw at the end of its turn. On a successful save, the effect ends for it.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'cube',
      length: 50,
    },
    {
      id: 3061,
      german: 'Verständigung',
      english: 'Sending',
      image: 'sending',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'rounds',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: false,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du sendest eine kurze Nachricht von höchstens fünfundzwanzig Wörtern an ein dir bekanntes Wesen. Die Kreatur hört die Nachricht in ihrem Geist, erkennt dich als Absender, wenn sie dich kennt, und kann sofort auf die gleiche Weise antworten. Der Zauber ermöglicht es Kreaturen mit einem Intelligenzwert von mindestens 1, die Bedeutung deiner Nachricht zu verstehen.</p>
        <p>Du kannst die Nachricht über eine beliebige Entfernung und sogar in andere Existenzebenen schicken, aber wenn sich das Ziel auf einer anderen Ebene als du befindet, besteht eine 5 %ige Chance, dass die Nachricht nicht ankommt.</p>
        `,
      description_en: `
        <p>You send a short message of twenty-five words or less to a creature with which you are familiar. The creature hears the message in its mind, recognizes you as the sender if it knows you, and can answer in a like manner immediately. The spell enables creatures with Intelligence scores of at least 1 to understand the meaning of your message.</p>
        <p>You can send the message across any distance and even to other planes of existence, but if the target is on a different plane than you, there is a 5 percent chance that the message doesn’t arrive.</p>
        `,
      school: 'evocation',
      isRanged: true,
      range: 9999,
      hasAreaOfEffect: false,
    },
    {
      id: 3062,
      german: 'Wasser atmen',
      english: 'Water Breathing',
      image: 'waterBreathing',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'cleric', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 24,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: true,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Dieser Zauber gewährt bis zu zehn bereitwilligen Kreaturen in Reichweite, die du siehst, die Fähigkeit, unter Wasser zu atmen, bis der Zauber endet. Die betroffenen Kreaturen behalten auch ihre normale Atmungsweise bei.</p>
        `,
      description_en: `
        <p>This spell grants up to ten willing creatures you can see within range the ability to breathe underwater until the spell ends. Affected creatures also retain their normal mode of respiration.</p>
        `,
      school: 'transmutation',
      isRanged: true,
      range: 30,
      hasAreaOfEffect: false,
    },
    {
      id: 3063,
      german: 'Wasserwand',
      english: 'Wall of Water',
      image: 'wallOfWater',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 10,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du beschwörst eine Wand aus Wasser auf dem Boden an einem Punkt in Reichweite, den du sehen kannst. Die Wand kann bis zu 9 Meter lang, 3 Meter hoch und 30 Zentimeter dick sein, oder du kannst eine ringförmige Wand mit einem Durchmesser von 6 Metern, 3 Metern Höhe und 30 Zentimetern Dicke erschaffen. Die Mauer verschwindet, wenn der Zauber endet. Jeder Angriff mit einer Fernkampfwaffe, der den Raum der Mauer betritt, hat einen Nachteil beim Angriffswurf, und der Feuerschaden wird halbiert, wenn der Feuereffekt die Mauer durchquert, um sein Ziel zu erreichen. Zauber, die Kälteschaden verursachen und die Wand durchqueren, lassen den Bereich der Wand, den sie durchqueren, fest gefrieren (mindestens ein 1,5-Meter-Quadratbereich wird gefroren). Jeder gefrorene Abschnitt von 1,5 Metern im Quadrat hat RK 5 und 15 Trefferpunkte. Wird ein gefrorener Abschnitt auf 0 Trefferpunkte reduziert, wird er zerstört. Wenn ein Abschnitt zerstört wird, füllt sich die Wand nicht mehr mit Wasser.</p>
          `,
      description_en: `
        <p>You conjure up a wall of water on the ground at a point you can see within range. You can make the wall up to 30 feet long, 10 feet high, and 1 foot thick, or you can make a ringed wall up to 20 feet in diameter, 20 feet high, and 1 foot thick. The wall vanishes when the spell ends. The wall’s space is difficult terrain.</p>
        <p>Any ranged weapon attack that enters the wall’s space has disadvantage on the attack roll, and fire damage is halved if the fire effect passes through the wall to reach its target. Spells that deal cold damage that pass through the wall cause the area of the wall they pass through to freeze solid (at least a 5-foot square section is frozen). Each 5-foot-square frozen section has AC 5 and 15 hit points. Reducing a frozen section to 0 hit points destroys it. When a section is destroyed, the wall’s water doesn’t fill it.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 30,
    },
    {
      id: 3064,
      german: 'Wiederbeleben',
      english: 'Revivify',
      image: 'revivify',
      level: 3,
      isCustom: false,
      classes: ['test', 'cleric', 'paladin'],
      timeToCast: 1,
      cost: 'action',
      duration: 0,
      durationType: 'instant',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Du berührst eine Kreatur, die innerhalb der letzten Minute gestorben ist. Diese Kreatur erwacht mit 1 Trefferpunkt wieder zum Leben. Dieser Zauber kann weder eine Kreatur wiederbeleben, die an Altersschwäche gestorben ist, noch kann er fehlende Körperteile wiederherstellen.</p>
        `,
      description_en: `
        <p>You touch a creature that has died within the last minute. That creature returns to life with 1 hit point. This spell can’t return to life a creature that has died of old age, nor can it restore any missing body parts.</p>
        `,
      school: 'necromancy',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },
    {
      id: 3065,
      german: 'Windwall',
      english: 'Wind Wall',
      image: 'windWall',
      level: 3,
      isCustom: false,
      classes: ['test', 'druid', 'ranger', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'minutes',
      needsConcentration: true,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: true,
      savingThrowAttribute: 'strength',
      doesDamage: true,
      damage: [{ diceType: 8, diceNumber: 3, damageType: 'bludgeoning' }],
      doesHeal: false,
      description_de: `
        <p>Eine Wand aus starkem Wind erhebt sich an einem von dir gewählten Punkt in Reichweite aus dem Boden. Die Wand kann bis zu 15 Meter lang, 5 Meter hoch und 30 Zentimeter dick sein. Du kannst die Wand auf jede beliebige Weise formen, solange sie einen durchgehenden Pfad entlang des Bodens bildet. Die Mauer bleibt für die Dauer des Zaubers bestehen.</p>
        <p>Wenn die Mauer erscheint, muss jede Kreatur in ihrem Bereich einen Rettungswurf auf Stärke machen. Bei einem misslungenen Rettungswurf erleidet die Kreatur 3W8 Wuchtschaden, bei einem erfolgreichen Rettungswurf die Hälfte.</p>
        <p>Der starke Wind hält Nebel, Rauch und andere Gase in Schach. Kleine oder kleinere fliegende Kreaturen oder Objekte können die Wand nicht durchdringen. Lose, leichte Materialien, die in die Wand gebracht werden, fliegen nach oben. Pfeile, Bolzen und andere gewöhnliche Geschosse, die auf Ziele hinter der Mauer abgefeuert werden, werden nach oben abgelenkt und verfehlen automatisch. (Felsbrocken, die von Riesen oder Belagerungsmaschinen geschleudert werden, und ähnliche Geschosse sind davon nicht betroffen). Kreaturen in gasförmiger Form können die Mauer nicht durchdringen.</p>
        `,
      description_en: `
        <p>A wall of strong wind rises from the ground at a point you choose within range. You can make the wall up to 50 feet long, 15 feet high, and 1 foot thick. You can shape the wall in any way you choose so long as it makes a continuous path along the ground. The wall lasts for the duration.</p>
        <p>When the wall appears, each creature within its area must make a Strength saving throw. A creature takes 3d8 bludgeoning damage on a failed save, or half as much damage on a successful one.</p>
        <p>The strong wind keeps fog, smoke, and other gases at bay. Small or smaller flying creatures or objects can’t pass through the wall. Loose, lightweight materials brought into the wall fly upward. Arrows, bolts, and other ordinary projectiles launched at targets behind the wall are deflected upward and automatically miss. (Boulders hurled by giants or siege engines, and similar projectiles, are unaffected). Creatures in gaseous form can’t pass through the wall.</p>
      `,
      school: 'evocation',
      isRanged: true,
      range: 120,
      hasAreaOfEffect: true,
      areaOfEffectType: 'line',
      length: 50,
    },
    {
      id: 3066,
      german: 'Winziger Diener',
      english: 'Tiny Servant',
      image: 'tinyServant',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'warlock', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Der Zauber erschafft eine unsichtbare, geistlose, formlose Kraft, die auf dein Geheiß einfache Aufgaben erfüllt bis der Zauber endet. Der Diener erscheint in einem nicht besetzten Bereich am Boden innerhalb der Reichweite. Er hat RK 10, 1 Trefferpunkt, eine Stärke von 2 und kann nicht angreifen. Wenn er auf 0 TP fällt, endet der Zauber.</p>
        <p>Einmal in jedem deiner Züge kannst du als Bonusaktion dem Diener den geistigen Befehl geben, sich bis zu 4,50 Meter zu bewegen und mit einem Gegenstand zu interagieren. Der Diener kann einfache Aufgaben übernehmen, die ein menschlicher Diener leisten könnte. Er kann Dinge holen, reinigen und reparieren, Kleider zusammenlegen, Feuer anzünden, Essen servieren, und Wein einschenken. Sobald du den Befehl erteilst, erfüllt der Diener sie nach besten Möglichkeiten bis er die Aufgabe erfüllt. Dann wartet er auf den nächsten Befehl.</p>
        <p>Wenn du dem Diener befiehlst, eine Aufgabe zu erfüllen, für die er sich mehr als 18 Meter von dir entfernen müsste, dann endet der Zauber.</p>
        `,
      description_en: `
        <p>You create an invisible, mindless, shapeless force that performs simple tasks at your command until the spell ends. The servant springs into existence in an unoccupied space on the ground within range. It has AC 10, 1 hit point, and a Strength of 2, and it can’t attack. If it drops to 0 hit points, the spell ends.</p>
        <p>Once on each of your turns as a bonus action, you can mentally command the servant to move up to 15 feet and interact with an object. The servant can perform simple tasks that a human servant could do, such as fetching things, cleaning, mending, folding clothes, lighting fires, serving food, and pouring wine. Once you give the command, the servant performs the task to the best of its ability until it completes the task, then waits for your next command.</p>
        <p>If you command the servant to perform a task that would move it more than 60 feet away from you, the spell ends.</p>
      `,
      school: 'transmutation',
      isRanged: true,
      range: 60,
      hasAreaOfEffect: false,
    },
    {
      id: 3067,
      german: 'Zungen',
      english: 'Tongues',
      image: 'tongues',
      level: 3,
      isCustom: false,
      classes: ['test', 'bard', 'cleric', 'warlock', 'sorcerer', 'wizard'],
      timeToCast: 1,
      cost: 'action',
      duration: 1,
      durationType: 'hours',
      needsConcentration: false,
      isRitual: false,
      needsVerbal: true,
      needsSomatic: true,
      needsMaterial: true,
      needsAttackRoll: false,
      needsSavingThrow: false,
      doesDamage: false,
      doesHeal: false,
      description_de: `
        <p>Dieser Zauberspruch verleiht der Kreatur, die du berührst, die Fähigkeit, jede gesprochene Sprache zu verstehen, die sie hört. Wenn das Ziel spricht, versteht außerdem jede Kreatur, die mindestens eine Sprache beherrscht und das Ziel hören kann, was es sagt.</p>
        `,
      description_en: `
        <p>This spell grants the creature you touch the ability to understand any spoken language it hears. Moreover, when the target speaks, any creature that knows at least one language and can hear the target understands what it says.</p>
      `,
      school: 'divination',
      isRanged: false,
      range: 5,
      hasAreaOfEffect: false,
    },

    // #endregion

    // #region LEVEL 4

    // #endregion
  ];
}
